"""
Recorded Future TLP:Amber

##################################### TERMS OF USE ###########################################
# The following code is provided for demonstration purpose only, and should not              #
# be used without independent verification. Recorded Future makes no representations         #
# or warranties, express, implied, statutory, or otherwise, regarding any aspect of          #
# this code or of the information it may retrieve, and provides it both strictly “as-is”     #
# and without assuming responsibility for any information it may retrieve. Recorded Future   #
# shall not be liable for, and you assume all risk of using, the foregoing. By using this    #
# code, Customer represents that it is solely responsible for having all necessary licenses, #
# permissions, rights, and/or consents to connect to third party APIs, and that it is solely #
# responsible for having all necessary licenses, permissions, rights, and/or consents to     #
# any data accessed from any third party API.                                                #
##############################################################################################
"""

# standard library
import re
import itertools
from collections.abc import Iterable
from datetime import datetime, timezone

# third-party
from tcex import TcEx
from tcex.api.tc.v3.indicators.indicator import Indicator
from tcex.api.tc.v3.groups.group import Group
from tcex.api.tc.v3.tql.tql_operator import TqlOperator
from tcex.exit import ExitCode
from psengine.collective_insights import (
    RFCollectiveInsights,
    RFInsight,
    CollectiveInsightsError,
)
from psengine.config import Config

# first-party
from app_inputs import TCFiltersModel
from job_app import JobApp
from constants import (
    APP_ID,
    APP_VERSION,
    COLLECTIVE_INSIGHTS_TAG,
    DETECTION_NAME,
    DETECTION_TYPE,
    INCIDENT_TYPE,
    INDICATOR_MAP,
    INDICATOR_TYPES,
    REGEX_MITRE_PATTERN,
)


class App(JobApp):
    """Job App"""

    def __init__(self, _tcex: TcEx):
        """Initialize class properties."""
        super().__init__(_tcex)

    def setup(self) -> None:
        """Perform prep/setup logic."""
        # Recorded Future constants
        self.rf_config = Config()
        self.rf_config.rf_token = self.in_.rf_token
        self.rf_config.app_id = f"{APP_ID}/{APP_VERSION}"
        self.rf_ci = RFCollectiveInsights(self.rf_config)
        self.tc_tag = self.tcex.api.tc.v3.tag(name=COLLECTIVE_INSIGHTS_TAG)
        self.tags_error = False

    def run(self) -> None:
        """Run main App logic."""

        rf_insights = []
        tc_entities = []

        self.tcex.log.info("Fetching ThreatConnect groups and associated indicators")
        for group in self.get_groups(self.in_):
            incident_id = group.model.id
            incident_name = group.model.name
            indicators, malwares, mitre_codes = self.get_group_associations(group)

            if not indicators:
                self.tcex.log.info(
                    f"No compatible indicators associated to '{incident_name}', skipping"
                )
                continue

            # Add ThreatConnect Entities to Collective Insights payload
            self.tcex.log.info(
                f"Found {len(indicators)} indicators associated to group '{incident_name}'"
            )
            tc_entities.extend([indicator for indicator in indicators])
            rf_insights.extend(
                [
                    self.format_threatconnect_entity(
                        entity=indicator,
                        incident_id=incident_id,
                        incident_name=incident_name,
                        malwares=malwares,
                        mitre_codes=mitre_codes,
                    )
                    for indicator in indicators
                ]
            )

        # Exit if no groups
        if not rf_insights:
            self.tcex.log.info(
                "No ThreatConnect indicators to upload to Recorded Future Collective Insights"
            )
            exit_msg = "Exiting process with status: 0"
            self.tcex.exit.exit(ExitCode.SUCCESS, exit_msg)

        ci_size = len(rf_insights)
        self.tcex.log.info(f"Sending {ci_size} indicators to Collective Insights API")
        try:
            self.tcex.log.debug(
                f"Indicators to submit: {[[insight.json() for insight in rf_insights]]}"
            )
            response = self.rf_ci.submit(insight=rf_insights, debug=self.in_.debug)
            self.tcex.log.info(f"Collective Insights API Response: {response}")
        except CollectiveInsightsError as err:
            fail_message = "Error submitting indicators to Collective Insights API"
            self.tcex.log.error(
                f"Error submitting indicators to Collective Insights API: {err}"
            )
            self.tcex.exit.exit(ExitCode.FAILURE, fail_message)

        # Tag ThreatConnect entities after good submit to Collective Insights API
        if not self.in_.debug and self.in_.tag_indicators:
            self.tcex.log.info("Tagging ThreatConnect entities")
            for entity in tc_entities:
                self.tag_threatconnect_entity(tc_entity=entity)

        self.tcex.log.info("Finished ThreatConnect Collective Insights upload")

        # Set exit message and code
        ci_status = response.get("result", {}).get("status", "")
        exit_code = ExitCode.SUCCESS
        exit_msg = f"Collective Insights response: {ci_status}."
        if self.tags_error:
            exit_msg += " Error Applying tag to ThreatConnect entities."
            exit_code = ExitCode.PARTIAL_FAILURE

        if not self.in_.tql:
            self.tcex.app.results_tc(
                "last_modified", datetime.now(timezone.utc).isoformat()
            )

        self.tcex.exit.exit(exit_code, exit_msg)

    def get_groups(self, model: TCFiltersModel, max_results=None) -> Iterable[Group]:
        """Retrieve groups from ThreatConnect based on filter params."""
        groups = self.tcex.api.tc.v3.ti.groups()
        if model.tql:
            # Check to see if user picked one or more owners as we
            # will add those into the TQL. The owner is not already in
            # the TQL as that is checked by pydantic. Can't use params
            # for adding in owner here as that only supports 1 owner and
            # we might have multiple to add.
            tql = model.tql
            if model.owners:
                tql += (
                    " and ownerName IN ("
                    + ", ".join(f"'{item}'" for item in self.in_.owners or [])
                    + ")"
                )

            # Set TQL directly.
            groups.tql.raw_tql = tql
        else:
            groups.filter.owner_name(TqlOperator.IN, model.owners)

            if model.group_types:
                self.tcex.log.info(
                    f"Adding filter rating {TqlOperator.GEQ} {model.group_types}"
                )
                groups.filter.type_name(TqlOperator.IN, model.group_types)

            if model.tags:
                self.tcex.log.debug(f"filtering for tags: {model.tags}")
                groups.filter.has_tag.name(TqlOperator.IN, model.tags)

            if model.last_modified:
                self.tcex.log.info(
                    f"Adding filter modified since = {model.last_modified}"
                )
                groups.filter.last_modified(TqlOperator.GT, model.last_modified)

        if max_results:
            yield from itertools.islice(groups, max_results)
        else:
            yield from groups

    def get_group_associations(self, group: Group) -> tuple:
        """Get group associations for Malwares, MITRE, and indicators.

        Args:
            group (Group): ThreatConnect Group

        Returns:
            indicators, malwares, mitre_codes (list, list, list): Tuple of group associations
                applicable to Collective Insights
        """
        vulns = [
            association
            for association in group.associated_groups
            if association.model.type == "Vulnerability"
        ]
        malwares = [
            association.model.name
            for association in group.associated_groups
            if association.model.type == "Malware"
        ]
        mitre_codes = [
            re.findall(REGEX_MITRE_PATTERN, association.model.name)[0]
            for association in group.associated_groups
            if association.model.type == "Attack Pattern"
        ]
        indicators = [
            indicator
            for indicator in group.associated_indicators
            if indicator.model.type in INDICATOR_TYPES
        ]

        # Add vulns to list of indicators
        indicators.extend(vulns)

        return (indicators, malwares, mitre_codes)

    def format_threatconnect_entity(
        self,
        entity: Indicator | Group,
        incident_id: str,
        incident_name: str,
        malwares: list,
        mitre_codes: list,
    ) -> dict:
        """Format TC Entity into Recorded Future Collective Insights Payload.

        Args:
            entity (Indicator | Group): TCEntity (Indicator or Group in case of Vulnerability)
            incident_id (str): ThreatConnect Group ID
            incident_name (str): ThreatConnect Group Name
            malwares (list): Associated group malwares
            mitre_codes (list): Associated group mitre T-codes

        Returns:
            insight (RFInsight): Formatted insight to send to Collective Insights
        """
        indicator_value = (
            entity.model.summary if isinstance(entity, Indicator) else entity.model.name
        )
        indicator_type = INDICATOR_MAP[entity.model.type]
        if indicator_type == "url" and "amp;" in indicator_value:
            indicator_value = indicator_value.replace("amp;", "&")

        timestamp_dt = entity.model.last_modified.isoformat()
        timestamp = timestamp_dt[: timestamp_dt.index("+")] + "Z"
        insight = RFInsight(
            timestamp=timestamp,
            ioc_value=indicator_value,
            ioc_type=indicator_type,
            incident_id=incident_id,
            incident_name=incident_name,
            incident_type=INCIDENT_TYPE,
            detection_name=DETECTION_NAME,
            detection_type=DETECTION_TYPE,
            malwares=malwares if malwares else None,
            mitre_codes=mitre_codes if mitre_codes else None,
        )
        return insight

    def tag_threatconnect_entity(self, tc_entity: Indicator | Group):
        """Tag ThreatConnect Entity

        Args:
            tc_entity (Indicator, Group): TCEntity sent to Collective Insights API
        """
        try:
            tc_entity.stage_tag(self.tc_tag)
            tc_entity.update()
        except Exception:
            self.tags_error = True
            self.tcex.log.error("Error applying tag to indicator, skipping.")
            pass
