##################################### TERMS OF USE ###########################################
# The following code is provided for demonstration purpose only, and should not be used      #
# without independent verification. Recorded Future makes no representations or warranties,  #
# express, implied, statutory, or otherwise, regarding any aspect of this code or of the     #
# information it may retrieve, and provides it both strictly “as-is” and without assuming    #
# responsibility for any information it may retrieve. Recorded Future shall not be liable    #
# for, and you assume all risk of using, the foregoing. By using this code, Customer         #
# represents that it is solely responsible for having all necessary licenses, permissions,   #
# rights, and/or consents to connect to third party APIs, and that it is solely responsible  #
# for having all necessary licenses, permissions, rights, and/or consents to any data        #
# accessed from any third party API.                                                         #
##############################################################################################

import logging
from copy import deepcopy
from typing import Dict, List, Union

from .constants import URL_TO_PORTAL
from .errors import AnalystNoteError


class RFNote:
    """Class to create an Analyst Note object."""

    def __init__(self, raw_note: dict):
        self.log = logging.getLogger(__name__)
        self._raw = raw_note

        if not self.raw or not isinstance(self.raw, dict):
            raise AnalystNoteError('Invalid note format. Note is empty!')

    def __str__(self) -> str:
        """Str repr."""
        return 'Note ID: {}, Source: {}, Published: {}, Title: {}'.format(
            self.note_id,
            self.source_name,
            self.published,
            self.title,
        )

    def __repr__(self) -> str:
        """Repr method."""
        return 'RFNote(Note ID: {}, Published: {}, Title:{})'.format(
            self.note_id,
            self.published,
            self.title,
        )

    def __eq__(self, other) -> bool:
        """Verify if two notes are the same."""
        return self.note_id == other.note_id

    def __hash__(self) -> int:
        """Hash for set function."""
        return hash(self.note_id)

    def json(self) -> dict:
        """Dict repr."""
        return self.raw

    @property
    def raw(self) -> dict:
        """Get raw note data.

        Returns:
            dict: raw note alert
        """
        return deepcopy(self._raw)

    @property
    def note_id(self) -> str:
        """Get the note_id."""
        return self.raw.get('id')

    @property
    def portal_url(self) -> str:
        """Get the link to portal."""
        return URL_TO_PORTAL.format(self.note_id)

    # Fields from attributes

    @property
    def attributes(self) -> dict:
        """Get the attributes."""
        return self.raw.get('attributes', {})

    @property
    def published(self) -> str:
        """Get the published."""
        return self.attributes.get('published')

    @property
    def title(self) -> str:
        """Get the title."""
        return self.attributes.get('title')

    @property
    def text(self) -> str:
        """Get the text."""
        return self.attributes.get('text')

    @property
    def note_entities(self) -> List[Dict]:
        """Get the note_entities."""
        return self.attributes.get('note_entities', [])

    @property
    def validation_urls(self) -> List[str]:
        """Get the validation_urls."""
        return [url.get('name') for url in self.attributes.get('validation_urls', [])]

    @property
    def attachment(self) -> str:
        """Get the attachment."""
        return self.attributes.get('attachment')

    @property
    def header_image(self) -> str:
        """Get the header_image."""
        return self.attributes.get('header_image', {}).get('name')

    @property
    def recommended_queries_obj(self) -> Dict[str, str]:
        """Get the recommended_queries_obj."""
        return self.attributes.get('recommended_queries_obj', {})

    # Fields from source

    @property
    def source(self) -> dict:
        """Get the source."""
        return self.raw.get('source', {})

    @property
    def source_id(self) -> str:
        """Get the source_id."""
        return self.source.get('id')

    @property
    def source_name(self) -> str:
        """Get the source_name."""
        return self.source.get('name')

    # Fields from events

    @property
    def events(self) -> list:
        """Get the events."""
        return self.raw.get('events', [])

    @property
    def topics(self) -> list:
        """Return the topics list."""
        return self.attributes.get('topic', [])

    @property
    def detection_rule_type(self) -> Union[str, None]:
        """Return the attachment type if present, else None."""
        topics_type = ('sigma rule', 'yara rule', 'snort rule')
        topics = set(topic['name'].lower() for topic in self.topics if topic.get('name'))

        return next(
            (topic_type.split()[0] for topic_type in topics_type if topic_type in topics),
            None,
        )
