##################################### TERMS OF USE ###########################################
# The following code is provided for demonstration purpose only, and should not be used      #
# without independent verification. Recorded Future makes no representations or warranties,  #
# express, implied, statutory, or otherwise, regarding any aspect of this code or of the     #
# information it may retrieve, and provides it both strictly “as-is” and without assuming    #
# responsibility for any information it may retrieve. Recorded Future shall not be liable    #
# for, and you assume all risk of using, the foregoing. By using this code, Customer         #
# represents that it is solely responsible for having all necessary licenses, permissions,   #
# rights, and/or consents to connect to third party APIs, and that it is solely responsible  #
# for having all necessary licenses, permissions, rights, and/or consents to any data        #
# accessed from any third party API.                                                         #
##############################################################################################

import json
import logging
import os
from typing import Dict, List, Tuple, Union

from psengine.analyst_notes.errors import (
    AnalystNoteAttachmentError,
    AnalystNoteDeleteError,
    AnalystNoteLookupError,
    AnalystNotePreviewError,
    AnalystNotePublishError,
    AnalystNoteSearchError,
)
from psengine.config.config import Config
from psengine.constants import (
    DEFAULT_KEY,
    ERR_MSG_KEY,
    ROOT_DIR,
    STANZA_FIELDS_KEY,
    TYPE_KEY,
    VALIDATOR_KEY,
)
from psengine.endpoints import (
    EP_ANALYST_NOTE_ATTACHMENT,
    EP_ANALYST_NOTE_DELETE,
    EP_ANALYST_NOTE_ID,
    EP_ANALYST_NOTE_PREVIEW,
    EP_ANALYST_NOTE_PUBLISH,
    EP_ANALYST_NOTE_SEARCH,
)
from psengine.exceptions import WriteFileError
from psengine.helpers.helpers import (
    OSHelpers,
    TimeHelpers,
    basic_args_validation,
    connection_exceptions,
)
from psengine.logger import debug_call
from psengine.rf_client import RFClient

from .rf_note import RFNote

ANALYST_NOTES_STANZA_NAME = 'analyst_notes'
OUTPUT_DIR_SETTING = 'output_dir'
DEFAULT_ALERTS_OUTPUT_DIR = os.path.join(ROOT_DIR, 'analyst_notes')
LIMIT_SETTING = 'limit'
LIMIT_DEFAULT = 10
PUBLISHED_SETTING = 'published'
PUBLISHED_DEFAULT = '-1d'
TOPICS_SETTING = 'topics'
TOPICS_DEFAULT = []

ANALYST_NOTES_MGR_PARAMS = {
    STANZA_FIELDS_KEY: {
        OUTPUT_DIR_SETTING: {
            DEFAULT_KEY: DEFAULT_ALERTS_OUTPUT_DIR,
            TYPE_KEY: str,
            VALIDATOR_KEY: OSHelpers.check_dir,
            ERR_MSG_KEY: "[{}] '{}' directory is invalid or has restricted permissions".format(
                ANALYST_NOTES_STANZA_NAME,
                OUTPUT_DIR_SETTING,
            ),
        },
        LIMIT_SETTING: {
            DEFAULT_KEY: LIMIT_DEFAULT,
            TYPE_KEY: int,
            VALIDATOR_KEY: lambda x: x > 0,
            ERR_MSG_KEY: "[{}] '{}' is not a positive integer".format(
                ANALYST_NOTES_STANZA_NAME,
                LIMIT_SETTING,
            ),
        },
        PUBLISHED_SETTING: {
            DEFAULT_KEY: PUBLISHED_DEFAULT,
            TYPE_KEY: str,
            VALIDATOR_KEY: TimeHelpers.is_rel_time_valid,
            ERR_MSG_KEY: "[{}] '{}' is not a valid relative time".format(
                ANALYST_NOTES_STANZA_NAME,
                PUBLISHED_SETTING,
            ),
        },
        TOPICS_SETTING: {
            DEFAULT_KEY: TOPICS_DEFAULT,
            TYPE_KEY: list,
        },
    },
}


class RFAnalystNoteMgr:
    """Manages requests for Recorded Future analyst notes."""

    def __init__(self, cfg: Config = None, **kwargs):
        """Initialize the RFAnalystNoteMgr.

        Args:
            cfg (Config, optional): Config object with a analyst_notes stanza. Defaults to None.
        """
        self.log = logging.getLogger(__name__)
        self.config = cfg if cfg is not None else Config()
        self.config.parse_stanza(ANALYST_NOTES_STANZA_NAME, ANALYST_NOTES_MGR_PARAMS)
        self.config.update_settings(
            [(ANALYST_NOTES_STANZA_NAME, ANALYST_NOTES_MGR_PARAMS)],
            kwargs,
        )
        self.rf_client = RFClient(self.config)

    @debug_call
    @basic_args_validation
    def search(
        self,
        published: str = '',
        entity: str = '',
        author: str = '',
        title: str = '',
        topic: Union[str, list] = '',
        label: str = '',
        source: str = '',
        taggedText: bool = False,
        serialization: str = '',
        limit: int = 0,
        from_: int = 0,
        **kwargs,
    ) -> List[RFNote]:
        """Execute a search for the analyst notes based on the parameters provided. Every parameter
        that has not been set up will be discarded.
        If more than one topic is specified, a search for each topic is executed and the RFNotes
        will be deduplicated.

        The ``from_`` field is automatically translated into ``from`` field.

        ``limit`` is the maximum number of references, not notes.

        Args:
            - published (str): Notes published after this date. Defaults to -1d.
            - entity (str): Notes referring entity, RF ID.
            - author (str): Notes by author, RF ID.
            - title (str): Notes by title.
            - topic (Union[str, list]): Notes by topic, RF ID.
            - label (str): Notes by label, by name.
            - source (str): source of note.
            - taggedText (bool): Should text contain tags. Defaults to False.
            - serialization (str): Entity serializer (id, min, full, raw).
            - limit (int): Maximum number of references from which notes are fetched, at most 1000.
            - from_ (int): Offset from first result. The API may only return the first 1000
              results, meaning limit + from_ cannot exceed 1000.


        Returns:
            List[RFNote]: List of deduplicated RFNotes.
        """
        responses = []

        if not limit:
            limit = self.limit

        if not published:
            published = self.published

        param = {key: val for key, val in locals().items() if val and key != 'self'}
        if param.get('from_'):
            param['from'] = param['from_']
            del param['from_']

        if isinstance(topic, list) and len(topic):
            for t in topic:
                param['topic'] = t
                responses.append(self._search(param))
            return list(set(RFNote(result) for response in responses for result in response))

        return list(set(RFNote(result) for result in self._search(param)))

    @debug_call
    @basic_args_validation
    @connection_exceptions(
        ignore_status_code=[404],
        exception_to_raise=AnalystNoteLookupError,
    )
    def lookup(self, note_id: str) -> RFNote:
        """Lookup an analyst note by ID.

        Args:
            - note_id (str): The ID of the analyst note to lookup

        Returns:
            str: The response from the Recorded Future API

        Raises:
            AnalystNoteLookupError if connection error occours
        """
        self.log.info('Looking up analyst note: {}'.format(note_id))
        response = self.rf_client.get_request(EP_ANALYST_NOTE_ID + note_id).json()

        return RFNote(response)

    @debug_call
    @basic_args_validation
    @connection_exceptions(
        ignore_status_code=[404],
        exception_to_raise=AnalystNoteDeleteError,
    )
    def delete(self, note_id: str) -> Union[bool, None]:
        """Delete Analyst Note.

        Args:
            - note_id (str)

        Returns:
            Union[bool, None]: True if delete ok else None

        Raises:
            AnalystNoteDeleteError: if connection error

        """
        self.log.info('Deleting note {}'.format(note_id))
        self.rf_client.delete_request(url=EP_ANALYST_NOTE_DELETE.format(note_id))
        self.log.debug('Note {} deleted succesfully'.format(note_id))
        return True

    @debug_call
    @basic_args_validation
    @connection_exceptions(
        ignore_status_code=[404],
        exception_to_raise=AnalystNotePreviewError,
    )
    def preview(self, data: dict) -> RFNote:
        """Preview of the RFNote. This method does not create a note, it just return how the note
        will look like.

        Args:
            - data (dict): data to be posted. Should look like::

                data = {
                    "title": "Malicious Activity from My Domain",
                    "text": "My domain test.com has been seen communicating with C2 8.8.8.8",
                    "context_entities": ["idn:test.com", "ip:8.8.8.8"],
                    "note_entities": [],
                    "entities": []
                }

            If the entities are not known can be omitted::

                data = {
                    "title": "Malicious Activity from My Domain",
                    "text": "My domain test.com has been seen communicating with C2 8.8.8.8",
                }

        Returns:
            RFNote: rfnote that will be created.

        Raises:
            TypeError: if data is not valid type
        """
        self.log.info('Preview started.')
        resp = self.rf_client.post_request(EP_ANALYST_NOTE_PREVIEW, data=json.dumps(data)).json()
        self.log.debug('Preview output: {}'.format(json.dumps(resp, indent=4)))

        return RFNote(resp)

    @debug_call
    @basic_args_validation
    @connection_exceptions(ignore_status_code=[404], exception_to_raise=AnalystNotePublishError)
    def publish(self, data: dict) -> Dict[str, str]:
        """Publish of data. This method does create a note and returns the id.

        Args:
            - data (dict): example::

                data = {
                    "title": "test pytest",
                    "text": "test related to test.com",
                    "context_entities": ["idn:test.com"],
                    "note_entities": [],
                    "entities": [],
                }

        Returns:
            dict: id that will be returned

        Raises:
            TypeError: if data is not valid type
        """
        self.log.info('Publish started.')
        resp = self.rf_client.post_request(EP_ANALYST_NOTE_PUBLISH, data=json.dumps(data)).json()
        self.log.debug('Publish output: {}'.format(json.dumps(resp, indent=4)))

        return resp

    @debug_call
    @basic_args_validation
    @connection_exceptions(ignore_status_code=[404], exception_to_raise=AnalystNoteAttachmentError)
    def fetch_attachment(self, note_id: str) -> Tuple[bytes, str]:
        """Get Analyst Note Attachment. To work with the attachment is the same no matter the ext::

            # note with pdf attachment
            attachment, extension = note_mgr.fetch_attachment('tPtLVw')
            note_mgr.save_attachment('tPtLVw', attachment, extension)

            # note with yar attachment
            attachment, extension = note_mgr.fetch_attachment('oJeqDP')
            note_mgr.save_attachment('oJeqDP', attachment, extension)

        Args:
            - note_id (str): id of the note

        Returns:
            Tuple[bytes, str]: content of the attachment in bytes and extension of the file

        Raises:
            AnalystNoteAttachmentError if connection error occours
            TypeError: if note_id is not a str or empty
        """
        self.log.info("Looking up analyst note's {} attachment".format(note_id))
        response = self.rf_client.get_request(EP_ANALYST_NOTE_ATTACHMENT.format(note_id))

        ext = (
            'pdf'
            if 'application/pdf' in response.headers.get('Content-Type')
            else response.headers.get('Content-Disposition', 'xxx')[-3:]
        )
        return response.content, ext

    @debug_call
    @basic_args_validation
    def save_attachment(self, note_id: str, data: Union[bytes, str], ext: str) -> None:
        """Save yara, sigma, snort or pdf to file. The file will use the extension provided and the
        ``note_id`` to create the filename.

        Args:
            - note_id (str): ``RFNote`` id
            - data (Union[bytes, str]): data, returned from ``fetch_attachment``
            - ext (str): extension of the attachment, returned by ``fetch_attachment``


        """
        self._save_attachment(note_id, data, ext)

    @debug_call
    @basic_args_validation
    def save_note(self, rf_note: RFNote) -> None:
        """Save RFNote object on file. The file will have the name of the note id.

        Args:
            - rf_note (RFNote): note to save.

        """
        self._save_attachment(
            note_id=rf_note.note_id,
            data=json.dumps(rf_note.json(), indent=4),
            ext='json',
        )

    @basic_args_validation
    def _save_attachment(self, note_id: str, data: Union[bytes, str], ext) -> None:
        """Save attachment from bytes or note itself from json.

        Args:
            note_id (str): id of the note, will be the filename
            data (bytes | str): content to save
            ext (str): extension of the file.

        Raises:
            WriteFileError: if saving to file fails
        """
        try:
            self.log.debug("Saving file related to '{}' to disk".format(note_id))

            dir_path = OSHelpers.mkdir(self.output_dir)
            note_path = os.path.join(dir_path, '{}.{}'.format(note_id, ext))
            open_mode = 'wb' if isinstance(data, bytes) else 'w'

            with open(note_path, open_mode) as file:
                file.write(data)
        except OSError as err:
            raise WriteFileError('Failed to save file to disk. Cause: {}'.format(err.args))

    @connection_exceptions(ignore_status_code=[404], exception_to_raise=AnalystNoteSearchError)
    def _search(self, param: dict) -> dict:
        """Search for Analayst notes.

        Raises:
            AnalystNoteSearchError: if errror with the request

        Return:
            dict: json data for 'data.results' for the analayst notes found by the search.

        """
        url = '{}?{}'.format(
            EP_ANALYST_NOTE_SEARCH,
            '&'.join('{}={}'.format(key, val) for key, val in param.items()),
        )
        self.log.info('Searching analyst notes with query: {}'.format(param))
        response = self.rf_client.get_request(url).json()['data']['results']
        return response

    @property
    def output_dir(self) -> str:
        """Getter for output_dir."""
        output_dir = self.config.settings.get(ANALYST_NOTES_STANZA_NAME, {}).get(OUTPUT_DIR_SETTING)
        if output_dir is None:
            output_dir = DEFAULT_ALERTS_OUTPUT_DIR
            self.log.warning(
                '{} was not set, using default: {}'.format(OUTPUT_DIR_SETTING, output_dir),
            )
        return output_dir

    @property
    def limit(self) -> str:
        """Getter for limit."""
        limit = self.config.settings.get(ANALYST_NOTES_STANZA_NAME, {}).get(LIMIT_SETTING)
        if limit is None:
            limit = LIMIT_DEFAULT
            self.log.warning('{} was not set, using default: {}'.format(LIMIT_SETTING, limit))
        return limit

    @property
    def published(self) -> str:
        """Getter for published."""
        published = self.config.settings.get(ANALYST_NOTES_STANZA_NAME, {}).get(PUBLISHED_SETTING)
        if published is None:
            published = PUBLISHED_DEFAULT
            self.log.warning(
                '{} was not set, using default: {}'.format(PUBLISHED_SETTING, published),
            )
        return published

    @property
    def topics(self) -> str:
        """Getter for topics."""
        topics = self.config.settings.get(ANALYST_NOTES_STANZA_NAME, {}).get(TOPICS_SETTING)
        if topics is None:
            topics = TOPICS_DEFAULT
            self.log.warning('{} was not set, using default: {}'.format(TOPICS_SETTING, topics))
        return topics
