##################################### TERMS OF USE ###########################################
# The following code is provided for demonstration purpose only, and should not be used      #
# without independent verification. Recorded Future makes no representations or warranties,  #
# express, implied, statutory, or otherwise, regarding any aspect of this code or of the     #
# information it may retrieve, and provides it both strictly “as-is” and without assuming    #
# responsibility for any information it may retrieve. Recorded Future shall not be liable    #
# for, and you assume all risk of using, the foregoing. By using this code, Customer         #
# represents that it is solely responsible for having all necessary licenses, permissions,   #
# rights, and/or consents to connect to third party APIs, and that it is solely responsible  #
# for having all necessary licenses, permissions, rights, and/or consents to any data        #
# accessed from any third party API.                                                         #
##############################################################################################

import json
import logging

from requests import ConnectionError, ConnectTimeout, HTTPError, ReadTimeout

from ..config import Config
from ..constants import (
    DEFAULT_KEY,
    ERR_MSG_KEY,
    RF_TOKEN_KWARG,
    STANZA_FIELDS_KEY,
    TYPE_KEY,
    VALIDATOR_KEY,
)
from ..endpoints import EP_COLLECTIVE_INSIGHTS_DETECTIONS
from ..rf_client import RFClient
from .errors import CollectiveInsightsError
from .rf_insight import RFInsight

COLLECTIVE_INSIGHTS_STANZA_NAME = 'collective_insights'

ENABLED_SETTING = 'enabled'
ENABLED_DEFAULT = False

ORGANIZATION_IDS_SETTING = 'organization_ids'
ORGANIZATION_IDS_DEFAULT = None

DEBUG_SETTING = 'debug'
DEBUG_DEFAULT = True

SUMMARY_DEFAULT = True

COLLECTIVE_INSIGHTS_STANZA_PARAMS = {
    STANZA_FIELDS_KEY: {
        ENABLED_SETTING: {
            DEFAULT_KEY: ENABLED_DEFAULT,
            TYPE_KEY: bool,
            VALIDATOR_KEY: lambda x: x in [True, False],
            ERR_MSG_KEY: "[{}] '{}' must be one of [true, false]".format(
                COLLECTIVE_INSIGHTS_STANZA_NAME,
                ENABLED_SETTING,
            ),
        },
        ORGANIZATION_IDS_SETTING: {
            DEFAULT_KEY: ORGANIZATION_IDS_DEFAULT,
            TYPE_KEY: list,
            VALIDATOR_KEY: lambda x: isinstance(x, list) and all(isinstance(y, str) for y in x),
            ERR_MSG_KEY: "[{}] '{}' must be a list of string organization ids".format(
                COLLECTIVE_INSIGHTS_STANZA_NAME,
                ORGANIZATION_IDS_SETTING,
            ),
        },
        DEBUG_SETTING: {
            DEFAULT_KEY: DEBUG_DEFAULT,
            TYPE_KEY: bool,
            VALIDATOR_KEY: lambda x: x in [True, False],
            ERR_MSG_KEY: "[{}] '{}' must be one of [true, false]".format(
                COLLECTIVE_INSIGHTS_STANZA_NAME,
                DEBUG_SETTING,
            ),
        },
    },
}


class RFCollectiveInsights:
    """Class for interacting with the Recorded Future Collective Insights API."""

    def __init__(self, cfg: Config = None, **kwargs):
        """Initialize RFCollectiveInsights.

        Args:
            cfg (Config, optional): Config object. Defaults to None.
            rf_token (str, optional): Recorded Future API token. Defaults to None.
            organization_ids (list, optional): List of organization ids. Defaults to None.
            debug (bool, optional): Enable debug insight submission. Defaults to True.
        """
        self.log = logging.getLogger(__name__)
        self.config = cfg if cfg is not None else Config()
        rf_token = kwargs.pop(RF_TOKEN_KWARG) if RF_TOKEN_KWARG in kwargs else None
        self.config.parse_stanza(COLLECTIVE_INSIGHTS_STANZA_NAME, COLLECTIVE_INSIGHTS_STANZA_PARAMS)
        self.config.update_settings(
            [(COLLECTIVE_INSIGHTS_STANZA_NAME, COLLECTIVE_INSIGHTS_STANZA_PARAMS)],
            kwargs,
        )
        self.rf_client = (
            RFClient(self.config, rf_token=rf_token)
            if rf_token is not None
            else RFClient(self.config)
        )
        self.config.log_settings(COLLECTIVE_INSIGHTS_STANZA_NAME)

    def submit(
        self,
        insight: list or RFInsight,
        debug: bool = None,
        organization_ids: list = None,
    ):
        """Submit a detection/insight to Recorded Future Collective Insights API.

        Args:
            insight (list[RFInsight] or RFInsight): A detection/insight
            debug (bool, optional): Determines if submission will show in SecOPS dashboard.
            organization_ids (list, optional): Org ID. Defaults to None.

        Raises:
            TypeError: when supplied params are invalid
            ValueError: when supplied params are invalid
            CollectiveInsightsError: when submission fails

        Returns:
            dict: response from Recorded Future API
        """
        insight = insight if isinstance(insight, list) else [insight]

        if not len(insight):
            raise ValueError('insight must be a list of RFInsight objects or an RFInsight object')

        if not all(isinstance(ins, RFInsight) for ins in insight):
            raise TypeError('insight must be a list of RFInsight objects or an RFInsight object')

        param = self._prepare_params_for_submission(insight, debug, organization_ids)
        try:
            self.log.info(
                'Submitting {} insight(s) with DEBUG: {}'.format(
                    len(insight),
                    param['options']['debug'],
                ),
            )
            response = self.rf_client.post_request(
                url=EP_COLLECTIVE_INSIGHTS_DETECTIONS,
                data=json.dumps(param),
            )
        except (ConnectionError, ConnectTimeout, ReadTimeout, HTTPError) as err:
            raise CollectiveInsightsError(str(err))

        try:
            json_response = response.json()
        except json.JSONDecodeError as err:
            raise CollectiveInsightsError(str(err))

        status = json_response.get('result', {}).get('status', 'N/A')
        processed = json_response.get('result', {}).get('summary', {}).get('processed', 'N/A')
        self.log.info('Status: {}, Processed: {}'.format(status, processed))
        return json_response

    def _prepare_params_for_submission(
        self,
        insight: list or RFInsight,
        debug: bool = None,
        organization_ids: list = None,
    ) -> dict:
        params = {'options': {}}

        params['data'] = (
            [insight.json()] if isinstance(insight, RFInsight) else [ins.json() for ins in insight]
        )

        organization_ids = (
            organization_ids if organization_ids is not None else self.organization_ids
        )

        if organization_ids is not None:
            if isinstance(organization_ids, list):
                if len(organization_ids) > 0:
                    params['organization_ids'] = organization_ids
            else:
                raise TypeError(
                    'organization_ids must be a list of string organization ids or None',
                )

        params['options']['debug'] = self.debug if debug is None else debug

        # We always have summary of the submission
        params['options']['summary'] = SUMMARY_DEFAULT

        self.log.debug('Params for submission: \n{}'.format(json.dumps(params, indent=2)))

        return params

    @property
    def enabled(self):
        """Getter for enabled."""
        enabled_value = self.config.settings.get(COLLECTIVE_INSIGHTS_STANZA_NAME, {}).get(
            ENABLED_SETTING,
        )
        return enabled_value

    @enabled.setter
    def enabled(self, value) -> None:
        """Setter for enabled."""
        if not isinstance(value, bool):
            raise TypeError('enabled must be a bool')
        self.config.update_setting(COLLECTIVE_INSIGHTS_STANZA_NAME, ENABLED_SETTING, value)

    @property
    def organization_ids(self):
        """Getter for organization_ids."""
        organization_ids_value = self.config.settings.get(COLLECTIVE_INSIGHTS_STANZA_NAME, {}).get(
            ORGANIZATION_IDS_SETTING,
        )
        return organization_ids_value

    @organization_ids.setter
    def organization_ids(self, value) -> None:
        """Setter for organization_ids."""
        if not isinstance(value, list) or not value:
            raise TypeError('organization_ids must be a list of string names')
        if not COLLECTIVE_INSIGHTS_STANZA_PARAMS[STANZA_FIELDS_KEY][ORGANIZATION_IDS_SETTING][
            VALIDATOR_KEY
        ](value):
            raise ValueError('Invalid value for organization_ids: {}'.format(value))
        self.config.update_setting(COLLECTIVE_INSIGHTS_STANZA_NAME, ORGANIZATION_IDS_SETTING, value)

    @property
    def debug(self):
        """Getter for debug."""
        debug_value = self.config.settings.get(COLLECTIVE_INSIGHTS_STANZA_NAME, {}).get(
            DEBUG_SETTING,
        )
        return debug_value

    @debug.setter
    def debug(self, value) -> None:
        """Setter for debug."""
        if not isinstance(value, bool):
            raise TypeError('debug must be a bool')
        self.config.update_setting(COLLECTIVE_INSIGHTS_STANZA_NAME, DEBUG_SETTING, value)
