##################################### TERMS OF USE ###########################################
# The following code is provided for demonstration purpose only, and should not be used      #
# without independent verification. Recorded Future makes no representations or warranties,  #
# express, implied, statutory, or otherwise, regarding any aspect of this code or of the     #
# information it may retrieve, and provides it both strictly “as-is” and without assuming    #
# responsibility for any information it may retrieve. Recorded Future shall not be liable    #
# for, and you assume all risk of using, the foregoing. By using this code, Customer         #
# represents that it is solely responsible for having all necessary licenses, permissions,   #
# rights, and/or consents to connect to third party APIs, and that it is solely responsible  #
# for having all necessary licenses, permissions, rights, and/or consents to any data        #
# accessed from any third party API.                                                         #
##############################################################################################

import datetime
import logging

from dateutil.parser import ParserError, parse

from .constants import (
    DETECTION_SUB_FORMAT_MAPPING,
    DETECTION_TYPE_RULE,
    TIMESTAMP_FORMAT,
    VALID_DETECTION_RULE_SUB_TYPES,
    VALID_DETECTION_TYPES,
    VALID_ENTITY_TYPES,
)


class RFInsight:
    """Insight class to be used for Recorded Future Collective Insights."""

    def __init__(
        self,
        ioc_value: str,
        ioc_type: str,
        detection_type: str,
        detection_sub_type: str = None,
        detection_id: str = None,
        detection_name: str = None,
        ioc_field: str = None,
        ioc_source_type: str = None,
        timestamp: str = None,
        incident_id: str = None,
        incident_name: str = None,
        incident_type: str = None,
        mitre_codes: list or str = None,
        malwares: list or str = None,
        **kwargs,
    ):
        self.log = logging.getLogger(__name__)
        self._ioc_value = ioc_value
        self._ioc_type = ioc_type
        self._detection_type = detection_type
        self._detection_sub_type = detection_sub_type
        self._detection_id = detection_id
        self._detection_name = detection_name
        self._ioc_field = ioc_field
        self._ioc_source_type = ioc_source_type
        self._timestamp = self._validate_timestamp(timestamp)
        self._incident_id = incident_id
        self._incident_name = incident_name
        self._incident_type = incident_type
        self._mitre_codes = mitre_codes
        self._malwares = malwares
        self._kwargs = kwargs

        self._validate()

    def _validate(self) -> None:
        """Validate the detection."""
        required_attrs = ('ioc_value', 'ioc_type', 'detection_type')
        for attr in required_attrs:
            if not getattr(self, attr, None):
                raise ValueError('{} is required.'.format(attr))

        valid_values = {'ioc_type': VALID_ENTITY_TYPES, 'detection_type': VALID_DETECTION_TYPES}
        for attr, valid_type in valid_values.items():
            value = getattr(self, attr)
            if value not in valid_type:
                raise ValueError('{} is invalid.'.format(attr))

        for attr in ('mitre_codes', 'malwares'):
            value = getattr(self, attr, None)
            if value and not isinstance(value, (list, str)):
                raise TypeError('{} must be a list of strings or a string'.format(attr))

        self._validate_detection()

    def _validate_detection(self) -> None:
        if self.detection_type == DETECTION_TYPE_RULE:
            if self.detection_sub_type not in VALID_DETECTION_RULE_SUB_TYPES:
                raise ValueError(
                    "detection_sub_type must be one of {} when detection_type is '{}'".format(
                        VALID_DETECTION_RULE_SUB_TYPES,
                        DETECTION_TYPE_RULE,
                    ),
                )

            # Because its a detection rule the detection_id must start with 'doc:'
            # it is the way we connect the analyst note to the detection. Since it is
            # an RF made sigma/yara/snort rule
            if not self.detection_id or not self.detection_id.startswith('doc:'):
                raise ValueError(
                    "detection_id is required when detection_type is '{}'"
                    "and must start with 'doc:'.".format(DETECTION_TYPE_RULE),
                )

    def _validate_timestamp(self, timestamp):
        if timestamp:
            try:
                datetime.datetime.strptime(timestamp, TIMESTAMP_FORMAT)
                return timestamp
            except ValueError:
                try:
                    wrong_timestamp = parse(timestamp, fuzzy=True)
                    return wrong_timestamp.strftime(TIMESTAMP_FORMAT)
                except ParserError:
                    raise ValueError('timestamp must be in the format YYYY-MM-DDTHH:MM:SSSZ')

    def json(self) -> dict:
        """Returns the detection as a JSON/dict.

        Returns:
            dict: The detection as a JSON/dict
        """
        ioc = {'ioc': {}, 'incident': {}, 'detection': {}}

        for attr, keys in DETECTION_SUB_FORMAT_MAPPING.items():
            value = getattr(self, attr, None)
            if value:
                ioc[keys[0]][keys[1]] = value

        ioc['detection']['sub_type'] = getattr(self, 'detection_sub_type', '') or ''

        for attr in ['malwares', 'mitre_codes']:
            value = getattr(self, attr, None)
            if value:
                ioc[attr] = [value] if isinstance(value, str) else value

        if self.timestamp:
            ioc['timestamp'] = self.timestamp
        if self.kwargs:
            ioc.update(self.kwargs)

        ioc = {key: value for key, value in ioc.items() if value}

        return ioc

    def __str__(self) -> str:
        """Returns the detection as a summarized string.

        Returns:
            str: The detection as a summarized string
        """
        return '{}, {}, {}, {}'.format(
            self.ioc_value,
            self.ioc_type,
            self.detection_type,
            self.detection_sub_type,
        )

    @property
    def ioc_value(self) -> str:
        """Returns the IOC value."""
        return self._ioc_value

    @property
    def ioc_type(self) -> str:
        """Returns the IOC type."""
        return self._ioc_type

    @property
    def detection_type(self) -> str:
        """Returns the detection type."""
        return self._detection_type

    @property
    def detection_sub_type(self) -> str:
        """Returns the detection sub type."""
        return self._detection_sub_type

    @property
    def detection_id(self) -> str:
        """Returns the detection id."""
        return self._detection_id

    @property
    def detection_name(self) -> str:
        """Returns the detection name."""
        return self._detection_name

    @property
    def ioc_field(self) -> str:
        """Returns the IOC field."""
        return self._ioc_field

    @property
    def ioc_source_type(self) -> str:
        """Returns the IOC source type."""
        return self._ioc_source_type

    @property
    def timestamp(self) -> str:
        """Returns the timestamp."""
        return self._timestamp

    @property
    def incident_id(self) -> str:
        """Returns the incident id."""
        return self._incident_id

    @property
    def incident_name(self) -> str:
        """Returns the incident name."""
        return self._incident_name

    @property
    def incident_type(self) -> str:
        """Returns the incident type."""
        return self._incident_type

    @property
    def mitre_codes(self) -> list:
        """Returns the MITRE codes."""
        return self._mitre_codes

    @property
    def malwares(self) -> list:
        """Returns the malwares."""
        return self._malwares

    @property
    def kwargs(self) -> dict:
        """Returns the kwargs."""
        return self._kwargs
