##################################### TERMS OF USE ###########################################
# The following code is provided for demonstration purpose only, and should not be used      #
# without independent verification. Recorded Future makes no representations or warranties,  #
# express, implied, statutory, or otherwise, regarding any aspect of this code or of the     #
# information it may retrieve, and provides it both strictly “as-is” and without assuming    #
# responsibility for any information it may retrieve. Recorded Future shall not be liable    #
# for, and you assume all risk of using, the foregoing. By using this code, Customer         #
# represents that it is solely responsible for having all necessary licenses, permissions,   #
# rights, and/or consents to connect to third party APIs, and that it is solely responsible  #
# for having all necessary licenses, permissions, rights, and/or consents to any data        #
# accessed from any third party API.                                                         #
##############################################################################################

import configparser
import json
import logging
import os
import re
import sys
from distutils import util
from json import JSONDecodeError

from ..constants import (
    DEFAULT_KEY,
    ERR_MSG_KEY,
    ROOT_DIR,
    STANZA_FIELDS_KEY,
    TYPE_KEY,
    VALIDATOR_KEY,
)
from ..exceptions import WriteFileError
from .errors import ConfigError

DEFAULT_SETTINGS_FILE = 'settings.ini'


class Config:
    """Main class to store & access the configuration."""

    def __init__(self, cfg_filepath: str = None):
        self.log = logging.getLogger(__name__)
        self.__rf_token = None
        self.__app_id = None
        self.__platform_id = None
        self.settings = {}

        if cfg_filepath is not None:
            if not os.path.isabs(cfg_filepath):
                cfg_filepath = os.path.join(ROOT_DIR, cfg_filepath)
            self.cfg_filepath = cfg_filepath
            self.cfg = self._read_cfg_file(cfg_filepath)
            self.settings = self._parse()
        else:
            self.cfg = configparser.ConfigParser()
            self.settings = {}

    def log_settings(self, stanza_name) -> None:
        """Log settings for a given stanza."""
        try:
            # Get calling class name from the stack
            caller = sys._getframe(1).f_locals['self'].__class__.__name__
        except (AttributeError, KeyError, ValueError) as err:
            # Unable to get calling class name, use the stanza name instead
            caller = stanza_name
            self.log.debug(err)

        self.log.info('{} initialized settings:'.format(caller))
        for setting, value in self.settings.get(stanza_name).items():
            self.log.info('  {}: {}'.format(setting, value))

    def save(self) -> None:
        """Save current configuration to disk.
        This will overwrite the configuration file that is loaded.

        Raises:
            WriteFileError: Raised if settings cannot be saved
        """
        self.log.debug('Saving settings')
        out_cfg = configparser.ConfigParser()
        for section, settings in self.settings.items():
            if settings != {}:
                out_cfg.add_section(section)
                for key, val in settings.items():
                    if isinstance(val, (dict, list)):
                        out_cfg.set(section, key, json.dumps(val))
                    else:
                        out_cfg.set(section, key, str(val))

        try:
            with open(self.cfg_filepath, 'w') as outfile:
                out_cfg.write(outfile)
        except OSError as err:
            raise WriteFileError('Failed to save configuration to file. Cause: {}'.format(err))

        self.log.debug('Settings saved')

    def update_setting(self, stanza, setting, value) -> None:
        """Update setting for a stanza.

        Args:
            stanza (str): Stanza to update
            setting (str): Setting within stanza to update
            value (any): Value to update setting to
        """
        self.log.debug('Updating [{}].{} with {}'.format(stanza, setting, value))
        if stanza not in self.settings:
            self.settings[stanza] = {}
        self.settings[stanza][setting] = value

    def update_settings(self, stanzas: list, kwargs: dict) -> None:  # noqa: C901
        """Update settings to supplied kwargs for a list of stanzas.
        s
                Args:
                    stanzas (list): List of stanza settings dictionaries
                    kwargs (dict): Kwargs dict from class constructor

        Raises:
                    ValueError: Stanzas must be a list of dicts
        """
        if not all(isinstance(stanza, tuple) for stanza in stanzas):
            raise ValueError('Stanzas must be a list of (name, dict) tuples')
        if not all(isinstance(name, str) and isinstance(stanza, dict) for name, stanza in stanzas):
            raise ValueError('stanzas must be a list of dicts')
        for key, value in kwargs.items():
            updated = False
            for name, stanza in stanzas:
                if key in stanza[STANZA_FIELDS_KEY]:
                    if TYPE_KEY in stanza[STANZA_FIELDS_KEY][key] and not isinstance(
                        value,
                        stanza[STANZA_FIELDS_KEY][key][TYPE_KEY],
                    ):
                        raise TypeError(
                            'Invalid type for {}.{}: {}'.format(
                                name,
                                key,
                                stanza[STANZA_FIELDS_KEY][key][TYPE_KEY],
                            ),
                        )
                    if (
                        VALIDATOR_KEY in stanza[STANZA_FIELDS_KEY][key]
                        and ERR_MSG_KEY in stanza[STANZA_FIELDS_KEY][key]
                    ) and not stanza[STANZA_FIELDS_KEY][key][VALIDATOR_KEY](value):
                        raise ValueError(stanza[STANZA_FIELDS_KEY][key][ERR_MSG_KEY])
                    self.update_setting(name, key, value)
                    updated = True
            if not updated:
                self.log.warning('Ignoring unrecognized parameter: {}'.format(key))

    def parse_stanza(self, stanza_name, parameters) -> None:
        """Helper function to easily parse a stanzas. Simply pass in
        the stanza you wish to parse and a parameters dictionary
        specifying how to parse it.


        Args:
            stanza_name (str): stanza name
            parameters (dict): required parameters

        Raises:
            ConfigError: if parsing or validation fails
        """
        if stanza_name not in self.settings:
            # Stanza does not exist in the supplied config
            # create it with no fields/parameters. This happens in the following cases:
            # Config loaded without any settings.ini file or the stanza was not present.
            self.settings[stanza_name] = {}

        for field, constraints in parameters[STANZA_FIELDS_KEY].items():
            if field not in self.settings[stanza_name]:
                # Field is not present in the supplied config, so simply use the default value
                # without validating it
                if DEFAULT_KEY not in constraints:
                    raise ConfigError(
                        'Library missing default value for field: {}.{}'.format(stanza_name, field),
                    )
                value = constraints[DEFAULT_KEY]
            else:
                # Field is present in the supplied config, so set it to the correct type
                # Raise TypeError if type conversion fails
                value = self.__set_by_type(
                    field,
                    self.settings[stanza_name][field],
                    constraints[TYPE_KEY],
                )
                # Validator will raise ValueError if validation fails
                self.__run_validator(parameters[STANZA_FIELDS_KEY], field, value)
            self.settings[stanza_name][field] = value

    def _read_cfg_file(self, cfg_file):
        cfg_file = self._validate_cfg_filepath(cfg_file)
        self.log.debug("Reading configuration file '{}'".format(cfg_file))
        p_cfg = configparser.ConfigParser(interpolation=configparser.ExtendedInterpolation())
        p_cfg.read(cfg_file)

        return p_cfg

    def _validate_cfg_filepath(self, cfg_filepath):
        if os.path.exists(cfg_filepath):
            return cfg_filepath

        cfg_filepath = os.path.join(ROOT_DIR, cfg_filepath)
        if not os.path.exists(cfg_filepath):
            raise ConfigError(
                'Configuration file not found {}'.format(os.path.join(ROOT_DIR, cfg_filepath)),
            )

        return cfg_filepath

    def _parse(self):
        settings = {}
        for section in self.cfg.sections():
            section_settings = dict(self.cfg[section])
            settings[section] = section_settings

        self.log.debug("Configuration file '{}' loaded".format(self.cfg_filepath))

        return settings

    def __run_validator(self, fields, setting_name, setting_value) -> None:
        """Run validator for setting and raise ConfigError if validation fails.

        Args:
            fields (dict): required fields
            setting_name (str): setting name
            setting_value (any): setting value

        Raises:
            ValueError: if validation fails
        """
        if VALIDATOR_KEY in fields[setting_name]:
            validator = fields[setting_name][VALIDATOR_KEY]
            if not validator(setting_value):
                if ERR_MSG_KEY not in fields[setting_name]:
                    self.log.warning(
                        '{} has validation function but no error message'.format(setting_name),
                    )
                    msg = "Invalid value for '{}'".format(setting_name)
                else:
                    msg = fields[setting_name][ERR_MSG_KEY]
                raise ValueError(msg)

    def __set_by_type(self, field, param_value, expected_param_type):  # noqa: C901
        """Do type check and make sure the parameter is casted into that type.

        Args:
            param_value (): value of the parameter
            param_type (type): type of the parameter (str, int, bool, list, dict

        Raises:
            ValueError: if type set fails

        Returns:
            any: casted parameter
        """
        if isinstance(param_value, expected_param_type):
            # Parameter is already set to the expected type, so just return it
            return param_value

        err_message = "Invalid value '{}' for {}. Must be {} {}".format(
            param_value,
            field,
            'an' if expected_param_type == int else 'a',
            expected_param_type,
        )

        if expected_param_type in (str, tuple):
            # Most types can be converted to string, so don't do repr(param_value) here
            # If the value is not a string, raising a TypeError prevents more confusing errors later
            if not isinstance(param_value, expected_param_type):
                raise TypeError(err_message)

        elif expected_param_type == bool:
            try:
                casted_param_value = util.strtobool(param_value) == 1
            except (ValueError, AttributeError):
                raise TypeError(err_message)

        elif expected_param_type in (list, dict, int):
            try:
                casted_param_value = json.loads(param_value)
            except (TypeError, JSONDecodeError):
                raise TypeError(err_message)
            if not isinstance(casted_param_value, expected_param_type):
                raise TypeError(err_message)

        else:
            raise TypeError('Unknown type {}'.format(expected_param_type))

        return casted_param_value

    @property
    def rf_token(self):
        """rf_token getter."""
        return self.__rf_token

    @rf_token.setter
    def rf_token(self, token) -> None:
        """Set Recorded Future API key.

        Args:
            token (str): Recorded Future API key

        Raises:
            TypeError: If token is not a string
        """
        self.log.debug('Setting RF API Key')
        if not isinstance(token, str):
            raise TypeError('token must have type str')
        self.__rf_token = token

    @property
    def app_id(self):
        """app_id getter."""
        return self.__app_id

    @app_id.setter
    def app_id(self, app_id) -> None:
        """Set App ID which is used by RFClient in the UserAgent header.
        App ID must contain an integration version in the format of M.m.p.

        Args:
            app_id (string): app id, for example: ps-elk/1.0.0, ps-elk/9.9.9

        Raises:
            ConfigError: If no app_id was specified
            ConfigErrror: If app_id format is invalid

        Raises:
            TypeError: If app_id is not a string
            ValueError: If app_id is not the correct format
        """
        self.log.debug('Setting App ID')
        if not isinstance(app_id, str):
            raise TypeError('app_id must have type str')
        regexp = re.compile(r'^\S+\/\d+((\.\d+)*?)$')
        if regexp.search(app_id):
            self.__app_id = app_id
        else:
            raise ValueError('App ID must be in format M.m.p, for example: get-alerts/1.0.0')

    @property
    def platform_id(self):
        """platform_id getter."""
        return self.__platform_id

    @platform_id.setter
    def platform_id(self, platform_id) -> None:
        """Set platform ID which is used by RFClient in the UserAgent header.
        Platform ID must contain the platform name with the version in the format of M.m.p.

        Args:
            platform_id (str): platform id, for example: elasticsearch/8.0.3

        Raises:
            TypeError: If platform_id is not a string
            ValueError: If platform_id format is invalid
        """
        self.log.debug('Setting Platform ID')
        if not isinstance(platform_id, str):
            raise TypeError('platform_id must have type str')

        platform_id = platform_id.strip()
        regex_with_id = re.compile(r'^\S+\/\d+((\.\d+)*?)$')
        regex_no_id = re.compile(r'^([A-Z]|[a-z])+$')
        if regex_with_id.search(platform_id) or regex_no_id.search(platform_id):
            self.__platform_id = platform_id
        else:
            raise ValueError(
                'Platform ID must be in format platform/version, for example: Splunk/8.0.0',
            )
