##################################### TERMS OF USE ###########################################
# The following code is provided for demonstration purpose only, and should not be used      #
# without independent verification. Recorded Future makes no representations or warranties,  #
# express, implied, statutory, or otherwise, regarding any aspect of this code or of the     #
# information it may retrieve, and provides it both strictly “as-is” and without assuming    #
# responsibility for any information it may retrieve. Recorded Future shall not be liable    #
# for, and you assume all risk of using, the foregoing. By using this code, Customer         #
# represents that it is solely responsible for having all necessary licenses, permissions,   #
# rights, and/or consents to connect to third party APIs, and that it is solely responsible  #
# for having all necessary licenses, permissions, rights, and/or consents to any data        #
# accessed from any third party API.                                                         #
##############################################################################################

import json
import logging
import os

from requests import ConnectionError, ConnectTimeout, HTTPError, ReadTimeout

from ..config import Config
from ..constants import (
    _API_RESPONSE_LIMIT,
    DEFAULT_KEY,
    ERR_MSG_KEY,
    RF_TOKEN_KWARG,
    ROOT_DIR,
    STANZA_FIELDS_KEY,
    TYPE_KEY,
    VALIDATOR_KEY,
)
from ..endpoints import EP_DETECTION_RULES
from ..helpers.helpers import OSHelpers
from ..rf_client import RFClient
from .errors import DetectionRulesError
from .rf_detection_rule import RFDetectionRule

DETECTION_STANZA_NAME = 'detection'
DEFAULT_DETECTION_OUTPUT_DIR = os.path.join(ROOT_DIR, 'detection_rules')
OUTPUT_DIR_SETTING = 'output_directory'
LIMIT_SETTING = 'limit'
RULE_TYPE_SETTING = 'rule_type'
START_DATE_SETTING = 'start_date'
DEFAULT_RULE_TYPE_SETTING = None
DEFAULT_START_DATE = None
ENTITIES_ARG = 'entities'

DETECTION_RULE_TYPES = ['sigma', 'snort', 'yara']

DETECTION_RULE_MGR_PARAMS = {
    STANZA_FIELDS_KEY: {
        LIMIT_SETTING: {
            DEFAULT_KEY: _API_RESPONSE_LIMIT,
            TYPE_KEY: int,
            VALIDATOR_KEY: lambda x: x > 0,
            ERR_MSG_KEY: "[{}] '{}' is not a positive integer".format(
                DETECTION_STANZA_NAME,
                LIMIT_SETTING,
            ),
        },
        RULE_TYPE_SETTING: {
            DEFAULT_KEY: DEFAULT_RULE_TYPE_SETTING,
            TYPE_KEY: str,
            VALIDATOR_KEY: lambda x: x in DETECTION_RULE_TYPES,
            ERR_MSG_KEY: "[{}] '{}' must be one of {}".format(
                DETECTION_STANZA_NAME,
                RULE_TYPE_SETTING,
                DETECTION_RULE_TYPES,
            ),
        },
        START_DATE_SETTING: {DEFAULT_KEY: DEFAULT_START_DATE, TYPE_KEY: str},
        OUTPUT_DIR_SETTING: {
            DEFAULT_KEY: DEFAULT_DETECTION_OUTPUT_DIR,
            TYPE_KEY: str,
            VALIDATOR_KEY: OSHelpers.check_dir,
            ERR_MSG_KEY: "[{}] '{}' directory is invalid or has restricted permissions".format(
                DETECTION_STANZA_NAME,
                OUTPUT_DIR_SETTING,
            ),
        },
    },
}


class RFDetectionMgr:
    """Class for managing detection rules."""

    def __init__(self, cfg: Config = None, **kwargs):
        """Initialize RFDetectionMgr.

        Args:
            cfg (Config, optional): Config object. Inits empty Config if none provided
            limit (int, optional): Max number of rules to fetch. Defaults to 1000
            rule_type (str, optional): Rule type to fetch (yara, sigma, snort)
            start_date (str, optional): Start date for rules
            output_directory (str, optional): Rule output dir. Default sys.path[0]/detection_rules
        """
        self.log = logging.getLogger(__name__)
        self.config = cfg if cfg is not None else Config()
        rf_token = kwargs.pop(RF_TOKEN_KWARG) if RF_TOKEN_KWARG in kwargs else None
        self.rf_client = (
            RFClient(self.config, rf_token=rf_token)
            if rf_token is not None
            else RFClient(self.config)
        )
        self.config.parse_stanza(DETECTION_STANZA_NAME, DETECTION_RULE_MGR_PARAMS)
        self.config.update_settings(
            [
                (DETECTION_STANZA_NAME, DETECTION_RULE_MGR_PARAMS),
            ],
            kwargs,
        )
        self.config.log_settings(DETECTION_STANZA_NAME)

    def _validate_fetch_rules_args(self, **kwargs) -> None:
        """Validates the arguments to fetch_rules.

        Argument names match setting names wherever possible

        Args:
            entities (list str): List of string entity IDs to search
            rule_type (str): Rule type (yara, sigma, snort) to search
            limit (int): Limit for rules to search
            start_date (str): Start date to search

        Raises:
            TypeError: If an argument has an incorrect type
            ValueError: If an argument has an incorrect value
        """
        if ENTITIES_ARG in kwargs and (
            not isinstance(kwargs[ENTITIES_ARG], list)
            or not all(isinstance(x, str) for x in kwargs[ENTITIES_ARG])
        ):
            raise TypeError('{} must be a list of string entity IDs'.format(ENTITIES_ARG))
        if RULE_TYPE_SETTING in kwargs and not isinstance(kwargs[RULE_TYPE_SETTING], str):
            raise TypeError('{} must be a string'.format(RULE_TYPE_SETTING))
        if RULE_TYPE_SETTING in kwargs and kwargs[RULE_TYPE_SETTING] not in DETECTION_RULE_TYPES:
            raise ValueError(
                '{} must be one of {}'.format(RULE_TYPE_SETTING, ', '.join(DETECTION_RULE_TYPES)),
            )
        if LIMIT_SETTING in kwargs and not isinstance(kwargs[LIMIT_SETTING], int):
            raise TypeError('{} must be an integer'.format(LIMIT_SETTING))
        if LIMIT_SETTING in kwargs and kwargs[LIMIT_SETTING] < 0:
            raise ValueError('{} must be non-negative'.format(LIMIT_SETTING))
        if START_DATE_SETTING in kwargs and not isinstance(kwargs[START_DATE_SETTING], str):
            raise TypeError('{} must be a string'.format(START_DATE_SETTING))

    def fetch_rules(self, **kwargs):
        """Fetches all detection rules for a given query.

        Args:
            entities (list str): List of string entity IDs to search
            rule_type (str): Rule type (yara, sigma, snort) to search
            limit (int): Limit for rules to search
            start_date (str): Start date to search

        Returns:
            rulesList (list): list of rules returned
        """
        url = EP_DETECTION_RULES
        counts = 0
        rule_list = []
        finished = False
        offset = None
        self._validate_fetch_rules_args(**kwargs)
        self.log.info(
            'Searching for detection rules with params:\n'
            + '  {}: {}\n  {}: {}\n  {}: {}\n  {}: {}'.format(
                ENTITIES_ARG,
                kwargs.get(ENTITIES_ARG),
                RULE_TYPE_SETTING,
                kwargs.get(RULE_TYPE_SETTING) or self.rule_type,
                LIMIT_SETTING,
                kwargs.get(LIMIT_SETTING) or self.limit,
                START_DATE_SETTING,
                kwargs.get(START_DATE_SETTING) or self.start_date,
            ),
        )
        while not finished:
            try:
                resp = self.make_request(url, offset, **kwargs)
            except (HTTPError, ConnectTimeout, ConnectionError, ReadTimeout) as dre:
                raise DetectionRulesError('Error fetching rules: {}'.format(str(dre)))
            try:
                counts, rule_list, finished, offset = self._paginate_response(
                    counts,
                    resp,
                    finished,
                    rule_list,
                    offset,
                )
            except KeyError as ke:
                raise DetectionRulesError('Error processing response: {}'.format(str(ke)))

            self.log.debug('count dict: {}\n next_offset: {}'.format(str(counts), str(offset)))

        detection_rules = [RFDetectionRule(self, rule) for rule in rule_list]

        return detection_rules

    def _paginate_response(self, counts, resp, finished, rule_list, offset):
        """Page responses and return rules returned, finished flag, and next offset.

        Args:
            counts (int): count of rules returned
            resp (dict): response from Detection Rules API
            finished (bool): flag to indicate if search is complete
            rule_list (list): list of rules returned
            offset (str): offset string for pagination

        Returns:
            counts (int): count of rules returned
            rule_list (list): list of rules returned
            finished (bool): flag to indicate if search is complete
            offset (str): offset string for pagination
        """
        counts = counts + resp['count']
        rule_list = rule_list + resp['result']
        offset = None
        if resp['count'] == 0 or resp['count'] < self.limit:
            finished = True
            self.log.info('All available rules found')
        elif counts >= self.limit:
            finished = True
            self.log.info('More than (' + str(self.limit) + ') rules found.')
        elif 'next_offset' in resp:
            offset = resp['next_offset']
        else:
            self.log.error('*** ERROR!  Unexpected Error Found in count/offset logic')
            finished = True
        return (counts, rule_list, finished, offset)

    def make_request(self, url, offset, **kwargs):
        """Makes Recorded Future API request and returns the response.

        Args:
            url (str): url from endpoints.py for detection rules api
            offset (str): offset string for pagination
            entities (list str): List of string entity IDs to search
            rule_type (str): Rule type (yara, sigma, snort) to search
            limit (int): Limit for rules to search
            start_date (str): Start date to search

        Raises:
            DetectionRulesError: Error returned for Detection Rules

        Returns:
            resp: response from Detection Rules API
        """
        data = self.prepare_request(offset, **kwargs)
        try:
            response = self.rf_client.post_request(url, data)
        except (HTTPError, ConnectTimeout, ConnectionError, ReadTimeout) as dre:
            raise DetectionRulesError('Error fetching rules: {}'.format(str(dre)))

        try:
            resp = response.json()
        except json.JSONDecodeError as jde:
            raise DetectionRulesError('Error decoding json: {}'.format(str(jde)))

        if 'count' not in resp or 'result' not in resp:
            raise DetectionRulesError('Invalid response from API: {}'.format(resp))
        return resp

    def prepare_request(self, offset, **kwargs):
        """Prepares the request for the Detection Rule API.

        Args:
            offset (str, optional): Offset for pagination. Defaults to None.
            entities (list str): List of string entity IDs to search
            rule_type (str): Rule type (yara, sigma, snort) to search
            limit (int): Limit for rules to search
            start_date (str): Start date to search

        Returns:
            requestString (str): JSON string of request
        """
        limit = kwargs.get(LIMIT_SETTING) or self.limit
        request_body = {
            'tagged_entities': False,
            'limit': limit,
        }
        if offset is not None:
            request_body['offset'] = offset
        filter_params = {}
        if kwargs.get(ENTITIES_ARG) is not None:
            filter_params['entities'] = kwargs.get(ENTITIES_ARG)
        start_date = kwargs.get(START_DATE_SETTING) or self.start_date
        if start_date is not None:
            filter_params['created'] = {'after': self.start_date}
        rule_type = kwargs.get(RULE_TYPE_SETTING) or self.rule_type
        if rule_type is not None:
            filter_params['types'] = [rule_type]
        if len(filter_params) > 0:
            request_body['filter'] = filter_params

        return json.dumps(request_body)

    @property
    def output_directory(self) -> str:
        """Getter for output_dir.

        Returns:
            str: output_dir
        """
        output_directory = self.config.settings.get(DETECTION_STANZA_NAME, {}).get(
            OUTPUT_DIR_SETTING,
        )
        if output_directory is None:
            output_directory = DEFAULT_DETECTION_OUTPUT_DIR
            self.log.warning(
                'No output directory provided. Using default: {}'.format(
                    DEFAULT_DETECTION_OUTPUT_DIR,
                ),
            )

        return output_directory

    @output_directory.setter
    def output_directory(self, value: str) -> None:
        """Setter for output_dir."""
        if not DETECTION_RULE_MGR_PARAMS[STANZA_FIELDS_KEY][OUTPUT_DIR_SETTING][VALIDATOR_KEY](
            value,
        ):
            raise ValueError('Invalid value for {}: "{}"'.format(OUTPUT_DIR_SETTING, value))
        self.config.update_setting(DETECTION_STANZA_NAME, OUTPUT_DIR_SETTING, value)

    @property
    def limit(self) -> int:
        """Getter for limit.

        Returns:
            int: limit
        """
        limit = self.config.settings.get(DETECTION_STANZA_NAME, {}).get(LIMIT_SETTING)
        if limit is None:
            limit = _API_RESPONSE_LIMIT
            self.log.warning('No {} provided. Using default: {}'.format(LIMIT_SETTING, str(limit)))

        return limit

    @limit.setter
    def limit(self, value: int) -> None:
        """Setter for limit."""
        if not isinstance(value, int):
            raise TypeError('{} must be an integer'.format(LIMIT_SETTING))
        if not DETECTION_RULE_MGR_PARAMS[STANZA_FIELDS_KEY][LIMIT_SETTING][VALIDATOR_KEY](value):
            raise ValueError('Invalid value for {}: "{}"'.format(LIMIT_SETTING, value))
        self.config.update_setting(DETECTION_STANZA_NAME, LIMIT_SETTING, value)

    @property
    def rule_type(self) -> str:
        """Getter for rule_type.

        Returns:
            str: rule_type
        """
        rule_type = self.config.settings.get(DETECTION_STANZA_NAME, {}).get(RULE_TYPE_SETTING)
        if rule_type is None:
            self.log.warning('No {} set'.format(RULE_TYPE_SETTING))

        return rule_type

    @rule_type.setter
    def rule_type(self, value: str) -> None:
        """Setter for rule_type."""
        if not isinstance(value, str):
            raise TypeError('{} must be a string'.format(RULE_TYPE_SETTING))
        if not DETECTION_RULE_MGR_PARAMS[STANZA_FIELDS_KEY][RULE_TYPE_SETTING][VALIDATOR_KEY](
            value,
        ):
            raise ValueError('Invalid value for {}: "{}"'.format(RULE_TYPE_SETTING, value))
        self.config.update_setting(DETECTION_STANZA_NAME, RULE_TYPE_SETTING, value)

    @property
    def start_date(self) -> str:
        """Getter for start_date.

        Returns:
            str: start_date
        """
        start_date = self.config.settings.get(DETECTION_STANZA_NAME, {}).get(START_DATE_SETTING)
        if start_date is None:
            self.log.warning('No {} set'.format(START_DATE_SETTING))

        return start_date

    @start_date.setter
    def start_date(self, value: str) -> None:
        """Setter for start_date."""
        if not isinstance(value, str):
            raise TypeError('{} must be a string'.format(START_DATE_SETTING))
        self.config.update_setting(DETECTION_STANZA_NAME, START_DATE_SETTING, value)
