##################################### TERMS OF USE ###########################################
# The following code is provided for demonstration purpose only, and should not be used      #
# without independent verification. Recorded Future makes no representations or warranties,  #
# express, implied, statutory, or otherwise, regarding any aspect of this code or of the     #
# information it may retrieve, and provides it both strictly “as-is” and without assuming    #
# responsibility for any information it may retrieve. Recorded Future shall not be liable    #
# for, and you assume all risk of using, the foregoing. By using this code, Customer         #
# represents that it is solely responsible for having all necessary licenses, permissions,   #
# rights, and/or consents to connect to third party APIs, and that it is solely responsible  #
# for having all necessary licenses, permissions, rights, and/or consents to any data        #
# accessed from any third party API.                                                         #
##############################################################################################

import logging
import os

from ..exceptions import WriteFileError
from ..helpers.helpers import OSHelpers


class RFDetectionRule:
    """Recorded Future Detection Rules class."""

    def __init__(self, rule_mgr, rule):
        """Initializes the RFDetectionRule object."""
        self.log = logging.getLogger(__name__)
        self.__rule_mgr = rule_mgr
        self.rule_entities = []

        self.__raw = rule
        (
            self.__title,
            self.__description,
            self.__created_date,
            self.__updated_date,
            self.__doc_id,
            self.__rules,
        ) = self._parse()

    def __str__(self):
        """Returns a string representation of the object."""
        rules = []
        for rule in self.rules:
            rule_info = '    file_name={}\n'.format(rule['file_name'])
            rule_info += '    content={}'.format(rule['content'])
            rule_info += '    entities={}\n'.format(rule['entities'])
            rules.append(rule_info)

        str_rep = 'RFDetectionRule(\n'
        str_rep += 'title:\n  {}\n'.format(self.title)
        str_rep += 'description:\n  {}\n'.format(self.description)
        str_rep += 'created_date:\n  {}\n'.format(self.created_date)
        str_rep += 'updated_date:\n  {}\n'.format(self.updated_date)
        str_rep += 'doc_id:\n  {}\n'.format(self.doc_id)
        str_rep += 'rules:\n{})'.format('\n'.join(rules) if rules else 'None')

        return str_rep

    def _parse(self):
        title = None
        description = None
        created_date = None
        updated_date = None
        doc_id = None
        rules = []

        if self.__raw is not None:
            if 'rules' in self.__raw:
                rules = [
                    {
                        'file_name': rule.get('file_name'),
                        'content': rule.get('content'),
                        'entities': [e['name'] for e in rule.get('entities', [])],
                    }
                    for rule in self.__raw['rules']
                ]
            title = self.__raw.get('title')
            description = self.__raw.get('description')
            created_date = self.__raw.get('created')
            updated_date = self.__raw.get('updated')
            doc_id = self.__raw.get('id')

        return (
            title,
            description,
            created_date,
            updated_date,
            doc_id,
            rules,
        )

    def write_file(self, output_directory=None, filename=None) -> None:
        """Writes a rule to file. A rule may have multiple contents, and become multiple files.

        Uses output directory set in the rule mgr and filename from the rule,
        unless otherwise specified.

        Args:
            rule (str): rule contents to write to file
            filename (str): string filename, should end in .csv

        Raises:
            WriteFileError: if there was an error writing the file

        Side effects:
            Writes rule to file(s)
        """
        output_directory = output_directory or self.__rule_mgr.output_directory
        if len(self.rules) == 0:
            self.log.warning('No content found for rule {}'.format(self.title))
            return
        self.log.info('Writing rule to file')

        try:
            OSHelpers.mkdir(output_directory)
        except (ValueError, WriteFileError) as err:
            raise WriteFileError('Could not create output directory: {}'.format(err))

        for idx, rule in enumerate(self.rules):
            if filename is not None:
                if idx > 1:
                    filename = '{}_{}'.format(idx, filename)
            else:
                filename = rule['file_name']
            try:
                full_path = os.path.join(output_directory, filename)
                with open(full_path, 'w') as outfile:
                    outfile.write(rule['content'])
                    self.log.info('Wrote: {}'.format(full_path))
            except OSError as err:
                raise WriteFileError("Could not write file '{}': {}".format(filename, err))

    @property
    def title(self):
        """Get rule title.

        Returns:
            str: rule title
        """
        return self.__title

    @property
    def description(self):
        """Get rule description.

        Returns:
            str: rule description
        """
        return self.__description

    @property
    def created_date(self):
        """Get rule created date.

        Returns:
            str: rule created date
        """
        return self.__created_date

    @property
    def updated_date(self):
        """Get rule updated date.

        Returns:
            str: rule updated date
        """
        return self.__updated_date

    @property
    def doc_id(self):
        """Get rule doc id.

        Returns:
            str: rule doc id
        """
        return self.__doc_id

    @property
    def rules(self):
        """Get rule rules.

        Returns:
            list: rule dicts
        """
        return self.__rules
