##################################### TERMS OF USE ###########################################
# The following code is provided for demonstration purpose only, and should not be used      #
# without independent verification. Recorded Future makes no representations or warranties,  #
# express, implied, statutory, or otherwise, regarding any aspect of this code or of the     #
# information it may retrieve, and provides it both strictly “as-is” and without assuming    #
# responsibility for any information it may retrieve. Recorded Future shall not be liable    #
# for, and you assume all risk of using, the foregoing. By using this code, Customer         #
# represents that it is solely responsible for having all necessary licenses, permissions,   #
# rights, and/or consents to connect to third party APIs, and that it is solely responsible  #
# for having all necessary licenses, permissions, rights, and/or consents to any data        #
# accessed from any third party API.                                                         #
##############################################################################################

from typing import Dict, List, Union

from psengine.enrich.constants import _MSG_KEYS


class RFEnrich:
    """Baseclass representing the object returned by an enrichment call: lookup or soar."""

    def __init__(self, ioc: str, ioc_type: str = '', result: Dict[str, Union[List, Dict]] = None):
        """Baseclass representing the object returned by an enrichment call: lookup or soar.

        Args:
            ioc (str): ioc enriched
            result (Dict[str, Union[List, Dict]], optional): It can be:
             - values extracted from enrichment payload
             - error message based on the exception raised (404, 500 etc)
            Defaults to {}.

        Raises:
            ValueError: if ioc is not a string
            ValueError: if enriched values is not a dictionary
        """
        if isinstance(ioc, str) and len(ioc):
            self.__ioc = ioc
        else:
            raise ValueError('IOC must be a non empty string')
        result = result if result is not None else {}
        if (
            isinstance(result, dict)
            and all(isinstance(val, (dict, list)) for val in result.values())
            and all(isinstance(key, str) for key in result)
        ):
            self.__result = result
        else:
            raise ValueError('Enriched result should be a dict of dict')

        self.__ioc_type = ioc_type

    def __repr__(self):
        """__repr__ method."""
        crop = '[...]' if self.is_enriched else ''

        return 'RFEnrich(ioc={}, is enriched={}, result={:.20}{})'.format(
            self.ioc,
            self.is_enriched,
            str(self.result),
            crop,
        )

    def __str__(self):
        """__str__ method."""
        return "'ioc': {}, 'ioc type': {}, 'is enriched': {}, 'result': {}".format(
            self.ioc,
            self.ioc_type,
            self.is_enriched,
            self.result,
        )

    @property
    def ioc(self):
        """Return ioc."""
        return self.__ioc

    @property
    def is_enriched(self) -> bool:
        """The following property is mostly as an helper attribute to quickly identify those IOCs
        that for any reason has not been enriched. It will be false if the API call returned 404,
        4xx and 5xx errors.

        Returns:
            bool: True if enriched, false otherwise.
        """
        return not (
            any(x in self.result for x in _MSG_KEYS)
            or not self.result
            or self.result.get('risk', {}).get('score') == 0
        )

    @property
    def result(self):
        """Return result of enrichment."""
        return self.__result

    @property
    def ioc_type(self):
        """Returns the type of the IOC enriched."""
        return self.__ioc_type
