##################################### TERMS OF USE ###########################################
# The following code is provided for demonstration purpose only, and should not be used      #
# without independent verification. Recorded Future makes no representations or warranties,  #
# express, implied, statutory, or otherwise, regarding any aspect of this code or of the     #
# information it may retrieve, and provides it both strictly “as-is” and without assuming    #
# responsibility for any information it may retrieve. Recorded Future shall not be liable    #
# for, and you assume all risk of using, the foregoing. By using this code, Customer         #
# represents that it is solely responsible for having all necessary licenses, permissions,   #
# rights, and/or consents to connect to third party APIs, and that it is solely responsible  #
# for having all necessary licenses, permissions, rights, and/or consents to any data        #
# accessed from any third party API.                                                         #
##############################################################################################

import json
import logging
import re
import urllib
from collections import UserDict
from copy import deepcopy
from itertools import groupby
from threading import Lock
from typing import Dict, List, Tuple, Union

import simplejson
from more_itertools import chunked
from requests import ConnectTimeout, HTTPError, ReadTimeout

from ..config.config import Config
from ..constants import DEFAULT_KEY, RF_TOKEN_KWARG, STANZA_FIELDS_KEY, TYPE_KEY
from ..endpoints import BASE_URL, EP_SOAR_ENRICHMENT
from ..helpers import JSONParsingHelper, MultiThreadingHelper
from ..rf_client import RFClient
from .constants import (
    _DOMAIN,
    _ERR_KEY,
    _ERR_MSG_HTTP,
    _ERR_MSG_IOC_TYPE,
    _ERR_MSG_JSON,
    _INFO_IOC_NOT_FOUND,
    _IOC_TYPES,
    _MSG,
    _MSG_HTTP,
    _MSG_JSON,
    _MSG_KEYS,
    _RE_CVE,
    _RE_DOMAIN,
    _RE_HASH,
    _RE_RF_ID,
    _RE_URL_DOM,
    _RE_URL_IP,
    _RF_ID,
    _SOAR_IOC_NAME_JSONPATH,
    _SOAR_POST_ROWS,
    _RE_IPv4,
    _RE_IPv6,
)
from .rf_enrich import RFEnrich

ENRICHMENT_STANZA_NAME = 'enrichment'
FIELDS_SETTING = 'fields'
FIELDS_DEFAULT = {
    'ip': ['risk.score'],
    'domain': ['risk.score'],
    'url': ['risk.score'],
    'hash': ['risk.score'],
    'vulnerability': ['risk.score'],
    'companybydomain': ['risk.score'],
    'company': ['risk.score'],
}
MAX_WORKERS_SETTING = 'max_workers'
MAX_WORKERS_DEFAULT = 0

ENRICHMENT_PARAMS = {
    STANZA_FIELDS_KEY: {
        FIELDS_SETTING: {
            DEFAULT_KEY: FIELDS_DEFAULT,
            TYPE_KEY: dict,
        },
        MAX_WORKERS_SETTING: {
            DEFAULT_KEY: MAX_WORKERS_DEFAULT,
            TYPE_KEY: int,
        },
    },
}

ENRICHMENT_OUTPUT_FNAME = 'rfenriched_'


class EnrichedContent(UserDict):
    """Custom dictionary that allows only values of type RFEnrich and keys to be strings."""

    def __init__(self, content: Dict[str, Union[List, Dict]]):
        """__init__ for EnrichedContent.

        Args:
            content (Dict[str, Union[List, Dict]]): content to be added

        Raises:
            TypeError: If content is not a dict
            ValueError: If key, value do not follow the _is_instance constraints.
        """
        super().__init__()
        if not isinstance(content, dict):
            raise TypeError('Content must be a dict.')

        for key, value in content.items():
            if self._is_instance(key, value):
                self.data[key] = value
            else:
                msg = 'Key must be the IOC enriched in value. Got key: {} and value: {}.'.format(
                    key,
                    value,
                )
                raise ValueError(msg)

    @staticmethod
    def _is_instance(key: str, value: RFEnrich) -> bool:
        """Check the instance type of key and value pair.
        Key must be the same as value.ioc so that an assignment as below does not occur:
            '1.1.1.1' : RFenrich(ioc='google.com', [...]).

        Args:
            key (str): Must be a string
            value (RFEnrich): Must be RFEnrich type

        Returns:
            bool: True if all conditions met, False otherwise.
        """
        return isinstance(key, str) and isinstance(value, RFEnrich) and key == value.ioc

    def __setitem__(self, key: str, value: Union[List, Dict]):
        """__setitem__ method which checks the type of the key and value using self._is_instance.

        Raises:
            ValueError: if type of key or value are wrong.
        """
        if self._is_instance(key, value):
            self.data[key] = value
        else:
            msg = 'Key must be the IOC enriched in value. Got key: {} and value: {}.'.format(
                key,
                value,
            )
            raise ValueError(msg)

    def __getitem__(self, key: str):
        """__getitem__ method."""
        if isinstance(key, str):
            return self.data.get(key, None)
        return None

    def __repr__(self):
        """__repr__ method."""
        items = ', '.join(f'{key!r}: {value!r}' for key, value in self.items())
        return f"EnrichedContent({{{items[:65]}{'[...]' if len(items) > 65 else ''}}})"

    def __str__(self):
        """__str__ method."""
        items = ', '.join(f'{key}: {{{value}}}' for key, value in self.items())
        return f'{{{items}}}'


class RFEnrichMgr:
    """Class for managing the RFEnrich objects."""

    def __init__(self, cfg: Config = None, **kwargs):
        """Initialize RFEnrichment object.

        Args:
            cfg (Config, optional): Config object. Inits empty Config if none provided
            fields (dict, optional): Fields to be looked up for each IOC type.
                                    Defaults to ["risk.score"] for each IOC type
            max_workers (int, optional): Max number of workers for ThreadPoolExecutor.
                                            Defaults to 0
            rf_token (str, optional): Recorded Future API token. Defaults to None
        """
        self.log = logging.getLogger(__name__)
        self.config = cfg if cfg is not None else Config()
        rf_token = kwargs.pop(RF_TOKEN_KWARG) if RF_TOKEN_KWARG in kwargs else None
        self.config.parse_stanza(ENRICHMENT_STANZA_NAME, ENRICHMENT_PARAMS)
        self.config.update_settings(
            [
                (ENRICHMENT_STANZA_NAME, ENRICHMENT_PARAMS),
            ],
            kwargs,
        )
        self.rf_client = (
            RFClient(self.config, rf_token=rf_token)
            if rf_token is not None
            else RFClient(self.config)
        )
        self.rf_client.set_urllib_log_level('ERROR')
        self.config.log_settings(ENRICHMENT_STANZA_NAME)
        self.__content = EnrichedContent({})
        self.lock = Lock()

    def __str__(self):
        """__str__ method."""
        return str(self.content)

    def __repr__(self):
        """__repr__ method."""
        return 'RFEnrichMgr(max_workers: {}, fields: {}, content: {})'.format(
            self.max_workers,
            self.fields,
            repr(self.content),
        )

    def __len__(self):
        """__len__ method."""
        return len(self.content)

    def __iter__(self):
        """__iter__ method."""
        return iter(self.content)

    def __getitem__(self, key: str):
        """__getitem__ method. Used for getting element with manager['8.8.8.8']."""
        if isinstance(key, str):
            return self.content.get(key, None)
        return None

    def items(self) -> Tuple[str, list]:
        """Emulate the items method of dictionary to return key, value.

        Yields:
            tuple: key, value pair
        """
        for key in self:
            yield key, self[key]

    def json(self) -> Dict:
        """Return dict as json representation of the results."""
        return {key: value.result for key, value in self.items()}

    def lookup(self, ioc: Union[str, List[str]]) -> None:
        """Lookup of an IOC or list of IOCs.

        Args:
            ioc (Union[str, List[str]]): iocs to enrich

        Raises:
            ValueError: if ioc is not a string or list.
        """
        if not isinstance(ioc, (list, str)) or not len(ioc):
            raise ValueError('ioc must be a string or a list of strings with no empty values.')

        ioc = [ioc] if isinstance(ioc, str) else ioc

        if self.max_workers >= 1:
            futures = MultiThreadingHelper.multithread_it(
                self.max_workers,
                self._lookup_operations,
                iterator=ioc,
            )
            for future in futures:
                ioc, ioc_type, payload = future.result()
                self.content[ioc] = RFEnrich(ioc, ioc_type, payload)
        else:
            for single_ioc in ioc:
                enriched_ioc, ioc_type, payload = self._lookup_operations(single_ioc)
                self.content[single_ioc] = RFEnrich(enriched_ioc, ioc_type, payload)

    def company_lookup(self, iocs: Union[str, List[str]]) -> None:
        """Lookup of an IOC or list of IOCs against the company endpoint.
        The IOC can be a domain, which will point to the company/by_domain or a Recorded Future ID
        which will point to the company/{id} endpoint.

        Args:
            ioc (Union[str, List[str]]): iocs to enrich

        Raises:
            ValueError: if ioc is not a string or list.
        """
        if not isinstance(iocs, (list, str)) or not len(iocs):
            raise ValueError('ioc must be a string or a list of strings with no empty values.')

        iocs = [iocs] if isinstance(iocs, str) else iocs

        if self.max_workers >= 1:
            futures = MultiThreadingHelper.multithread_it(
                self.max_workers,
                self._company_lookup_operations,
                iterator=iocs,
            )
            for future in futures:
                iocs, ioc_type, payload = future.result()
                self.content[iocs] = RFEnrich(iocs, ioc_type, payload)
        else:
            for single_ioc in iocs:
                enriched_ioc, ioc_type, payload = self._company_lookup_operations(single_ioc)
                self.content[single_ioc] = RFEnrich(enriched_ioc, ioc_type, payload)

    def soar(self, iocs: Union[str, List[str]]) -> None:
        """Perform SOAR request, assinging the result to the IOC object.

        Args:
            iocs (Union[str, List[str]]): list of IOCs to enrich or single string to enrich.

        Raises:
            TypeError: If iocs is not a string or list
            ValueError: If all iocs do not belong to any IOC group supported.
        """
        if isinstance(iocs, str):
            iocs = [iocs]
        elif not isinstance(iocs, list):
            raise TypeError('The IOC to enrich must be a list of IOCs or a single IOC as string.')

        grouped_data = self._soar_categorize_iocs(iocs)
        if not grouped_data:
            raise ValueError('No IOCs types have been identified to be sent to the SOAR API.')

        self._soar(grouped_data)

    def company_soar(self, company_domains: Union[str, List[str]]) -> None:
        """Perform SOAR request for company via domain, assinging the result to the IOC object.

        Args:
            iocs (Union[str, List[str]]): list of IOCs to enrich or single string to enrich.

        Raises:
            TypeError: If iocs is not a string or list
        """
        if not isinstance(company_domains, (str, list)):
            raise TypeError('company_domains has to be a string or list')

        doms = [company_domains] if isinstance(company_domains, str) else company_domains
        grouped_data = {'companybydomain': doms}

        self._soar(grouped_data, specific_ioc_type='company')

    def _company_lookup_operations(self, ioc: str) -> Tuple[str, str, Dict]:
        """Perform all the operations when lookup is called. Write the result in self.content.

        Args:
            ioc (str): ioc to enrich

        Returns:
            Tuple[str, str, Dict]: tuple containing the IOC, the type of the IOC (ip, domain, etc)
                and the payload.
        """
        allowed_iocs = (_RF_ID, _DOMAIN)
        ioc_type = self._identify_ioc_type(ioc)

        if ioc_type not in allowed_iocs:
            raise ValueError(
                'Only domains and company IDs can be enriched via comapny_lookup, got: {}'.format(
                    ioc_type,
                ),
            )

        ioc_type = 'company' if ioc_type == _RF_ID else 'companybydomain'
        url = self._company_lookup_build_url(ioc, ioc_type)
        payload = self._make_request(url)

        # Do not parse payload if there was an error during the API connection
        if all(msg_key not in payload for msg_key in _MSG_KEYS):
            payload = JSONParsingHelper.parse_payload(payload, self.fields[ioc_type])

        self.log.debug('{} with fields {} returned: {}'.format(ioc, self.fields[ioc_type], payload))

        return ioc, ioc_type, payload

    def _identify_ioc_type(self, ioc: str) -> str:
        """Get the type of IOC based on the regex.

        Args:
            ioc (str): IOC to identify

        Returns:
            str: type of IOC
        """
        if not isinstance(ioc, str):
            return None

        regex_choice = {
            'domain': _RE_DOMAIN,
            'url_dom': _RE_URL_DOM,
            'url_ip': _RE_URL_IP,
            'ipv4': _RE_IPv4,
            'ipv6': _RE_IPv6,
            'hash': _RE_HASH,
            'vulnerability': _RE_CVE,
            _RF_ID: _RE_RF_ID,
        }
        for endpoint, regex in regex_choice.items():
            match = re.fullmatch(regex, ioc, re.IGNORECASE)
            if endpoint.startswith('ip'):
                endpoint = 'ip'
            elif endpoint.startswith('url'):
                endpoint = 'url'

            if match:
                return endpoint
        return None

    def _make_request(self, lookup_url: str = None, soar_payload: dict = None) -> dict:
        """Make HTTP Request to: URL if lookup enrichment, to SOAR if payload is specified.

        Args:
            url (str): url to contact
            payload (dict) : payload for POST request

        Raises:
            http_err: HTTP Error from connection
            payload_err: Error from payload returned
            ValueError: if payload cannot be converted to JSON.

        Returns:
            dict: payload if response was ok, else dict of error message.
        """
        try:
            response = (
                self.rf_client.get_request(url=lookup_url)
                if lookup_url
                else self.rf_client.post_request(
                    url=EP_SOAR_ENRICHMENT,
                    data=json.dumps(soar_payload),
                )
            )

        except (HTTPError, ConnectionError, ConnectTimeout, ReadTimeout) as http_err:
            if http_err.response.status_code == 404:
                self.log.info('No information has been found for querying {}'.format(lookup_url))
                return _INFO_IOC_NOT_FOUND

            exp_name = type(http_err).__name__
            self.log.warning('Error connecting to: {}, error: {}'.format(lookup_url, exp_name))
            return_error = deepcopy(_ERR_MSG_HTTP)
            return_error[_ERR_KEY][_MSG] = exp_name + _MSG_HTTP
            return return_error
        except (json.JSONDecodeError, TypeError):
            raise TypeError('The given payload cannot be converted to JSON.')

        try:
            data = response.json()['data'] if lookup_url else response.json()['data']['results']
        except (KeyError, TypeError, simplejson.errors.JSONDecodeError) as payload_err:
            exp_name = type(payload_err).__name__
            msg = '{} in the data returned. Unable to parse it to JSON from {}'.format(
                exp_name,
                lookup_url,
            )
            self.log.warning(msg)
            return_error = deepcopy(_ERR_MSG_JSON)
            return_error[_ERR_KEY][_MSG] = exp_name + _MSG_JSON
            return return_error

        return data

    def _lookup_operations(self, ioc: str) -> Tuple[str, str, Dict]:
        """Perform all the operations when lookup is called. Write the result in self.content.

        Args:
            ioc (str): ioc to enrich

        Returns:
            Tuple[str, str, Dict]: tuple containing the IOC, the type of the IOC (ip, domain, etc)
                and the payload.
        """
        payload, ioc_type = self._lookup_perform_request(ioc)

        # Do not parse payload if there was an error during the API connection
        if all(msg_key not in payload for msg_key in _MSG_KEYS):
            payload = JSONParsingHelper.parse_payload(payload, self.fields[ioc_type])

        self.log.debug('{} with fields {} returned: {}'.format(ioc, self.fields[ioc_type], payload))
        return ioc, ioc_type, payload

    def _lookup_perform_request(self, ioc: str) -> Tuple[dict, Union[str, None]]:
        """Perform enrichment lookup if the ioc type can be understood,
        else return ({}, None).

        Args:
            ioc (str): ioc to enrich

        Returns:
            Tuple[dict, Union[str, None]]: Tuple of payload and ioc type
        """
        payload = _ERR_MSG_IOC_TYPE
        ioc_type = self._identify_ioc_type(ioc)

        if ioc_type:
            url = self._lookup_build_url(ioc, ioc_type)
            payload = self._make_request(url)

        return payload, ioc_type

    def _lookup_build_url(self, ioc: str, ioc_type: str) -> Union[str, None]:
        """Create the Lookup URL.

        Args:
            ioc (str): ioc to enrich
            ioc_type (str): type of ioc

        Returns:
            Union[str, None]: url string if ioc_type is present else None
        """
        if ioc_type:
            with self.lock:
                if not self.fields.get(ioc_type):
                    self.log.warning(
                        '{} is a {}. The related fields are missing. Using defaults.'.format(
                            ioc,
                            ioc_type,
                        ),
                    )
                    self.fields[ioc_type] = FIELDS_DEFAULT[ioc_type]

            ioc = urllib.parse.quote(ioc, safe='.') if ioc_type == 'url' else ioc
            url_field = '%2C'.join({field.split('.')[0] for field in self.fields[ioc_type]})

            return '{}/v2/{}/{}?fields={}'.format(BASE_URL, ioc_type, ioc, url_field)

        return None

    def _company_lookup_build_url(self, ioc: str, ioc_type: str) -> str:
        """Create the Lookup URL for Company and Company by Domain entities.

        Args:
            ioc (str): ioc to enrich
            ioc_type (str): type of ioc

        Returns:
            str: url string
        """
        with self.lock:
            if not self.fields.get(ioc_type):
                self.log.warning(
                    '{} is a {}. The related fields are missing. Using defaults.'.format(
                        ioc,
                        ioc_type,
                    ),
                )
                self.fields[ioc_type] = FIELDS_DEFAULT[ioc_type]

        url_field = '%2C'.join({field.split('.')[0] for field in self.fields[ioc_type]})
        ioc_type = 'company/by_domain' if ioc_type == 'companybydomain' else ioc_type
        return '{}/v2/{}/{}?fields={}'.format(BASE_URL, ioc_type, ioc, url_field)

    def _soar(self, grouped_data: Dict[str, List[str]], specific_ioc_type: str = None) -> None:
        """Breakdown of Multithreaded or Single threaded calls.

        Args:
            grouped_data (dict): data to be submitted, key is the ioc type, value is a list of IOCs
            specific_ioc_type (str, optional): if type of data is not the same as the IOc type
                for example company are looked up as companybydomain. Defaults to None.
        """
        for ioc_type in grouped_data:
            if ioc_type not in self.fields:
                self.log.warning(
                    'There is an attempt to enrich some {}, '
                    'however the field for this type has not been defined. Using defauts.'.format(
                        ioc_type,
                    ),
                )
                try:
                    self.fields[ioc_type] = FIELDS_DEFAULT[ioc_type]
                except KeyError:
                    msg = f'The IOCs {grouped_data[ioc_type]} have the wrong type {ioc_type}.'
                    raise ValueError(msg)

        if self.max_workers:
            iterator = [
                {ioc_type: data}
                for ioc_type, ioc_list in grouped_data.items()
                for data in chunked(ioc_list, _SOAR_POST_ROWS)
            ]
            MultiThreadingHelper().multithread_it(
                self.max_workers,
                self._soar_operations,
                iterator=iterator,
                specific_ioc_type=specific_ioc_type,
            )

        else:
            for ioc_type, ioc_list in grouped_data.items():
                for ioc_chunck in chunked(ioc_list, _SOAR_POST_ROWS):
                    self._soar_operations(
                        {ioc_type: ioc_chunck},
                        specific_ioc_type=specific_ioc_type,
                    )

    def _soar_operations(self, data: dict.items, specific_ioc_type: str = None) -> None:
        """SOAR operations broken into use case. We need to change the fields temporarely
        to get the entity.name field in order to know which IOC the payload belongs to and assign
        it to the right RFEnrich object. In the final `result` the entity.name field will not be
        shown unless specifically requested by the user.

        It is considered an error if the payload does not have entity.name in it.

        Args:
            data (dict.items): IOCs (values) grouped by IOC type (key)
            *_ : this is needed for MultiThreading, not actually used within the function.
        """
        ioc_type = next(iter(data.keys()))
        payloads = self._make_request(soar_payload=data)

        fields = (
            [_SOAR_IOC_NAME_JSONPATH] + self.fields[ioc_type]
            if _SOAR_IOC_NAME_JSONPATH not in self.fields[ioc_type]
            else self.fields[ioc_type]
        )
        for payload in payloads:
            result = JSONParsingHelper.parse_payload(payload, fields)

            ioc = result.get('entity', {}).get('name')
            if ioc:
                if _SOAR_IOC_NAME_JSONPATH not in self.fields[ioc_type]:
                    self._soar_cleanup_result(result)

                type_of_ioc = specific_ioc_type if specific_ioc_type else ioc_type
                self.content[ioc] = RFEnrich(ioc, type_of_ioc, result)
                self.log.debug('{} returned: {}'.format(ioc, result))
            else:
                self.log.warning(
                    'The following result does not have an entity.name value path, '
                    'which is likely an error in the API response.: {}.'.format(result),
                )

    def _soar_cleanup_result(self, result: dict) -> None:
        """Inplace cleanup of the result dictionary.
        This is needed since we need to add the 'entity.name' to have a reference of which IOC
        the part of the payload is related to. If the user has specifically requested for this
        field, it will be left untouched.

        Args:
            result (dict): payload
        """
        try:
            del result['entity']['name']
        except KeyError:
            pass
        else:
            if not result.get('entity'):
                del result['entity']

    def _soar_categorize_iocs(self, iocs: List[str]) -> Dict[str, List[str]]:
        """Find IOC type of IOCs and organize them in a dictionary with
        key the ioc type and value the iocs belonging to that ioc type.
        This is used by SOAR API endpoint as POST payload.

        Args:
            iocs (List[str]): list of iocs

        Returns:
            Dict[str, List[str]]: dict of {ioc_type: [iocs], ioc_type: [iocs]}
        """
        data = []

        for ioc in iocs:
            ioc_type = self._identify_ioc_type(ioc)
            if ioc_type:
                data.append((ioc, ioc_type))

        data.sort(key=lambda ioc_type: ioc_type[1])

        return {
            ioc_type: [iocs[0] for iocs in ioc_group]
            for ioc_type, ioc_group in groupby(data, key=lambda ioc_type: ioc_type[1])
            if ioc_type
        }

    @property
    def content(self) -> EnrichedContent:
        """Return content attribute."""
        return self.__content

    @property
    def max_workers(self) -> int:
        """Get number of threads used."""
        max_workers = self.config.settings.get(ENRICHMENT_STANZA_NAME, {}).get(MAX_WORKERS_SETTING)
        if max_workers is None:
            max_workers = MAX_WORKERS_DEFAULT
            self.log.warning('max_workers not set, using default: {}'.format(max_workers))
        return max_workers

    @property
    def fields(self) -> dict:
        """Get Enrichment fields."""
        enrich_fields = self.config.settings.get(ENRICHMENT_STANZA_NAME, {}).get(FIELDS_SETTING)
        if not enrich_fields:
            enrich_fields = FIELDS_DEFAULT
            self.log.warning(
                'fields was not set, using default {}'.format(json.dumps(enrich_fields)),
            )
        return enrich_fields

    @max_workers.setter
    def max_workers(self, max_workers: int) -> None:
        """Set number of threads to be used if 0 it will be single thread (MainThread)."""
        if not isinstance(max_workers, int):
            raise TypeError('{} must be an integer value'.format(MAX_WORKERS_SETTING))

        if max_workers < 0:
            raise ValueError('{} must be at least positive'.format(MAX_WORKERS_SETTING))

        self.config.update_setting(ENRICHMENT_STANZA_NAME, MAX_WORKERS_SETTING, max_workers)

    @fields.setter
    def fields(self, enrich_fields: dict) -> None:
        """Set Enrichment fields.

        Args:
            enrich_fields (dict): format {"ip": ["risk.rule"], "hash": [...]}

        Raises:
            ValueError: fields is not a dict
            ValueError: if fields is empty
            ValueError: if a key of Enriched field is not in the supported IOC types
        """
        if not isinstance(enrich_fields, dict):
            msg = 'Enrichment fields must be a dictionary of format like {"ip": ["risk.score"]}.'
            raise ValueError(msg)

        if not len(enrich_fields):
            msg = 'Enrichment fields cannot be empty.'
            raise ValueError(msg)

        if any(key not in _IOC_TYPES for key in enrich_fields):
            msg = 'Enrichment field keys must be any of the supported types {}.'.format(_IOC_TYPES)
            raise ValueError(msg)

        if any(not isinstance(value, list) for value in enrich_fields.values()):
            msg = 'Enrichment field values must be lists.'
            raise ValueError(msg)

        self.config.update_setting(ENRICHMENT_STANZA_NAME, FIELDS_SETTING, enrich_fields)
