##################################### TERMS OF USE ###########################################
# The following code is provided for demonstration purpose only, and should not be used      #
# without independent verification. Recorded Future makes no representations or warranties,  #
# express, implied, statutory, or otherwise, regarding any aspect of this code or of the     #
# information it may retrieve, and provides it both strictly “as-is” and without assuming    #
# responsibility for any information it may retrieve. Recorded Future shall not be liable    #
# for, and you assume all risk of using, the foregoing. By using this code, Customer         #
# represents that it is solely responsible for having all necessary licenses, permissions,   #
# rights, and/or consents to connect to third party APIs, and that it is solely responsible  #
# for having all necessary licenses, permissions, rights, and/or consents to any data        #
# accessed from any third party API.                                                         #
##############################################################################################

import json
import logging
from typing import Union

from requests.exceptions import ConnectionError, ConnectTimeout, HTTPError, ReadTimeout

from ..config.config import Config
from ..endpoints import EP_ENTITY_MATCH
from ..rf_client import RFClient
from .errors import MatchApiError, MatchError


class RFMatchMgr:
    """Manages requests for Recorded Future Entity Match API."""

    def __init__(self, cfg: Config = None, rf_token: str = None):
        self.log = logging.getLogger(__name__)
        self.config = cfg if cfg is not None else Config()
        self.rf_client = (
            RFClient(self.config, rf_token=rf_token)
            if rf_token is not None
            else RFClient(self.config)
        )

    def match(self, entity_name, entity_type=None):
        """Matches a text string on entity match API.

        Args:
            entity_name (str): entity name as string
            entity_type (str, optional): defaults to None, is ignored on None

        Raises:
            MatchApiError: raised if Entity Match API call fails

        Returns:
            list: list of dicts with id, name, type for exact name matches
        """
        request_body = {'name': entity_name}
        if entity_type is not None:
            request_body['type'] = [entity_type]
        try:
            response = self.rf_client.post_request(EP_ENTITY_MATCH, json.dumps(request_body))
        except (HTTPError, ConnectionError, ReadTimeout, ConnectTimeout) as e:
            raise MatchApiError(str(e))
        return response.json()

    def resolve_entity_id(self, entity_name: str, entity_type: str = None):
        """Resolves an entity name and type to an ID.

        Args:
            entity (str, tuple): entity string ID or tuple of (name, type)

        Raises:
            MatchError: raised when an entity name matches none or multiple entities
            ValueError: raised when entity is not a string or tuple

        Returns:
            str: entity ID
        """
        matches = self.match(entity_name, entity_type=entity_type)
        if len(matches) == 0:
            message = "No match found for string '{0}'".format(entity_name)
            raise MatchError(message)
        elif len(matches) > 1:
            exact_count = 0
            resolved_id = None
            for match in matches:
                if match['name'] == entity_name:
                    if entity_type is not None and match['type'] != entity_type:
                        continue
                    resolved_id = match['id']
                    exact_count += 1
            if (not resolved_id) or exact_count > 1:
                message = "Multiple matches found for '{}'".format(entity_name)
                if entity_type is None:
                    message += '. No type set. Consider specifying entity type'
                else:
                    message += " of type '{}'".format(entity_type)
                raise MatchError(message)
        else:
            resolved_id = matches[0]['id']

        return resolved_id

    def resolve_entity_ids(self, entities: Union[list, tuple]):
        """Resolves a list of entities to IDs.

        Args:
            entities (list): list of entity name strings or entity name, type tuples

        Raises:
            TypeError: raised when entities is not a sequence

        Returns:
            generator: generator of entity IDs. If no match is found, None is yielded
        """
        if not isinstance(entities, list):
            raise TypeError('entities must be a list. Has type {}'.format(type(entities)))
        results = []
        for entity in entities:
            if isinstance(entity, str):
                entity_name = entity
                entity_type = None
            elif isinstance(entity, (list, tuple)):
                if len(entity) != 2:
                    raise ValueError('entity must be a string or sequence of (name, type)')
                entity_name = entity[0]
                entity_type = entity[1]
            try:
                resolved_id = self.resolve_entity_id(entity_name, entity_type=entity_type)
                results.append(resolved_id)
            except MatchError:
                self.log.info('No match found for entity {}'.format(entity))
                results.append(None)

        return results
