##################################### TERMS OF USE ###########################################
# The following code is provided for demonstration purpose only, and should not be used      #
# without independent verification. Recorded Future makes no representations or warranties,  #
# express, implied, statutory, or otherwise, regarding any aspect of this code or of the     #
# information it may retrieve, and provides it both strictly “as-is” and without assuming    #
# responsibility for any information it may retrieve. Recorded Future shall not be liable    #
# for, and you assume all risk of using, the foregoing. By using this code, Customer         #
# represents that it is solely responsible for having all necessary licenses, permissions,   #
# rights, and/or consents to connect to third party APIs, and that it is solely responsible  #
# for having all necessary licenses, permissions, rights, and/or consents to any data        #
# accessed from any third party API.                                                         #
##############################################################################################
import csv
import functools
import inspect
import json
import logging
import os
import pathlib
import platform
import re
import sys
from concurrent.futures import Future, ThreadPoolExecutor
from datetime import datetime, timedelta
from typing import Any, Callable, Dict, List, Union

from dateutil.parser import ParserError
from dateutil.parser import parse as date_parse
from jsonpath_ng import exceptions, ext, jsonpath
from requests.exceptions import (
    ConnectionError,
    ConnectTimeout,
    HTTPError,
    JSONDecodeError,
    ReadTimeout,
    SSLError,
)

from ..constants import ROOT_DIR
from ..exceptions import ReadFileError, WriteFileError

LOG = logging.getLogger('psengine.helpers')
VALID_TIME_REGEX = r'^(-?)([1-9]?[0-9]+[dDhH])$'


def connection_exceptions(ignore_status_code: List[int], exception_to_raise: Exception):
    """Decorator for handling HTTP related errors.

    Args:
        ignore_status_code (List[int]): list of status codes to be ignored - dont raise exception
        exception_to_raise (Exception): exception to raise in case of error. It should be based on
            the function that is decorated

    Raises:
        exception_to_raise

    Returns:
        Any: whatever the function decorated returns

    """

    def wrapper(func):
        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            self = args[0]

            try:
                return func(*args, **kwargs)
            except HTTPError as err:
                if err.response is not None and err.response.status_code in ignore_status_code:
                    msg = (
                        f"Requested data by {func.__name__} wasn't found or you cannot view it. "
                        f'Error: {err}'
                    )
                    self.log.info(msg)
                    return None

                msg = f'HTTPError in {func.__name__}. Error: {err}'
                self.log.error(msg)
                exception = exception_to_raise()
                exception.response = err.response
                raise exception from err

            except (ConnectTimeout, ConnectionError, ReadTimeout) as err:
                msg = f'Connection error in {func.__name__}. Error: {err}'
                self.log.error(msg)
                exception_to_raise.response = err.response
                raise exception_to_raise from err

            except (OSError, SSLError) as err:
                self.log.error(
                    f'Possible error with custom certificate {err} when calling {func.__name__}'
                )
                raise exception_to_raise from err

            except (JSONDecodeError, KeyError) as err:
                self.log.error(
                    'Incorrect data returned by {}. Error: {}'.format(func.__name__, err)
                )
                raise exception_to_raise from err

        return wrapped

    return wrapper


def basic_args_validation(func):
    """Validate data passed to the function
        - Any argument without type hint will not be validated
        - Any argument with type hint coming from typing will not be validated.

    If a param has a default value the type will be checked
    If a param does not have a default value the type will be checked and the value cannot be empty

    """
    func_signature = inspect.signature(func)
    arg_types = {
        name: param.annotation
        for name, param in func_signature.parameters.items()
        if param.annotation is not param.empty
    }
    arg_defaults = {
        name: param.default
        for name, param in func_signature.parameters.items()
        if param.default is not param.empty
    }

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        bound_args = func_signature.bind_partial(*args, **kwargs)
        bound_args.apply_defaults()
        self = args[0]

        for arg_name, arg_value in bound_args.arguments.items():
            expected_type = arg_types.get(arg_name)
            arg_value = bound_args.arguments[arg_name]

            # skip typing related hints
            if hasattr(expected_type, '__module__') and expected_type.__module__ == 'typing':
                continue

            if expected_type and not isinstance(arg_value, expected_type):
                msg = 'Argument {} is not of type {}.'.format(arg_name, expected_type)
                self.log.error(msg)
                raise TypeError(msg)

            if not (arg_name in arg_defaults or arg_value) and arg_name != 'kwargs':
                msg = 'Argument {} must not be empty.'.format(arg_name)
                self.log.error(msg)
                raise ValueError(msg)

        return func(*args, **kwargs)

    return wrapper


class TimeHelpers:
    """Helpers for time related functions."""

    @classmethod
    def rel_time_to_date(cls, relative_time):
        """Convert a relative time to a date. Minutes not supported
        For example:
        1h - > Return -1h from NOW
        1d - > Return -1d from NOW.

        Args:
            relative_time (str): 7d, 3h, etc..

        Raises:
            ValueError: if the relative time is invalid

        Returns:
            str: time delta, for example: 2022-08-08T13:11
        """
        logger = logging.getLogger(__name__)
        match = re.match(VALID_TIME_REGEX, relative_time)
        if match is None:
            raise ValueError(
                "Invalid relative time '{}'. Accepted format: [-|][integer][h|d]".format(
                    relative_time,
                ),
            )
        else:
            relative_time = match.groups()[-1]
            time_now = datetime.utcnow()
            digit = int(re.findall(r'^\d+', relative_time)[0])
            if relative_time.endswith('d'):
                subtracted = (time_now - timedelta(days=digit)).strftime('%Y-%m-%dT%H:%M')
            else:
                subtracted = (time_now - timedelta(hours=digit)).strftime('%Y-%m-%dT%H:%M')
            logger.debug('UTC Time now: {}'.format(time_now))
            logger.debug('Relative time -{} to date: {}'.format(relative_time, subtracted))

            return subtracted

    @classmethod
    def is_rel_time_valid(cls, rel_time) -> bool:
        """Helper function to determine if relative time expression
        is valid.

        Args:
            rel_time (str): relative time

        Returns:
            bool: True if valid, False otherwise
        """
        if rel_time is None:
            return False

        match = re.match(VALID_TIME_REGEX, rel_time)
        return bool(match)

    @classmethod
    def is_valid_time_range(cls, range_: str) -> bool:
        """Verifies if an ISO 8601 compliant time range was specified, for example:
            [2017-07-30,2017-07-31]
            (2017-07-30,2017-07-31)
            [2017-07-30,2017-07-31)
            [2017-07-30,)
            [,2017-07-31)
            https://www.elastic.co/guide/en/elasticsearch/reference/current/date.html.

        Args:
            range (str): time range

        Returns:
            bool: True if valid, False otherwise
        """
        if range_ is None:
            return False

        match = re.match(r'^(\[|\()(.*)?,\s*(.*)?(\]|\))$', range_)
        if match is None:
            return False
        else:
            start_time = match.groups()[1]
            end_time = match.groups()[2]
            try:
                if start_time != '' and not TimeHelpers.is_rel_time_valid(start_time):
                    date_parse(start_time)
                if end_time != '' and not TimeHelpers.is_rel_time_valid(end_time):
                    date_parse(end_time)
            except ParserError:
                return False
            return True


class FormattingHelpers:
    """Helpers for formatting related functions."""

    @classmethod
    def _ellipsize(cls, string, length):
        """Truncate and add ellipsis to a string.

        Args:
            string (str): string to truncate
            length (int): max length of string (including ellipsis)

        Returns:
            str: truncated string with ellipsis
        """
        if not isinstance(string, str):
            string = str(string)
        if len(string) > length - 1:
            return string[: length - 4] + '...'
        return string

    @classmethod
    def format_output(cls, lengths, args, right_alignment=True):
        """Generate columnar data.

        Default alignment is right, but can be changed to left

        Args:
            lengths (list): list of integer column lengths
            args (list): list of strings to format
            right_alignment (bool, optional): right align columns. Defaults to True.

        Returns:
            str: columnar data aligned and spaced
        """
        if len(lengths) != len(args):
            raise ValueError('Length of lengths and args must be the same')
        alignment_key = '>' if right_alignment else '<'
        spacing = ''
        for length in lengths:
            spacing += '{:' + alignment_key + str(length) + '} '

        outputs = []
        for length, arg in zip(lengths, args):
            outputs.append(cls._ellipsize(arg, length))

        return spacing.format(*outputs)

    @classmethod
    def cleanup_ai_insights(cls, ai_insights: str) -> str:
        """Clean up RF AI Insights to avoid markdown rendering issues.

        Args:
            ai_insights (str): ai insights

        Returns:
            str: cleaned up ai insights
        """
        # Replace \n so markdown doesn't add new lines (breaks rendering otherwise)
        ai_insights = ai_insights.replace('\n', ' ')

        # Replace '1. ' with '1.' so markdown doesn't indent the paragraph
        ai_insights = ai_insights.replace('1. ', '1.')

        return ai_insights


class OSHelpers:
    """Helpers for OS related functions."""

    @classmethod
    def os_platform(cls):
        """Get the OS platform information, for example: macOS-13.0-x86_64-i386-64bit.

        Returns:
            str: OS platform info, if unavailable return None
        """
        os_info = None
        os_info = platform.platform(aliased=True, terse=False)

        return os_info

    @classmethod
    def mkdir(cls, path):
        """Safely create a directory.

        Args:
            path (str): path to directory

        Raises:
            ValueError: if path is not a string
            ValueError: if path is empty
            WriteFileError: if directory is not writeable

        Returns:
            str: path to directory created
        """
        LOG.debug('Creating directory: {}'.format(path))
        if path == '':
            raise ValueError('path cannot be empty')
        if not os.path.isabs(path):
            path = os.path.join(sys.path[0], path)
        if os.path.isdir(path) and os.access(path, os.W_OK):
            return path
        try:
            # Create the directory, do not raise an error if it already exists
            pathlib.Path(path).mkdir(parents=True, exist_ok=True)
        except PermissionError:
            raise WriteFileError('Directory {} is not writeable'.format(path))
        # In case it already exists, check if it is writeable
        if not os.access(path, os.W_OK):
            raise WriteFileError('Directory {} is not writeable'.format(path))
        return path

    @classmethod
    def check_dir(cls, path) -> bool:
        """Check if a given path exists or could be created as a directory.

        Args:
            path (str): path to check

        Raises:
            ValueError: if path is not a string

        Returns:
            bool: True if path exists or is writeable, False otherwise
        """
        if not isinstance(path, str):
            raise TypeError('path must be a string')
        if path == '':
            return False
        if not os.path.isabs(path):
            path = os.path.join(sys.path[0], path)
        if os.path.exists(path):
            if not os.path.isdir(path):
                return False
            # check if directory is writable
            return bool(os.access(path, os.W_OK))
        else:
            # In case path is some/long/path, check that all existing dirs are writeable
            basepath = ''
            for subpath in os.path.split(path):
                basepath = os.path.join(basepath, subpath)
                if not os.path.exists(basepath):
                    return True
                if not os.access(basepath, os.W_OK):
                    return False
            return True


class FileHelpers:
    """Helpers for file related functions."""

    @classmethod
    def read_csv(cls, csv_file, as_dict=False, single_column=False):
        """Reads all rows from a CSV.

        It is the client's responsibility to ensure column headers are handled appropriately.

        Using ``as_dict`` will reader the CSV with ``csv.DictReader``, which treats the first row
        as column headers. For example with CSV::

            Name,ID,Level
            Patrick,321,4
            Ernest,123,8

        ``as_dict=True`` will return a list of dictionaries keyed by header names::

            [{'Name': 'Patrick', 'ID': '321', 'Level': '4'},
            {'Name': 'Ernest', 'ID': '123', 'Level': '8'}]

        ``single_column=True`` will return a list of only the first column of the CSV
        as strings (note that ``as_dict=False``, ``single_column=False`` returns a list of lists)::

            ['Name', 'Patrick', 'Ernest']

        ``as_dict=False`` and ``single_column=False`` will return a list of lists::

            [['Name', 'ID', 'Level'], ['Patrick', '321', '4'], ['Ernest', '123', '8']]

        **Args**:
            - csv_file (str): path to CSV file
            - as_dict (bool, optional): return entities as a dict. Defaults to False.
            - single_column (bool, optional): return only entities (not lists) from first column
            cannot be used with ``as_dict``. Defaults to False.

        **Raises**:
            ValueError: If both as_dict and single_column are True
            ReadFileError: If file is not found or has restricted access

        **Returns**:
            list of rows from CSV
        """
        if as_dict and single_column:
            raise ValueError('Cannot use as_dict and single_column together')
        if not os.path.abspath(csv_file):
            LOG.debug(
                '{} is not an absolute path. Attempting to find it in {}'.format(
                    csv_file,
                    ROOT_DIR,
                ),
            )
            file_path = os.path.join(ROOT_DIR, csv_file)
        else:
            file_path = csv_file
        try:
            with open(file_path, 'r') as file_obj:
                if as_dict is True:
                    reader = csv.DictReader(file_obj)
                    return list(reader)

                reader = csv.reader(file_obj)
                if single_column is True:
                    return [row[0] for row in reader]

                return list(reader)
        except OSError as oe:
            raise ReadFileError('Error reading entity file: {}'.format(str(oe)))

    @classmethod
    def write_file(cls, to_write, output_directory, fname) -> None:
        """Write bytes to file.

        Args:
            to_write (bytes): bytes to write to file
            output_directory (str): path to directory to write file
            fname (str): name of file to write

        Side effects:
            Writes to file
        """
        LOG.info('Writing file: {}'.format(fname))
        try:
            if not os.path.isdir(output_directory):
                output_directory = OSHelpers.mkdir(output_directory)
            full_path = os.path.join(output_directory, fname)
            with open(full_path, 'wb') as f:
                f.write(to_write)
        except OSError as err:
            raise WriteFileError("Error writing file '{}': {}".format(err.filename, str(err)))

        LOG.info('File written to: {}'.format(full_path))


class JSONParsingHelper:
    """Helpers for parsing JSON objects."""

    @classmethod
    def _value_adder(cls, data: Dict[str, Any], path_list: List[str], value: Any) -> None:
        """Set value in nested dictionary using list of keys."""
        cls._schema_builder(data, path_list[:-1])[path_list[-1]] = value

    @classmethod
    def _schema_builder(cls, data: Dict[str, Any], path_list: List[str]) -> Dict[str, Any]:
        """Create the schema along the path and return the innermost value."""
        for subpath in path_list:
            if subpath not in data:
                data[subpath] = {}
            # move inside the new structure within subpath
            data = data[subpath]
        return data

    @classmethod
    def _dict_to_list(cls, data: Dict[str, Any]) -> Union[List, Dict[str, Any]]:
        """If all keys in a dictionary are integers and form a consecutive
        sequence starting from 0, convert it to a list.
        """
        if data and isinstance(data, dict):
            keys = sorted(data.keys())

            if any(key.isdigit() for key in keys):
                return [cls._dict_to_list(data.get(key)) for key in keys]
            else:
                return {key: cls._dict_to_list(value) for key, value in data.items()}
        else:
            return data

    @classmethod
    def _cleanup_path(cls, full_path: Union[jsonpath.Fields, jsonpath.Child]) -> List[str]:
        return (
            str(full_path)
            .replace('[*]', '')
            .replace('[', '')
            .replace(']', '')
            .replace("'", '')
            .split('.')
        )

    @classmethod
    def parse_payload(cls, payload: Dict[str, Any], fields: List[str]) -> dict:
        """Parses a payload for specified fields, maintaining original structure.

        Args:
            payload (Dict[str, Any]): The input payload to parse.
            fields (List[str]): A list of fields to extract from the payload.

        Returns:
            Dict[str, Union[Dict[str, Any], List]]: dict with the parsed data, and original struct.
        """
        try:
            parsed = [ext.parse(field) for field in fields]
        except (exceptions.JsonPathParserError, exceptions.JsonPathLexerError) as err:
            raise ValueError('Invalid JSON Path: {}'.format(err))

        structure = {}
        for field in parsed:
            matches = [match for match in field.find(payload)]
            for match in matches:
                cls._value_adder(structure, cls._cleanup_path(match.full_path), match.value)

        return cls._dict_to_list(structure)


class RiskListParsingHelper:
    """Helpers for parsing Risklist files."""

    @classmethod
    def parse_risklist(cls, file: str, fields: List[str]) -> List[List[str]]:
        """Parse risklist from list of fields.

        Args:
            file (str): path to file
            fields (List[str]): list of fields to return.
                Example: ['Name', 'Risk',  'EvidenceDetails[?(@.Criticality > 2)].EvidenceString']

        Raises:
            (FileNotFoundError, IsADirectoryError, PermissionError, OSError, TypeError): file error

        Returns:
            List[List[str]]: Parsed data
        """
        logger = logging.getLogger(__name__)
        try:
            with pathlib.Path(file).open() as csv_file:
                res = [cls._extract_fields(line, fields) for line in csv.DictReader(csv_file)]
            return res
        except (
            FileNotFoundError,
            IsADirectoryError,
            PermissionError,
            OSError,
            TypeError,
        ) as file_error:
            logger.error(file_error)
            raise file_error

    @classmethod
    def _extract_fields(cls, line: Dict[str, Any], fields: List[str]) -> List[str]:
        """For each line of risklist extract the all the fields.

        Args:
            line (dict[str, Any]): line of file
            fields (List[str]): list of fields to return.

        Raises:
            ValueError: From JSONParsingHelper

        Returns:
            List[str]: extracted fields
        """
        logger = logging.getLogger(__name__)
        parsed_data = []

        for field in fields:
            if '.' in field:
                header_field = field.split('.')[0].split('[')[0]
                try:
                    data = json.loads(line[header_field])
                except (json.JSONDecodeError, TypeError, KeyError) as exp:
                    logger.error(exp)
                    raise exp

                try:
                    parsed_data.append(JSONParsingHelper().parse_payload(data, [field]))
                except ValueError:
                    logger.error('{} is not a valid JSON Path'.format(field))
                    raise ValueError('{} is not a valid JSON Path'.format(field)) from None
            else:
                try:
                    parsed_data.append(line[field])
                except KeyError:
                    pass
        return parsed_data


class MultiThreadingHelper:
    """Multithreading class."""

    @classmethod
    def multithread_it(cls, max_workers: int, func: Callable, *, iterator, **kargs) -> List[Future]:
        """The class can be used in the following way:

        Given a single thread code like::

            def _lookup_alert(self, alert_id, index, total_num_of_alerts):
                ...

            def all_alerts(self, alerts):
                res = []
                for index, alert_id in enumerate(alert_ids_to_fetch):
                    res.append(self._lookup_alert(alert_id, index, len(alert_ids_to_fetch)))

        It can be rewritten like::

            def _lookup_alert(self, alert_id, index, total_num_of_alerts):
                ...

            def all_alerts(self, alerts):
                results = MultiThreadingHelper().multithread_it(
                    self.max_workers,
                    self._lookup_alert,
                    iterator=alert_ids_to_fetch,
                    total_num_of_alerts=len(alert_ids_to_fetch)
                )

        The result can be viewed with::

            for result in results:
                print(result.result())

        Args:
            max_workers (int): Number of threads to use.
            func (Callable): Function to be executed in parallel.
            iterator (_type_): The list of elements to be dispatched to the threads.
                Example: This can be the list of alerts to download, the IOCs, etc.
            **kwargs: Any other argument that the function needs for execution.
                Example: The Alert type, or the IOC type.

        Returns:
            list[Future]: An object defined in concurrent.futures.Future.
            Refer to official python docs for more details:
            https://docs.python.org/3/library/concurrent.futures.html#concurrent.futures.Future
        """
        with ThreadPoolExecutor(max_workers=max_workers) as pool:
            futures = [pool.submit(func, element, **kargs) for element in iterator]

        return futures
