##################################### TERMS OF USE ###########################################
# The following code is provided for demonstration purpose only, and should not be used      #
# without independent verification. Recorded Future makes no representations or warranties,  #
# express, implied, statutory, or otherwise, regarding any aspect of this code or of the     #
# information it may retrieve, and provides it both strictly “as-is” and without assuming    #
# responsibility for any information it may retrieve. Recorded Future shall not be liable    #
# for, and you assume all risk of using, the foregoing. By using this code, Customer         #
# represents that it is solely responsible for having all necessary licenses, permissions,   #
# rights, and/or consents to connect to third party APIs, and that it is solely responsible  #
# for having all necessary licenses, permissions, rights, and/or consents to any data        #
# accessed from any third party API.                                                         #
##############################################################################################

import collections.abc
import copy
import json
import logging

import jsonpath_ng

from .errors import MultipleReadError, NoMatchError

alert_schema = {
    'alert': {'id': None, 'title': None, 'triggered': None, 'url': None, 'type': None},
    'rule': {},
    'counts': {},
    'review': {},
}

root_map = {
    'data.id': 'alert.id',
    'data.title': 'alert.title',
    'data.triggered': 'alert.triggered',
    'data.url': 'alert.url',
    'data.type': 'alert.type',
    'data.rule': 'rule',
    'data.counts': 'counts',
    'data.review': 'review',
}


class RFAlert:
    """Recorded Future alert parsing class."""

    def __init__(self):
        self.log = logging.getLogger(__name__)

    def parse(self, alert):
        """Parses raw input alert into the following fields:

        self.rfalert - common fields present in every alert

        Args:
            alert (dict): Raw alert from Recorded Future API

        Returns:
            dict: restructured Recorded Future alert
        """
        references = []
        raw_alert = copy.deepcopy(alert)
        common_fields = self._populate_root_keys(copy.deepcopy(alert_schema), alert)
        alertid = self.get_alertid(alert)
        self.log.debug('Parsing alert {}'.format(alertid))
        entities = self._get_array_vals(raw_alert, 'data.entities[*]')

        if len(entities) == 0:
            return [common_fields]
        for entity in entities:
            self._prune_empties(entity)
            docs, risk = self._get_docs_risk_objs(entity)
            if docs is None:
                self.log.debug('  Creating single copy')
                local_copy = copy.deepcopy(common_fields)
                local_copy.update(entity)
                if risk:
                    local_copy['risk'] = risk
                references.append(local_copy)
            else:
                for doc in docs:
                    refs = self._get_array_vals(doc, 'references[*]')
                    del doc['references']
                    self._prune_empties(doc)
                    copies = [copy.deepcopy(common_fields) for x in range(len(refs))]
                    for idx, ref in enumerate(refs):
                        copies[idx].update(ref)
                        copies[idx].update(entity)
                        copies[idx].update(doc)
                        if risk is not None:
                            copies[idx]['risk'] = risk
                    references += copies

        self.log.debug('Parsing complete')
        return references

    def get_alertid(self, alert):
        """Get an alert id.

        Args:
            alert (dict): Raw alert from Recorded Future API or a standard alert
            parsed by the parse() func

        Returns:
            str: alert id
        """
        try:
            alertid = alert['data']['id']
        except KeyError:
            try:
                alertid = alert['id']
            except KeyError:
                self.log.error(
                    'Cannot access alert ID for alert {}'.format(json.dumps(alert, indent=2)),
                )
                return None
        return alertid

    def rollup(self, alerts):
        """Combine fields from alerts preprocessed by PSEngine.

        Args:
            alerts (list): standard formatted alerts

        Returns:
            list: combined alert objects by alert ID
        """
        alerts_by_id = {}
        # create a dict keyed by alert ID, value is a list of all objects from that ID
        for alert in alerts:
            if alert['alert']['id'] in alerts_by_id:
                alerts_by_id[alert['alert']['id']].append(alert)
            else:
                alerts_by_id[alert['alert']['id']] = [alert]

        rollups = []
        for alert_id, alert_contents in alerts_by_id.items():
            self.log.debug('Rolling up alert {}'.format(alert_id))
            rollup = {
                'alert': alert_contents[0]['alert'],
                'rule': alert_contents[0]['rule'],
                'counts': alert_contents[0]['counts'],
                'review': alert_contents[0]['review'],
            }
            rollup['fragments'] = list(map(lambda x: x.get('fragment', None), alert_contents))
            rollup['entities'] = list(map(lambda x: x.get('entities', []), alert_contents))
            rollup['languages'] = list(map(lambda x: x.get('language', None), alert_contents))
            rollup['ids'] = list(map(lambda x: x.get('id', None), alert_contents))
            rollup['sources'] = list(map(lambda x: x.get('source', None), alert_contents))
            rollup['authors'] = list(map(lambda x: x.get('authors', None), alert_contents))
            rollup['titles'] = list(map(lambda x: x.get('title', None), alert_contents))
            rollup['urls'] = list(map(lambda x: x.get('url', None), alert_contents))
            rollup['subjects'] = list(map(lambda x: x.get('entity', None), alert_contents))
            rollup['trends'] = list(map(lambda x: x.get('trend', None), alert_contents))
            rollup['risks'] = list(map(lambda x: x.get('risk', None), alert_contents))
            rollup['noteId'] = list(map(lambda x: x.get('noteId', None), alert_contents))
            rollups.append(rollup)

        return rollups

    def _get_docs_risk_objs(self, entity):
        """Delete documents and risk objects in entity and return them.

        Args:
            entity (dict): Entity from a Recorded Future alert

        Returns:
            tuple: docs array, risk dict tuple
        """
        docs = None
        risk = None
        if 'risk' in entity and not self._is_empty(entity['risk']):
            # Get documents only from risk obj if it is present and not empty
            if 'documents' in entity['risk'] and not self._is_empty(entity['risk']['documents']):
                docs = copy.deepcopy(entity['risk']['documents'])
                del entity['risk']['documents']
            # Get risk obj. Entity info - risk obj is used to update parsed alert
            risk = copy.deepcopy(entity['risk'])
            del entity['risk']
        # Get docs obj outside of risk obj
        for array_path in self._get_array_paths(entity):
            if 'documents' in array_path and 'risk' not in array_path:
                docs = self._get_val(entity, array_path)
                self._delete_key(entity, array_path)

        return docs, risk

    def _assign_val(self, original, original_path, assignee, assignee_path):
        """Given 2 alerts, assigns a value from one path to the new path.

        Args:
            assignee (dict): unified alert to produce
            original (dict): alert from RF API
            assignee_path (string): dot-nested string indicating a JSONPath
            original_path (string): dot-nested string indicating a JSONPath
        """
        jsonpath_expr = jsonpath_ng.parse(original_path)
        match = jsonpath_expr.find(original)
        if len(match) > 1:
            raise MultipleReadError
        val = match[0].value
        jsonpath_expr = jsonpath_ng.parse(assignee_path)
        assignee = jsonpath_expr.update(assignee, val)

        return assignee

    def _populate_root_keys(self, assignee, original):
        """Sets fields in the alert that are in every alert.

        Args:
            assignee (dict): Alert to populate with new info
            original (dict): Alert to source info
        """
        for key, value in root_map.items():
            assignee = self._assign_val(original, key, assignee, value)

        return assignee

    def _get_leaves(self, alert):
        """Returns a list of all paths that are maximal in the alert.

        Args:
            alert (dict): Source alert to find paths in

        Returns:
            list: String list of paths with leaves in the alert
        """
        paths = self._get_paths(alert)
        paths = list(map(lambda x: x.replace('.[', '[*'), paths))

        return list(filter(lambda x: self._is_maximal(x, paths), paths))

    def _get_paths(self, source):
        """Returns a list of all JSONPaths in source.

        Args:
            source (dict): dictionary to find paths from

        Returns:
            list: list of string JSONPaths in the source
        """
        paths = []
        if isinstance(source, collections.abc.MutableMapping):
            for key, value in source.items():
                if key not in paths:
                    paths.append(key)
                for x in self._get_paths(value):
                    if key + '.' + x not in paths:
                        paths.append(key + '.' + x)
        elif isinstance(source, collections.abc.Sequence) and not isinstance(source, str):
            for x in source:
                for y in self._get_paths(x):
                    if '[].' + y not in paths:
                        paths.append('[].' + y)

        return paths

    def _is_maximal(self, path, paths):
        """Predicate, returns True when path is not a proper substring of any x in paths.

        For 2 strings s and t, s is a proper substring of t iff:
        1. s is a substring of t
        2. s does not equal t

        Args:
            path (str): path to check for maximality
            paths (str): list of paths to check path against

        Returns:
            bool: True if path is maximal among paths
        """
        answer = True
        for x in paths:
            if path in x and path != x:
                answer = False

        return answer

    def _get_array_paths(self, alert):
        """Given alert JSON, returns JSONPath Expression for top-level array.

        Args:
            paths (list): A list of JSONPath strings
        """
        paths = self._get_leaves(alert)
        array_paths = filter(lambda x: '[*]' in x, paths)
        truncated_array_paths = map(lambda x: x.split('[*]')[0], array_paths)
        jsonpath_list = list(set(list(truncated_array_paths)))

        return jsonpath_list

    def _get_array_vals(self, alert, path):
        """Gets values for all elements at path in alert.

        Args:
            alert (dict): Source to get elements from
            path (str): Path to find elements from

        Yields:
            any: Whatever value is at path in alert
        """
        # TODO: Use regex to verify path ends in '[*]'
        jsonpath_expr = jsonpath_ng.parse(path)
        matches = jsonpath_expr.find(alert)
        return [match.value for match in matches]

    def _get_val(self, alert, path):
        """Returns the first value at at path in alert.

        Args:
            alert (dict): Alert JSON
            path (string): JSONPath to desired value

        Returns:
            any: Value at that path in the alert
        """
        jsonpath_expr = jsonpath_ng.parse(path)
        try:
            val = jsonpath_expr.find(alert)[0].value
        except IndexError:
            raise NoMatchError('No match for path {}'.format(path))

        return val

    def _delete_key(self, alert, path) -> None:
        """Deletes the key at the end of the JSONPath path in alert.

        Args:
            alert (dict): Alert to delete path from
            key (str): Path to delete key
        """
        for access in path.split('.')[:-1]:
            alert = alert[access]
        try:
            del alert[path.split('.')[-1]]
        except KeyError:
            raise

    def _is_empty(self, item) -> bool:
        """Predicate that returns True when item contains no useful information.

        Args:
            item (any): The item to check whether empty

        Returns:
            bool: Boolean indicating if the item is empty
        """
        return bool(item == '' or item == 0 or item is None or item == [] or item == {})

    def _prune_empties(self, alert) -> None:
        """Deletes all empty top-level keys in alert.

        Args:
            alert (dict): a dict to delete empty keys from
        """
        to_delete = [key for key in alert if self._is_empty(alert[key])]

        for key in to_delete:
            del alert[key]
