##################################### TERMS OF USE ###########################################
# The following code is provided for demonstration purpose only, and should not be used      #
# without independent verification. Recorded Future makes no representations or warranties,  #
# express, implied, statutory, or otherwise, regarding any aspect of this code or of the     #
# information it may retrieve, and provides it both strictly “as-is” and without assuming    #
# responsibility for any information it may retrieve. Recorded Future shall not be liable    #
# for, and you assume all risk of using, the foregoing. By using this code, Customer         #
# represents that it is solely responsible for having all necessary licenses, permissions,   #
# rights, and/or consents to connect to third party APIs, and that it is solely responsible  #
# for having all necessary licenses, permissions, rights, and/or consents to any data        #
# accessed from any third party API.                                                         #
##############################################################################################

import json
import logging
import os
from datetime import datetime
from typing import Optional

from dateutil import parser
from requests import ConnectionError, ConnectTimeout, HTTPError, JSONDecodeError, ReadTimeout

from ..config.config import Config
from ..constants import (
    _API_RESPONSE_LIMIT,
    DEFAULT_KEY,
    ERR_MSG_KEY,
    RF_TOKEN_KWARG,
    ROOT_DIR,
    STANZA_FIELDS_KEY,
    TYPE_KEY,
    VALIDATOR_KEY,
)
from ..endpoints import EP_ALERT_LOOKUP, EP_ALERT_RULE, EP_ALERT_SEARCH, EP_ALERT_UPDATE
from ..exceptions import WriteFileError
from ..helpers.helpers import OSHelpers, TimeHelpers
from ..rf_client import RFClient
from .errors import AlertsFetchError, AlertsParseError, AlertUpdateError, NoRulesFoundError
from .rf_alert import RFAlert

ALERTS_STANZA_NAME = 'alerts'
TRIGGERED_SETTING = 'triggered'
TRIGGERED_DEFAULT = '-1d'
UPDATE_STATUS_SETTING = 'update_status'
UPDATE_STATUS_DEFAULT = False
FORMAT_SETTING = 'format'
FORMAT_DEFAULT = 'standard'
ALERT_RULES_SETTING = 'alert_rules'
ALERT_RULES_DEFAULT = ['*']
FETCH_STATUS_SETTING = 'fetch_status'
FETCH_STATUS_DEFAULT = 'no-action'
NEW_STATUS_SETTING = 'new_status'
NEW_STATUS_DEFAULT = 'pending'
OUTPUT_DIR_SETTING = 'output_dir'

DEFAULT_ALERTS_OUTPUT_DIR = os.path.join(ROOT_DIR, 'alerts')

ALERTS_OUTPUT_FNAME = 'rf_alerts_'

ALERT_STATUSES = [
    'unassigned',
    'assigned',
    'pending',
    'dismiss',
    'no-action',
    'actionable',
    'tuning',
]
ALERT_FORMATS = ['standard', 'raw']

ALERTS_STANZA_PARAMS = {
    STANZA_FIELDS_KEY: {
        FETCH_STATUS_SETTING: {
            DEFAULT_KEY: FETCH_STATUS_DEFAULT,
            TYPE_KEY: str,
            VALIDATOR_KEY: lambda x: x in ALERT_STATUSES,
            ERR_MSG_KEY: "[{}] '{}' must be one of {}".format(
                ALERTS_STANZA_NAME,
                FETCH_STATUS_SETTING,
                ALERT_STATUSES,
            ),
        },
        UPDATE_STATUS_SETTING: {
            DEFAULT_KEY: UPDATE_STATUS_DEFAULT,
            TYPE_KEY: bool,
            VALIDATOR_KEY: lambda x: x in [True, False],
            ERR_MSG_KEY: "[{}] '{}' must be one of [true, false]".format(
                ALERTS_STANZA_NAME,
                UPDATE_STATUS_SETTING,
            ),
        },
        NEW_STATUS_SETTING: {
            DEFAULT_KEY: NEW_STATUS_DEFAULT,
            TYPE_KEY: str,
            VALIDATOR_KEY: lambda x: x in ALERT_STATUSES,
            ERR_MSG_KEY: "[{}] '{}' must be one of {}".format(
                ALERTS_STANZA_NAME,
                NEW_STATUS_SETTING,
                ALERT_STATUSES,
            ),
        },
        FORMAT_SETTING: {
            DEFAULT_KEY: FORMAT_DEFAULT,
            TYPE_KEY: str,
            VALIDATOR_KEY: lambda x: x in ALERT_FORMATS,
            ERR_MSG_KEY: "[{}] '{}' must be one of {}".format(
                ALERTS_STANZA_NAME,
                FORMAT_SETTING,
                ALERT_FORMATS,
            ),
        },
        TRIGGERED_SETTING: {
            DEFAULT_KEY: TRIGGERED_DEFAULT,
            TYPE_KEY: str,
            VALIDATOR_KEY: lambda x: TimeHelpers.is_rel_time_valid(x)
            or TimeHelpers.is_valid_time_range(x),
            ERR_MSG_KEY: "[{}] '{}' must be a valid time range / relative time".format(
                ALERTS_STANZA_NAME,
                TRIGGERED_SETTING,
            ),
        },
        ALERT_RULES_SETTING: {
            DEFAULT_KEY: ALERT_RULES_DEFAULT,
            TYPE_KEY: list,
            VALIDATOR_KEY: lambda x: isinstance(x, list) and all(isinstance(y, str) for y in x),
            ERR_MSG_KEY: "[{}] '{}' must be a list of string rule names".format(
                ALERTS_STANZA_NAME,
                ALERT_RULES_SETTING,
            ),
        },
        OUTPUT_DIR_SETTING: {
            DEFAULT_KEY: DEFAULT_ALERTS_OUTPUT_DIR,
            TYPE_KEY: str,
            VALIDATOR_KEY: OSHelpers.check_dir,
            ERR_MSG_KEY: "[{}] '{}' directory is invalid or has restricted permissions".format(
                ALERTS_STANZA_NAME,
                OUTPUT_DIR_SETTING,
            ),
        },
    },
}


class RFAlertMgr:
    """Manages requests for Recorded Future alerts."""

    def __init__(self, cfg: Config = None, **kwargs):
        """Initialize RFAlertMgr.

        Args:
            cfg (Config, optional): Config object. Inits empty Config if none provided
            triggered (str, optional): Lookback time for alerts. Defaults to -1d
            format (str, optional): Output format for alerts. Defaults to standard
            alert_rules (str, optional): List of alert rules to fetch. Defaults to ["*"]
            update_status (bool, optional): Update alert status after fetching. Defaults to False
            fetch_status (str, optional): Status of alerts to fetch. Defaults to no-action
            new_status (str, optional): Status to update alerts after fetch. Defaults to pending
            output_dir (str, optional): Directory to write alerts to. Defaults to sys.path[0]/alerts
            rf_token (str, optional): Recorded Future API token. Defaults to None
        """
        self.log = logging.getLogger(__name__)
        self.raw_alerts = None
        self.config = cfg if cfg is not None else Config()
        rf_token = kwargs.pop(RF_TOKEN_KWARG) if RF_TOKEN_KWARG in kwargs else None
        self.config.parse_stanza(ALERTS_STANZA_NAME, ALERTS_STANZA_PARAMS)
        self.config.update_settings(
            [
                (ALERTS_STANZA_NAME, ALERTS_STANZA_PARAMS),
            ],
            kwargs,
        )
        self.rf_client = (
            RFClient(self.config, rf_token=rf_token)
            if rf_token is not None
            else RFClient(self.config)
        )
        self.rf_alert = RFAlert()
        self.config.log_settings(ALERTS_STANZA_NAME)

    def _fetch_rule_ids(self, alert_rules: list = None):
        """Fetches rule IDs for configured alert rules from Recorded Future API.

        Args:
            alert_rules (list, optional): List of alert rules to fetch. Defaults to ["*"]
        """
        if alert_rules is not None and not isinstance(alert_rules, list):
            raise TypeError('alert_rules must be a list')
        rule_ids = []
        alert_rules = alert_rules if alert_rules is not None else self.alert_rules

        self.log.debug('Fetching alert rule IDs')
        # If wildcard specified, no point in running the query below
        if '*' in alert_rules:
            return rule_ids

        for alert_rule in alert_rules:
            query = {'freetext': alert_rule}
            try:
                response = self.rf_client.make_paged_request(EP_ALERT_RULE, params=query)
            except (
                ConnectionError,
                ConnectTimeout,
                HTTPError,
                JSONDecodeError,
                ReadTimeout,
            ) as err:
                self.log.warning('Error looking up rule {}. Error: {}'.format(alert_rule, err))
                raise NoRulesFoundError(err)

            if len(response) == 0:
                self.log.warning('No rule IDs found for rule {}'.format(alert_rule))

            rule_ids += [rule['id'] for rule in response]

        self.log.debug('Rule IDs: {}'.format(rule_ids))
        if len(rule_ids) == 0:
            raise NoRulesFoundError('Rule names supplied but no rule IDs found')

        return rule_ids

    def fetch_rules(self):
        """Fetches rule IDs for configured alert rules from Recorded Future API."""
        params = {'limit': 1000}
        try:
            rule_response = self.rf_client.get_request(EP_ALERT_RULE, params=params)
        except (ConnectionError, ConnectTimeout, ReadTimeout, HTTPError) as err:
            raise AlertsFetchError(err)
        try:
            rule_info = rule_response.json()
        except JSONDecodeError as err:
            self.log.debug(
                '{} response cannot be decoded as JSON: {}'.format(
                    EP_ALERT_RULE,
                    rule_response.text,
                ),
            )
            raise AlertsFetchError(err)
        try:
            rules = rule_info['data']['results']
        except KeyError as err:
            self.log.debug(
                '{} response does not contain data.results: {}'.format(EP_ALERT_RULE, rule_info),
            )
            raise AlertsFetchError(err)

        return rules

    def _fetch_triggered_alerts(self, rule_ids, triggered):
        """Calls the alert search endpoint to retrieve a triggered alerts.

        Args:
            rule_ids (list): List of string rule IDs
            triggered (str): Timeframe for triggered param

        Raises:
            AlertsFetchError: Raised when the rule ID request fails for any reason

        Returns:
            list: Dict list of v2/alert/search response for provided rule IDs
        """
        self.log.debug("Get triggered alert ID's: Start")
        query = {
            'limit': _API_RESPONSE_LIMIT,
            'triggered': triggered,
        }
        if len(rule_ids) == 0:
            try:
                alerts = self.rf_client.make_paged_request(EP_ALERT_SEARCH, params=query)
            except (
                ConnectionError,
                ConnectTimeout,
                HTTPError,
                JSONDecodeError,
                ReadTimeout,
            ) as err:
                raise AlertsFetchError(err)
        elif len(rule_ids) > 0:
            alerts = []
            for rule in rule_ids:
                query['alertRule'] = rule
                try:
                    alerts_for_rule = self.rf_client.make_paged_request(
                        EP_ALERT_SEARCH,
                        params=query,
                    )
                    self.log.debug(
                        'For rule ID {} found {} alerts'.format(rule, len(alerts_for_rule)),
                    )
                    alerts += alerts_for_rule
                except (
                    ConnectionError,
                    ConnectTimeout,
                    HTTPError,
                    JSONDecodeError,
                    ReadTimeout,
                ) as err:
                    raise AlertsFetchError(err)

        self.log.debug('Found {} total alerts'.format(len(alerts)))

        return alerts

    def lookup_alert(self, alert_id):
        """Fetches a Recorded Future Alert based on the alert id.

        Args:
            alert_id (str): Alert ID to retrieve JSON for from RF API

        Raises:
            AlertsFetchError: Raised when the alert lookup failes for any reason

        Returns:
            dict: Alert details JSON
        """
        url = EP_ALERT_LOOKUP + alert_id
        try:
            response = self.rf_client.get_request(url)
        except (ConnectionError, ConnectTimeout, HTTPError, JSONDecodeError, ReadTimeout) as err:
            raise AlertsFetchError(err)

        try:
            response = response.json()
        except JSONDecodeError:
            self.log.debug('API response is:\n{}'.format(response.text))
            raise AlertsFetchError('Unable to parse API response as JSON')

        return response

    def _lookup_alerts(self, alert_ids):
        """Returns complete alert data for all alert IDs passed in.

        Args:
            alert_metadata (dict list): List of metadata for each alert to lookup from RF API

        Returns:
            list: List of alerts JSON
        """
        if len(alert_ids) == 0:
            self.log.warning('No alerts to lookup')
            return []

        alert_list = []
        total_num_of_alerts = len(alert_ids)

        self.log.info('Starting classic alert lookups. This might take a while...')
        for idx, alert_id in enumerate(alert_ids):
            self.log.info(
                'Lookup {0}/{1} -> Alert ID: {2}'.format(idx + 1, total_num_of_alerts, alert_id),
            )
            alert_json = self.lookup_alert(alert_id)
            alert_list.append(alert_json)

        self.log.info('Classic alert lookup completed')
        return alert_list

    def _filter_triggered_alerts(self, alerts_metadata, filter_id=None, fetch_status=None):
        """Filters v2/alert/search response using filter_id, fetch_status.

        Args:
            alerts_metadata (dict list): id, triggered dict list
            filter_id (str, optional): Alert ID to filter prior alerts. Defaults to None.
            fetch_status (str, optional): Status to filter from fetch. Defaults to None.
        """
        alerts_metadata.sort(key=lambda x: parser.parse(x['triggered']))
        filtered_by_id = alerts_metadata
        if filter_id is not None:
            filter_alert_list = list(filter(lambda x: x['id'] == filter_id, alerts_metadata))
            if len(filter_alert_list) == 0:
                self.log.warning('Filter ID {} not present in fetched alert IDs'.format(filter_id))
            else:
                filter_alert = filter_alert_list[0]
                filtered_by_id = list(
                    filter(
                        lambda x: parser.parse(x['triggered'])
                        > parser.parse(filter_alert['triggered']),
                        alerts_metadata,
                    ),
                )
                self.log.info(
                    "Filtered {} alerts using filter ID '{}'".format(
                        len(alerts_metadata) - len(filtered_by_id),
                        filter_id,
                    ),
                )
        if fetch_status is not None:
            filtered_by_id_and_status = list(
                filter(lambda x: x['review']['status'] == fetch_status, filtered_by_id),
            )
            self.log.info(
                "Filtered {} alerts not matching status '{}'".format(
                    len(filtered_by_id) - len(filtered_by_id_and_status),
                    fetch_status,
                ),
            )
            return filtered_by_id_and_status
        else:
            return filtered_by_id

    def fetch_alerts(self, filter_id=None, triggered=None, rule_names=None, fetch_status=None):
        """Gets raw alerts from RF API and parses by format using RFAlert.

        Args:
            filter_id (str, optional): If provided, remove all alerts up to and including this ID

        Returns:
            list: alert JSON list

        Raises:
            NoRulesFoundError, NoAlertsFoundError
        """
        self.log.info('Fetching alerts')

        if triggered is not None and not isinstance(triggered, str):
            raise TypeError('triggered must be a string')
        if rule_names is not None and not isinstance(rule_names, list):
            raise TypeError('rule_ids must be a list')
        if fetch_status is not None and not isinstance(fetch_status, str):
            raise TypeError('fetch_status must be a string')

        triggered = self._format_triggered(triggered or self.triggered)
        self.log.info('Triggered: {}'.format(triggered))

        rule_ids = (
            self._fetch_rule_ids(rule_names) if rule_names is not None else self._fetch_rule_ids()
        )
        fetch_status = fetch_status or self.fetch_status
        alerts_to_fetch = self._fetch_triggered_alerts(rule_ids, triggered)
        alerts_to_fetch = self._filter_triggered_alerts(
            alerts_to_fetch,
            filter_id=filter_id,
            fetch_status=fetch_status,
        )
        alerts_to_fetch = [alert['id'] for alert in alerts_to_fetch]
        raw_alerts = self._lookup_alerts(alerts_to_fetch)
        self.raw_alerts = raw_alerts  # alerts are saved for IDs to be used by get_fetched_ids()

        self.log.debug('Recorded Future {} alert format'.format(self.format))
        if self.format == 'raw':
            return raw_alerts
        else:
            parsed_alerts = []
            self.log.info('Parsing alerts. This might take a while...')
            for alert in raw_alerts:
                parsed_alerts += self.rf_alert.parse(alert=alert)
            return parsed_alerts

    def _format_triggered(self, triggered):
        if (
            not TimeHelpers.is_valid_time_range(triggered)
            and TimeHelpers.is_rel_time_valid(triggered)
            and not triggered.startswith('-')
        ):
            triggered = '-' + triggered
        return triggered

    def get_fetched_ids(self):
        """Returns fetched alert IDs from last call to fetch_alerts as list.

        Raises:
            AlertsParseError: Raised when alert IDs cannot be parsed

        Returns:
            list: alert IDs from last fetch
        """
        try:
            ids = [alert['data']['id'] for alert in self.raw_alerts]

        except KeyError as err:
            raise AlertsParseError('Cannot get IDs from alerts: KeyError: {}'.format(err)) from err

        except TypeError as err:
            raise AlertsParseError('Cannot get IDs from alerts: {}'.format(err)) from err

        return ids

    def set_alert_status(self, alert_ids=None, new_status=None) -> Optional[bool]:
        """Bulk set alert status.

        Args:
            alert_ids (list): List of alert IDs to update status for
            status (str): Status to update to for those alert IDs
        """
        if alert_ids is not None and not isinstance(alert_ids, list):
            raise TypeError('alert_ids must be a list')
        if new_status is not None and not isinstance(new_status, str):
            raise TypeError('new_status must be a string')
        if alert_ids is not None and new_status is None:
            raise ValueError('new_status must be provided if alert_ids is provided')
        if alert_ids is None and new_status is not None:
            raise ValueError('alert_ids must be provided if new_status is provided')
        do_update = True if alert_ids is not None else self.update_status
        new_status = new_status or self.new_status
        if not do_update:
            self.log.info('Update alert status on Recorded Future API disabled. Skipping')
            return False

        if alert_ids is None:
            try:
                alert_ids = self.get_fetched_ids()
            except AlertsParseError:
                alert_ids = []

        self.log.info('Set alert status to: {}'.format(new_status))
        if len(alert_ids) == 0:
            self.log.warning('No alerts to update')
            return False

        payload = [{'id': alert_id, 'status': new_status} for alert_id in alert_ids]

        payload = json.dumps(payload)
        self.update_alert(payload)

    def update_alert(self, payload) -> None:
        """Updates an alert's review content using the /v2/alert/update endpoint.

        Args:
            payload (dict): {
                                "id": "string",
                                "assignee": "string",
                                "status": "unassigned",
                                "note": "string"
                            }

        Raises:
            AlertUpdateError: When alert cannot be updated, provide descriptive
                              error for main
        """
        try:
            # TODO check if error array is empty, if not raise an exc
            self.rf_client.post_request(EP_ALERT_UPDATE, payload)
        except (HTTPError, ConnectionError, ConnectTimeout, ReadTimeout) as err:
            raise AlertUpdateError('Alert update failed: {}'.format(err)) from err

    def write_file(self, alerts, output_dir=None, filename=None) -> None:
        """Writes alerts payload to file.

        Args:
            alerts (list): Alerts from a call to fetch_alerts

        Raises:
            WriteFileError: Raised when file cannot be written
        """
        if alerts is None:
            self.log.warning("Empty 'alerts' argument passed to RFAlertMgr.write_file")
            return
        formatted_time = datetime.now().strftime('%Y-%m-%dT%H-%M-%S-%f')
        if filename is not None and not isinstance(filename, str):
            raise TypeError('filename must be a string')
        if output_dir is not None and not isinstance(output_dir, str):
            raise TypeError('output_dir must be a string')
        output_file = filename if filename else ALERTS_OUTPUT_FNAME + formatted_time + '.json'
        output_dir = output_dir or self.output_dir
        self.log.info('Writing alerts to file')
        try:
            dir_path = OSHelpers.mkdir(output_dir)
            full_path = os.path.join(dir_path, output_file)
            with open(full_path, 'w') as outfile:
                outfile.write(json.dumps(alerts, indent=2))
                self.log.info('  Wrote: {}'.format(full_path))
        except (OSError, WriteFileError) as err:
            raise WriteFileError(str(err))

    @property
    def triggered(self):
        """Getter for triggered."""
        triggered = self.config.settings.get(ALERTS_STANZA_NAME, {}).get(TRIGGERED_SETTING)
        if triggered is None:
            triggered = TRIGGERED_DEFAULT
            self.log.warning('triggered was not set, using default: {}'.format(triggered))
        return self._format_triggered(triggered)

    @triggered.setter
    def triggered(self, value) -> None:
        """Setter for triggered."""
        if not isinstance(value, str):
            raise TypeError('triggered must be a string')
        if not ALERTS_STANZA_PARAMS[STANZA_FIELDS_KEY][TRIGGERED_SETTING][VALIDATOR_KEY](value):
            raise ValueError('Invalid value for triggered: {}'.format(value))
        self.config.update_setting(ALERTS_STANZA_NAME, TRIGGERED_SETTING, value)

    @property
    def format(self):
        """Getter for format."""
        frmt = self.config.settings.get(ALERTS_STANZA_NAME, {}).get(FORMAT_SETTING)
        if frmt is None:
            frmt = FORMAT_DEFAULT
            self.log.warning('format was not set, using default: {}'.format(frmt))
        return frmt

    @format.setter
    def format(self, value) -> None:
        """Setter for format."""
        if not isinstance(value, str):
            raise TypeError('format must be a string')
        if not ALERTS_STANZA_PARAMS[STANZA_FIELDS_KEY][FORMAT_SETTING][VALIDATOR_KEY](value):
            raise ValueError('Invalid value for format: {}'.format(value))
        self.config.update_setting(ALERTS_STANZA_NAME, FORMAT_SETTING, value)

    @property
    def alert_rules(self):
        """Getter for alert_rules."""
        alert_rules = self.config.settings.get(ALERTS_STANZA_NAME, {}).get(ALERT_RULES_SETTING)
        if alert_rules is None:
            alert_rules = ALERT_RULES_DEFAULT
            self.log.warning('alert_rules was not set, using default: {}'.format(alert_rules))
        return alert_rules

    @alert_rules.setter
    def alert_rules(self, value) -> None:
        """Setter for alert_rules."""
        if not isinstance(value, list):
            raise TypeError('alert_rules must be a list of string names')
        if not ALERTS_STANZA_PARAMS[STANZA_FIELDS_KEY][ALERT_RULES_SETTING][VALIDATOR_KEY](value):
            raise ValueError('Invalid value for alert_rules: {}'.format(value))
        self.config.update_setting(ALERTS_STANZA_NAME, ALERT_RULES_SETTING, value)

    @property
    def update_status(self):
        """Getter for update_status."""
        update_status = self.config.settings.get(ALERTS_STANZA_NAME, {}).get(UPDATE_STATUS_SETTING)
        if update_status is None:
            update_status = UPDATE_STATUS_DEFAULT
            self.log.warning('update_status was not set, using default: {}'.format(update_status))
        return update_status

    @update_status.setter
    def update_status(self, value) -> None:
        """Setter for update_status."""
        if not isinstance(value, bool):
            raise TypeError('update_status must be a boolean')
        if not ALERTS_STANZA_PARAMS[STANZA_FIELDS_KEY][UPDATE_STATUS_SETTING][VALIDATOR_KEY](value):
            raise ValueError('Invalid value for update_status: {}'.format(value))
        self.config.update_setting(ALERTS_STANZA_NAME, UPDATE_STATUS_SETTING, value)

    @property
    def fetch_status(self):
        """Getter for fetch_status."""
        fetch_status = self.config.settings.get(ALERTS_STANZA_NAME, {}).get(FETCH_STATUS_SETTING)
        if fetch_status is None:
            fetch_status = FETCH_STATUS_DEFAULT
            self.log.warning('fetch_status was not set, using default: {}'.format(fetch_status))
        return fetch_status

    @fetch_status.setter
    def fetch_status(self, value) -> None:
        """Setter for fetch_status."""
        if not isinstance(value, str):
            raise TypeError('fetch_status must be a string')
        if not ALERTS_STANZA_PARAMS[STANZA_FIELDS_KEY][FETCH_STATUS_SETTING][VALIDATOR_KEY](value):
            raise ValueError('Invalid value for fetch_status: {}'.format(value))
        self.config.update_setting(ALERTS_STANZA_NAME, FETCH_STATUS_SETTING, value)

    @property
    def new_status(self):
        """Getter for new_status."""
        new_status = self.config.settings.get(ALERTS_STANZA_NAME, {}).get(NEW_STATUS_SETTING)
        if new_status is None:
            new_status = NEW_STATUS_DEFAULT
            self.log.warning('new_status was not set, using default: {}'.format(new_status))
        return new_status

    @new_status.setter
    def new_status(self, value) -> None:
        """Setter for new_status."""
        if not isinstance(value, str):
            raise TypeError('new_status must be a string')
        if not ALERTS_STANZA_PARAMS[STANZA_FIELDS_KEY][NEW_STATUS_SETTING][VALIDATOR_KEY](value):
            raise ValueError('Invalid value for new_status: {}'.format(value))
        self.config.update_setting(ALERTS_STANZA_NAME, NEW_STATUS_SETTING, value)

    @property
    def output_dir(self):
        """Getter for output_dir."""
        output_dir = self.config.settings.get(ALERTS_STANZA_NAME, {}).get(OUTPUT_DIR_SETTING)
        if output_dir is None:
            output_dir = DEFAULT_ALERTS_OUTPUT_DIR
            self.log.warning(
                '{} was not set, using default: {}'.format(OUTPUT_DIR_SETTING, output_dir),
            )
        return output_dir

    @output_dir.setter
    def output_dir(self, value) -> None:
        """Setter for output_dir."""
        if not isinstance(value, str):
            raise TypeError('output_dir must be a string')
        if not ALERTS_STANZA_PARAMS[STANZA_FIELDS_KEY][OUTPUT_DIR_SETTING][VALIDATOR_KEY](value):
            raise ValueError('Invalid value for {}: {}'.format(OUTPUT_DIR_SETTING, value))
        self.config.update_setting(ALERTS_STANZA_NAME, OUTPUT_DIR_SETTING, value)
