##################################### TERMS OF USE ###########################################
# The following code is provided for demonstration purpose only, and should not be used      #
# without independent verification. Recorded Future makes no representations or warranties,  #
# express, implied, statutory, or otherwise, regarding any aspect of this code or of the     #
# information it may retrieve, and provides it both strictly “as-is” and without assuming    #
# responsibility for any information it may retrieve. Recorded Future shall not be liable    #
# for, and you assume all risk of using, the foregoing. By using this code, Customer         #
# represents that it is solely responsible for having all necessary licenses, permissions,   #
# rights, and/or consents to connect to third party APIs, and that it is solely responsible  #
# for having all necessary licenses, permissions, rights, and/or consents to any data        #
# accessed from any third party API.                                                         #
##############################################################################################

import json
import logging
import time

from ..helpers.helpers import FormattingHelpers
from .constants import (
    ADDED_LEN,
    CREATED_FIELD,
    ID_FIELD,
    IS_READY_INCREMENT,
    NAME_FIELD,
    NAME_LEN,
    ORGANISATION_ID_FIELD,
    ORGANISATION_NAME_FIELD,
    STATUS_LEN,
    TYPE_FIELD,
    TYPE_LEN,
    UPDATED_FIELD,
)
from .errors import ListApiError, ListStateError


class RFList:
    """Recorded Future List class."""

    def __init__(self, list_mgr, info: dict):
        self.log = logging.getLogger(__name__)
        self.__list_mgr = list_mgr
        (
            self.__id,
            self.__name,
            self.__type,
            self.__created,
            self.__updated,
            self.__organisation_id,
            self.__organisation_name,
        ) = self._parse(info)

    def _parse(self, info):
        """Parse list info, expecting list create, info, or search response."""
        id_ = info.get(ID_FIELD, None)
        name = info.get(NAME_FIELD, None)
        type_ = info.get(TYPE_FIELD, None)
        created = info.get(CREATED_FIELD, None)
        last_updated = info.get(UPDATED_FIELD, None)
        org_id = info.get(ORGANISATION_ID_FIELD, None)
        org_name = info.get(ORGANISATION_NAME_FIELD, None)

        return (id_, name, type_, created, last_updated, org_id, org_name)

    def __str__(self):
        """String representation of the list.

        Returns:
            str: list data with standard info + entities
        """
        list_repr = (
            'id: '
            + self.__id
            + '\n'
            + 'name: '
            + self.__name
            + '\n'
            + 'type: '
            + self.__type
            + '\n'
            + 'created: '
            + self.__created
            + '\n'
            + 'last updated: '
            + self.__updated
            + '\n'
            + 'organisation id: '
            + self.__organisation_id
            + '\n'
            + 'organisation name: '
            + self.__organisation_name
            + '\n\n'
        )

        try:
            entities = self.entities()
        except ListApiError:
            return list_repr
        list_repr += '{} entities in list\n'.format(len(entities))

        if len(entities) > 0:
            list_repr += '\n'
            list_repr += FormattingHelpers.format_output(
                [TYPE_LEN, NAME_LEN, STATUS_LEN, ADDED_LEN],
                ['Type', 'Name', 'Status', 'Added'],
            )
            count = 0
            for entity in entities:
                list_repr += '\n'
                count += 1
                if count == 10:
                    list_repr += '...'
                    break
                list_repr += FormattingHelpers.format_output(
                    [TYPE_LEN, NAME_LEN, STATUS_LEN, ADDED_LEN],
                    [
                        entity['entity']['type'],
                        entity['entity']['name'],
                        entity['status'],
                        entity['added'],
                    ],
                )

        return list_repr

    def info(self):
        """Return list info from /list/{id}/info endpoint.

        If info is already cached, return cached info

        Raises:
            ListApiError: If list info is not available
                NOTE: RFList objects must be initialized with list info
                      and this exception should not be raised unless that changes

        Returns:
            dict: list info
        """
        if (
            self.__id is not None
            and self.__name is not None
            and self.__type is not None
            and self.__created is not None
            and self.__updated is not None
            and self.__organisation_id is not None
            and self.__organisation_name is not None
        ):
            return {
                'id': self.__id,
                'name': self.__name,
                'type': self.__type,
                'created': self.__created,
                'updated': self.__updated,
                'organisation_id': self.__organisation_id,
                'organisation_name': self.__organisation_name,
            }
        else:
            return self.__list_mgr._get_list_info(self.__id)

    def status(self):
        """Get list status.

        Raises:
            ListApiError: If list status is not available
            ListStateError: If list API response does not contain status

        Returns:
            str: list status
        """
        resp = self.__list_mgr._get_list_status(self.__id)
        status = resp.get('status', None)

        if status is None:
            self.log.debug('List status is None')
            self.log.debug('API response is: {}'.format(resp))
            raise ListStateError('List status is not available')

        return status

    def is_ready(self):
        """Check if list is ready.

        Raises:
            ListApiError: If list status is not available
            ListStateError: If list API response does not contain status

        Returns:
            bool: True if list is ready, False otherwise
        """
        status = self.status()

        return status == 'ready'

    def register_callback(self, func, *args, **kwargs):
        """Register a function for callback when list status is 'ready'.

        Logs a wait message at INFO level every 10 seconds

        Args:
            func (function): Callback function to register
            *args (list): Arguments to pass to callback function
            **kwargs (dict): Keyword arguments to pass to callback function

        Raises:
            ListApiError: If list status is not available
            ListStateError: If list API response does not contain status

        Returns:
            any: return value of func
        """
        seconds = -1
        while not self.is_ready():
            seconds += IS_READY_INCREMENT
            if seconds % 10 == 0:
                self.log.info('Waiting for list ready status')
            time.sleep(IS_READY_INCREMENT)
        return func(*args, **kwargs)

    def size(self):
        """Get the size of the list.

        Raises:
            ListApiError: If list size is not available
            ListStateError: If list API response does not contain size

        Returns:
            int: size of the list
        """
        status = self.__list_mgr._get_list_status(self.__id)
        size = status.get('size', None)

        if size is None:
            self.log.debug('List size is None')
            self.log.debug('API response was: {}'.format(json.dumps(status, indent=2)))
            raise ListStateError('List size is not available')
        else:
            return size

    def entities(self):
        """Get list entities.

        Raises:
            ListApiError: If list entities are not available

        Returns:
            list: list of {id, name, type} dicts of list entities
        """
        return self.__list_mgr._get_list_entities(self.__id)

    def add(self, entity, context=None):
        """Add entity to a list.

        Raises:
            ListApiError: If entity cannot be added to list

        Args:
            entity (str or tuple or list): ID or name + type Sequence of entity to add
            context (dict, optional): context object for entity. Defaults to None.

        Returns:
            dict: list/{id}/entity/add JSON response
        """
        return self.__list_mgr._add_to_list(self.__id, entity, context)

    def remove(self, entity):
        """Remove entity from list.

        Raises:
            ListApiError: If entity cannot be removed from list

        Args:
            entity (str or tuple or list): ID or name + type Sequence of entity to remove

        Returns:
            dict: list/{id}/entity/remove JSON response
        """
        return self.__list_mgr._remove_from_list(self.__id, entity)

    def json(self):
        """Return JSON representation of the list.

        Raises:
            ListApiError: If list info or entities are not available

        Returns:
            JSON: list data with standard info + entities
        """
        info = self.info()
        entities = self.entities()
        info['entities'] = entities

        return info

    def from_file(self, open_file, operation):
        """Perform list operation on entities from file.

        Use psengine ADD and REMOVE constants as operations

        Raises:
            ValueError: if operation is not ADD or REMOVE
                        if open_file is not CSV with correct headers

        Args:
            open_file (file): text file object in read mode, CSV of ID or Name, Type
            operation (str): ADD or REMOVE operation

        Returns:
            dict: {added/removed, unchanged, error} dict with counts
        """
        return self.__list_mgr._from_file(open_file, self.__id, operation)

    @property
    def id(self):
        """Get list id.

        Returns:
            str: list id
        """
        return self.__id

    @property
    def name(self):
        """Get list name.

        Returns:
            str: list name
        """
        return self.__name

    @property
    def type(self):
        """Get list type.

        Returns:
            str: list type
        """
        return self.__type

    @property
    def created(self):
        """Get list creation date.

        Returns:
            str: list creation date
        """
        return self.__created

    @property
    def updated(self):
        """Get list last updated date.

        Returns:
            str: list last updated date
        """
        return self.__updated

    @property
    def organisation_id(self):
        """Get list organisation id.

        Returns:
            str: list organisation id
        """
        return self.__organisation_id

    @property
    def organisation_name(self):
        """Get list organisation name.

        Returns:
            str: list organisation name
        """
        return self.__organisation_name
