##################################### TERMS OF USE ###########################################
# The following code is provided for demonstration purpose only, and should not be used      #
# without independent verification. Recorded Future makes no representations or warranties,  #
# express, implied, statutory, or otherwise, regarding any aspect of this code or of the     #
# information it may retrieve, and provides it both strictly “as-is” and without assuming    #
# responsibility for any information it may retrieve. Recorded Future shall not be liable    #
# for, and you assume all risk of using, the foregoing. By using this code, Customer         #
# represents that it is solely responsible for having all necessary licenses, permissions,   #
# rights, and/or consents to connect to third party APIs, and that it is solely responsible  #
# for having all necessary licenses, permissions, rights, and/or consents to any data        #
# accessed from any third party API.                                                         #
##############################################################################################

import csv
import io
import json
import logging
from collections.abc import Sequence
from json import JSONDecodeError

from requests.exceptions import ConnectionError, ConnectTimeout, HTTPError, ReadTimeout

from ..config.config import Config
from ..endpoints import EP_CREATE_LIST, EP_LIST, EP_SEARCH_LIST
from ..entity_match import MatchError, RFMatchMgr
from ..rf_client import RFClient
from .constants import ADD, REMOVE
from .errors import ListApiError
from .rf_list import RFList


class RFListMgr:
    """Manages requests for Recorded Future List API."""

    def __init__(self, cfg: Config = None, rf_token: str = None):
        """Initialize RFListMgr.

        Args:
            cfg (Config, optional): Config object. Inits empty Config if none provided.
            rf_token (str, optional): Recorded Future API token. Defaults to None.
        """
        self.log = logging.getLogger(__name__)
        self.config = cfg if cfg is not None else Config()
        self.rf_client = (
            RFClient(self.config, rf_token=rf_token)
            if rf_token is not None
            else RFClient(self.config)
        )
        self.match_mgr = RFMatchMgr(self.config)

    def _resolve_list_id(self, list_):
        """Resolves a list name to a list ID.

        Args:
            list_ (str, tuple): list string ID or tuple of (name, type)

        Raises:
            MatchError: raised when a list name matches none or multiple entities
            ValueError: raised when list_ is not a string or tuple

        Returns:
            str: list ID
        """
        if isinstance(list_, str):
            resolved_id = list_
        elif isinstance(list_, Sequence) and len(list_) == 2:
            list_name, list_type = list_
            matches = self.search_lists(list_name, list_type)
            len_matches = len(matches)
            if len_matches == 0:
                message = "No match found for string '{0}'".format(list_name)
                raise MatchError(message)
            elif len_matches > 1:
                exact_count = 0
                resolved_id = None
                for match in matches:
                    if match.name == list_name:
                        resolved_id = match.id
                        exact_count += 1
                if (not resolved_id) or exact_count > 1:
                    message = "Multiple matches found for string '{0}'".format(list_name)
                    raise MatchError(message)
            else:
                resolved_id = matches[0].id
        else:
            raise ValueError('Must specify either an ID or name and type')

        return resolved_id

    def _get_list_str(self, entity):
        """Returns either entity name or ID as string, for logging.

        Args:
            entity (str, tuple): entity string ID or tuple of (name, type)

        Returns:
            str: entity descriptor
        """
        if isinstance(entity, str):
            return entity
        if isinstance(entity, Sequence) and len(entity) == 2:
            return entity[0]

    def get_list_by_id(self, list_id):
        """Gets a list by its ID.

        Args:
            list_id (str): string list ID

        Raises:
            TypeError: if list_id is not a string
            ListApiError: if list API call fails, usually when a bad ID is supplied

        Returns:
            RFList: RFList object for list ID
        """
        self.log.debug('Getting list by ID: {}'.format(list_id))
        if not isinstance(list_id, str):
            raise TypeError('get_list_by_id argument must be a string')
        list_info = self._get_list_info(list_id)
        list_ = RFList(self, list_info)

        self.log.debug('Found list:\n{}'.format(list_.name))
        return list_

    def create_list(self, list_name: str, list_type='entity'):
        """Create list.

        Raises:
            ListApiError: if list API call fails

        Args:
            list_name (str): list name to use for new list
            list_type (str, optional): list type. Defaults to "entity".

            Supported list types are available on the support page for List API:
            https://support.recordedfuture.com/hc/en-us/articles/360058691913-List-API

        Returns:
            RFList: RFList object
        """
        self.log.debug("Creating list '{0}'".format(list_name))
        request_body = {'name': list_name, 'type': list_type}
        try:
            response = self.rf_client.post_request(EP_CREATE_LIST, json.dumps(request_body))
        except (HTTPError, ConnectionError, ReadTimeout, ConnectTimeout) as err:
            raise ListApiError(str(err))
        try:
            list_create_data = response.json()
        except JSONDecodeError as err:
            raise ListApiError('Malformed API response: ' + str(err))
        self.log.debug("List '{}' created".format(list_name))
        self.log.debug('  ID: {}'.format(list_create_data['id']))
        self.log.debug('  Type: {}'.format(list_create_data['type']))

        return RFList(self, list_create_data)

    def search_lists(self, list_name, list_type=None):
        """Search lists.

        Raises:
            ListApiError: if list API call fails

        Args:
            list_name (str): list name to search
            list_type (str, optional): list type. Defaults to None, ignored when None.

        Returns:
            dict: JSON response from list/search
        """
        request_body = {'name': list_name}
        if list_type is not None:
            request_body['type'] = list_type
            self.log.debug(
                "Searching list API for list '{}' with type '{}'".format(list_name, list_type),
            )
        else:
            self.log.debug("Searching list API for list '{}'".format(list_name))
            self.log.debug('  Use list_type parameter to narrow results')
        try:
            response = self.rf_client.post_request(EP_SEARCH_LIST, json.dumps(request_body))
        except (HTTPError, ConnectionError, ReadTimeout, ConnectTimeout) as err:
            raise ListApiError(str(err))
        try:
            list_search_data = response.json()
        except JSONDecodeError as err:
            raise ListApiError('Malformed API response: ' + str(err))
        term = 'lists' if len(list_search_data) != 1 else 'list'
        self.log.debug('Found {} matching {}'.format(len(list_search_data), term))

        return [RFList(self, list_) for list_ in list_search_data]

    def _get_list_info(self, list_):
        """Get basic information about a list.

        Raises:
            ListApiError: if list API call fails

        Args:
            list_ (str or tuple or list): ID or name + type Sequence of list to get info for

        Returns:
            dict: list/{id}/info JSON response
        """
        self.log.debug('Getting list info for {}'.format(self._get_list_str(list_)))
        resolved_id = self._resolve_list_id(list_)
        url = EP_LIST + '/{id}/info'.format(id=resolved_id)
        try:
            response = self.rf_client.get_request(url)
        except (HTTPError, ConnectionError, ReadTimeout, ConnectTimeout) as err:
            raise ListApiError(str(err))
        try:
            list_info_data = response.json()
        except JSONDecodeError as err:
            raise ListApiError('Malformed API response: ' + str(err))
        self.log.debug('  ID: {}'.format(list_info_data['id']))
        self.log.debug('  Type: {}'.format(list_info_data['type']))
        self.log.debug('  Created: {}'.format(list_info_data['created']))
        self.log.debug('  Updated: {}'.format(list_info_data['updated']))
        return list_info_data

    def _get_list_status(self, list_):
        """Get status information about a list.

        Raises:
            ListApiError: if list API call fails

        Args:
            list_ (str): ID or name + type Sequence of list to get status for

        Returns:
            dict: list/{id}/status JSON response
        """
        self.log.debug("Getting list status for '{}'".format(self._get_list_str(list_)))
        resolved_id = self._resolve_list_id(list_)
        url = EP_LIST + '/{id}/status'.format(id=resolved_id)
        try:
            response = self.rf_client.get_request(url)
        except (HTTPError, ConnectionError, ReadTimeout, ConnectTimeout) as err:
            raise ListApiError(str(err))
        try:
            list_status_data = response.json()
        except JSONDecodeError as err:
            raise ListApiError('Malformed API response: ' + str(err))
        return list_status_data

    def _get_list_entities(self, list_):
        """Get entities for a list.

        Raises:
            ListApiError: if list API call fails

        Args:
            list_ (str or tuple or list): ID or name + type Sequence of list to get entities for

        Returns:
            dict: list/{id}/entities JSON response
        """
        resolved_id = self._resolve_list_id(list_)
        url = EP_LIST + '/{id}/entities'.format(id=resolved_id)
        try:
            response = self.rf_client.get_request(url)
        except (HTTPError, ConnectionError, ReadTimeout, ConnectTimeout) as err:
            raise ListApiError(str(err))
        try:
            list_entities_data = response.json()
        except JSONDecodeError as err:
            raise ListApiError('Malformed API response: ' + str(err))
        return list_entities_data

    def _add_to_list(self, list_, entity, context=None):
        """Add entity to a list.

        Raises:
            MatchError: if entity in (name, type) form does not match exactly 1 entity
            MatchError: if list in (name, type) form does not match exactly 1 list
            ListApiError: if list API call fails

        Args:
            list_ (str): ID or name + type Sequence of list to add entity to
            entity (str or tuple or list): ID or name + type Sequence of entity to add
            context (dict, optional): context object for entity. Defaults to None.

        Returns:
            dict: list/{id}/entity/add JSON response
        """
        self.log.debug("Adding entity '{}' to list '{}'".format(entity, self._get_list_str(list_)))
        if isinstance(entity, str):
            resolved_entity_id = entity
        elif (isinstance(entity, (list, tuple))) and len(entity) == 2:
            resolved_entity_id = self.match_mgr.resolve_entity_id(entity[0], entity_type=entity[1])
        else:
            raise TypeError('Entity must be a string ID or tuple of (name, type)')
        resolved_list_id = self._resolve_list_id(list_)
        url = EP_LIST + '/{id}/entity/add'.format(id=resolved_list_id)
        request_body = {'entity': {'id': resolved_entity_id}}
        if context is not None:
            request_body['context'] = context
        try:
            response = self.rf_client.post_request(url, json.dumps(request_body))
        except (HTTPError, ConnectionError, ReadTimeout, ConnectTimeout) as err:
            raise ListApiError(str(err))
        try:
            list_add_data = response.json()
        except JSONDecodeError as err:
            raise ListApiError('Malformed API response: ' + str(err))
        return list_add_data

    def _remove_from_list(self, list_, entity):
        """Remove entity from list.

        Args:
            list_ (str or tuple or list): ID or name + type Sequence of list to remove entity from
            entity (str or tuple or list): ID or name + type Sequence of entity to remove

        Raises:
            MatchError: if entity in (name, type) form does not match exactly 1 entity
            MatchError: if list in (name, type) form does not match exactly 1 list
            ListApiError: if list API call fails

        Returns:
            dict: list/{id}/entity/remove JSON response
        """
        self.log.debug(
            "Removing entity '{}' from list '{}'".format(entity, self._get_list_str(list_)),
        )
        if isinstance(entity, str):
            resolved_entity_id = entity
        elif (isinstance(entity, (list, tuple))) and len(entity) == 2:
            resolved_entity_id = self.match_mgr.resolve_entity_id(entity[0], entity_type=entity[1])
        else:
            raise TypeError('Entity must be a string ID or name, type list or tuple')
        resolved_list_id = self._resolve_list_id(list_)
        url = EP_LIST + '/{id}/entity/remove'.format(id=resolved_list_id)
        request_body = {'entity': {'id': resolved_entity_id}}
        try:
            response = self.rf_client.post_request(url, json.dumps(request_body))
        except (HTTPError, ConnectionError, ReadTimeout, ConnectTimeout) as err:
            raise ListApiError(str(err))
        try:
            list_remove_data = response.json()
        except JSONDecodeError as err:
            raise ListApiError('Malformed API response: ' + str(err))
        return list_remove_data

    def _bulk_add(self, list_, entities):
        """Bulk add entities to a list.

        Args:
            list_ (str or tuple or list): ID or name + type Sequence of list to add entities to
            entities (list): list of entity string IDs or entity name + type Sequences to add

        Returns:
            dict: results JSON with added, unchanged, error keys
        """
        result = {'added': [], 'unchanged': [], 'error': []}
        total = len(entities)
        added = 0
        step = 10
        for entity in entities:
            entity_id = entity
            try:
                if isinstance(entity, str):
                    entity_id = entity
                elif (isinstance(entity, (list, tuple))) and len(entity) == 2:
                    entity_id = self.match_mgr.resolve_entity_id(entity[0], entity_type=entity[1])
                else:
                    raise TypeError('Entity must be a string ID or name, type list or tuple')
                response = self._add_to_list(list_, entity)
                if response['result'] == 'added':
                    result['added'].append(entity_id)
                elif response['result'] == 'unchanged':
                    result['unchanged'].append(entity_id)
            except (TypeError, MatchError, ListApiError) as err:
                result['error'].append({'message': str(err), 'id': entity_id})
            added += 1
            if (added / total) * 100 >= step:
                self.log.info('Added {}% of entities'.format(step))
                step += 10
        return result

    def _bulk_remove(self, list_, entities):
        """Bulk remove entities from list.

        Args:
            list_ (str or tupel or list): ID or name + type Sequence of list to remove entities from
            entities (list): list of entity string IDs or entity name + type Sequences to remove

        Returns:
            dict: results JSON with removed, unchanged, error keys
        """
        result = {'removed': [], 'unchanged': [], 'error': []}
        total = len(entities)
        removed = 0
        step = 10
        for entity in entities:
            entity_id = entity
            try:
                if isinstance(entity, str):
                    entity_id = entity
                elif (isinstance(entity, (list, tuple))) and len(entity) == 2:
                    entity_id = self.match_mgr.resolve_entity_id(entity[0], entity_type=entity[1])
                else:
                    raise TypeError('Entity must be a string ID or tuple of (name, type)')
                response = self._remove_from_list(list_, entity)
                if response['result'] == 'removed':
                    result['removed'].append(entity_id)
                elif response['result'] == 'unchanged':
                    result['unchanged'].append(entity_id)
            except (TypeError, MatchError, ListApiError) as err:
                result['error'].append({'message': str(err), 'id': entity_id})
            removed += 1
            if (removed / total) * 100 >= step:
                self.log.info('Removed {}% of entities'.format(step))
                step += 10
        return result

    def _get_entities_from_file(self, open_file):
        """Returns entities as a list from a CSV file for bulk add/remove operations.

        CSV must have either ID column or Name, Type columns

        Args:
            open_file (file-like object): file-like object for CSV file

        Raises:
            ValueError: if open_file is not a CSV file in read-mode with valid columns

        Returns:
            list: list of string IDs or (name, type) tuples
        """
        reader = csv.DictReader(open_file)
        fieldnames = reader.fieldnames
        if len(fieldnames) > 2:
            raise ValueError('Invalid CSV: Supply only ID or Name, Type columns in CSV file')
        if len(fieldnames) == 1:
            if fieldnames[0] != 'ID':
                raise ValueError("Invalid CSV: One column CSV must have column named 'ID'")
            entities = [row['ID'] for row in reader]
        elif len(fieldnames) == 2:
            if fieldnames[0] != 'Name' or fieldnames[1] != 'Type':
                raise ValueError(
                    "Invalid CSV: Two column CSV must have columns named 'Name' and 'Type'",
                )
            entities = [(row['Name'], row['Type']) for row in reader]

        return entities

    def _from_file(self, open_file, list_, op_type):
        """Adds to list or removes from list entities listed in a CSV file.

        Args:
            open_file (file-like object): file-like object for CSV file
            op_type (str): string operation, choices: add, remove

        Raises:
            ValueError: if op_type not in ['add', 'remove']
                        if open_file is not a CSV file in read-mode with valid columns

        Returns:
            list: list of {'added', 'removed', 'unchanged', 'error'} dicts
                  each key contains a list of entity IDs
        """
        if not isinstance(open_file, io.TextIOBase) or open_file.closed:
            raise ValueError("Please supply an open file object in read mode ('r')")
        self.log.debug(
            'Performing {} operation on list {}'.format(op_type, self._get_list_str(list_)),
        )
        choices = [ADD, REMOVE]
        if op_type not in choices:
            raise ValueError("Invalid operation type '{}'".format(op_type))
        entities = self._get_entities_from_file(open_file)
        if op_type == ADD:
            result = self._bulk_add(list_, entities)
        elif op_type == REMOVE:
            result = self._bulk_remove(list_, entities)

        self.log.info('Bulk operation complete')
        for key in result:
            self.log.info('{} entities with status {}'.format(len(result[key]), key))

        return result
