##################################### TERMS OF USE ###########################################
# The following code is provided for demonstration purpose only, and should not be used      #
# without independent verification. Recorded Future makes no representations or warranties,  #
# express, implied, statutory, or otherwise, regarding any aspect of this code or of the     #
# information it may retrieve, and provides it both strictly “as-is” and without assuming    #
# responsibility for any information it may retrieve. Recorded Future shall not be liable    #
# for, and you assume all risk of using, the foregoing. By using this code, Customer         #
# represents that it is solely responsible for having all necessary licenses, permissions,   #
# rights, and/or consents to connect to third party APIs, and that it is solely responsible  #
# for having all necessary licenses, permissions, rights, and/or consents to any data        #
# accessed from any third party API.                                                         #
##############################################################################################

import logging
import logging.config
import os
import sys

from ..exceptions import WriteFileError
from ..helpers.helpers import OSHelpers
from .errors import LoggingError

LOGGER_NAME = 'psengine'

DEFAULT_PSENGINE_OUTPUT = os.path.join('logs', 'psengine_recfut.log')
DEFAULT_ROOT_OUTPUT = os.path.join('logs', 'root_recfut.log')

MAX_BYTES = 20480 * 1024
BACKUP_COUNT = 5


class RFLogger:
    """Sets up logging and gives access to its functions."""

    def __init__(
        self,
        output: str = DEFAULT_PSENGINE_OUTPUT,
        root_output: str = DEFAULT_ROOT_OUTPUT,
        level=logging.INFO,
        propagate: bool = True,
        to_file=True,
        to_console=True,
        console_is_root=True,
    ):
        if to_file is False and to_console is False:
            raise ValueError('At least one of to_file or to_console must be set to True')

        if not isinstance(level, str) and not isinstance(level, int):
            raise TypeError('level must be a string or int')
        if isinstance(level, str):
            level = level.upper()
            if level not in ['NOTSET', 'DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL']:
                raise ValueError(
                    'level must be one of: NOTSET, DEBUG, INFO, WARNING, ERROR, CRITICAL',
                )
        if isinstance(level, int) and level not in [
            logging.NOTSET,
            logging.DEBUG,
            logging.INFO,
            logging.WARNING,
            logging.ERROR,
            logging.CRITICAL,
        ]:
            raise ValueError('level must be one of: NOTSET, DEBUG, INFO, WARNING, ERROR, CRITICAL')

        # Setup logging handlers
        self.logger = logging.getLogger(LOGGER_NAME)
        root_logger = logging.getLogger()

        if to_file:
            psengine_file_handler = self._create_file_hander(output)
            root_file_handler = self._create_file_hander(root_output)
            self.logger.addHandler(psengine_file_handler)
            root_logger.addHandler(root_file_handler)

        if to_console:
            console_handler = self._create_console_handler()
            if console_is_root:
                root_logger.addHandler(console_handler)
            else:
                self.logger.addHandler(console_handler)

        # If false then logging messages are not passed to the handlers of ancestor loggers
        self.logger.propagate = propagate

        logging.captureWarnings(True)

        sys.excepthook = self._log_uncaught_exception

        self.logger.setLevel(level=level)
        self.logger.debug('Logger initialized')

    def _create_file_hander(self, output):
        log_filename = self._setup_output(output)

        file_handler = logging.handlers.RotatingFileHandler(
            log_filename,
            maxBytes=MAX_BYTES,
            backupCount=BACKUP_COUNT,
        )
        formatter_file = logging.Formatter(
            fmt=(
                '%(asctime)s,%(msecs)03d [%(threadName)s] %(levelname)s [%(module)s] '
                '%(funcName)s:%(lineno)d - %(message)s'
            ),
            datefmt='%Y-%m-%d %H:%M:%S',
        )
        file_handler.setFormatter(formatter_file)

        return file_handler

    def _create_console_handler(self):
        console_handler = logging.StreamHandler()
        formatter_console = logging.Formatter(
            fmt='%(asctime)s,%(msecs)03d %(levelname)s [%(module)s] - %(message)s',
            datefmt='%Y-%m-%d %H:%M:%S',
        )
        console_handler.setFormatter(formatter_console)

        return console_handler

    def _setup_output(self, output):
        """Confirms path is valid, returns cwd path and log cfg fullpath.

        Args:
            output (str): logging output path

        Raises:
            LoggingError: Raised when logging path does not exist

        Returns:
            tuple: cwd path, log path tuple
        """
        if os.path.isabs(output):
            dir_name = os.path.dirname(output)
            full_path = output
        else:
            main_dir = sys.path[0]
            full_path = os.path.join(main_dir, output)
            dir_name = os.path.dirname(full_path)

        try:
            OSHelpers.mkdir(dir_name)
        except (WriteFileError, ValueError) as err:
            raise LoggingError(message='Unable to create logging directory. Cause: {}'.format(err))

        return full_path

    def _log_uncaught_exception(self, exc_type, exc_value, exc_traceback) -> None:
        if issubclass(exc_type, KeyboardInterrupt):
            sys.__excepthook__(exc_type, exc_value, exc_traceback)
            return

        self.logger.critical(
            'An unexpected error has occurred:\n========================\n',
            exc_info=(exc_type, exc_value, exc_traceback),
        )

    def get_logger(self) -> logging.Logger:
        """Returns self.logger object."""
        return self.logger
