##################################### TERMS OF USE ###########################################
# The following code is provided for demonstration purpose only, and should not be used      #
# without independent verification. Recorded Future makes no representations or warranties,  #
# express, implied, statutory, or otherwise, regarding any aspect of this code or of the     #
# information it may retrieve, and provides it both strictly “as-is” and without assuming    #
# responsibility for any information it may retrieve. Recorded Future shall not be liable    #
# for, and you assume all risk of using, the foregoing. By using this code, Customer         #
# represents that it is solely responsible for having all necessary licenses, permissions,   #
# rights, and/or consents to connect to third party APIs, and that it is solely responsible  #
# for having all necessary licenses, permissions, rights, and/or consents to any data        #
# accessed from any third party API.                                                         #
##############################################################################################

import json
import logging
import warnings
from copy import deepcopy

from psengine.playbook_alerts.constants import DEPRECATION_MSG

from .errors import PlaybookAlertError


class RFBasePlaybookAlert:
    """Base class for a Recorded Future Playbook Alert."""

    def __init__(self, category: str, raw_alert: dict):
        self.log = logging.getLogger(__name__)
        self._category = category.lower()
        self._raw = raw_alert

        if 'data' not in self.raw:
            self.log.debug('Raw payload: {}'.format(json.dumps(self._raw, indent=2)))
            raise PlaybookAlertError("Invalid alert format. 'data' object not found")

        self._data = self.raw.get('data')

    def __str__(self) -> str:
        """Summarized string representation of the alert.

        Returns:
            str: playbook alert data
        """
        return 'Playbook Alert ID: {}, Created: {}, Category: {}, Lookup Status: {}'.format(
            self.id,
            self.created,
            self.category,
            self.lookup_status,
        )

    def json(self) -> dict:
        """Return JSON representation of the alert.

        Returns:
            dict: alert data or NONE if no raw alert data is available
        """
        return self.raw

    # Generic PA fields

    @property
    def raw(self) -> dict:
        """Get raw alert data.

        Returns:
            dict: raw playbook alert
        """
        return deepcopy(self._raw)

    @property
    def data(self) -> dict:
        """Get the alert data.

        Returns:
            dict: raw playbook alert data block
        """
        return deepcopy(self._data)

    @property
    def id(self) -> str:
        """Get playbook alert id.

        Returns:
            str: playbook alert id
        """
        return self.data.get('playbook_alert_id', None)

    @property
    def category(self) -> str:
        """Get playbook alert category.

        Returns:
            str: playbook alert category
        """
        return self._category

    @property
    def lookup_status(self) -> dict:
        """Get playbook alert lookup status.

        Returns:
            dict: Alert lookup status or {} if not found in the raw payload
        """
        return self.raw.get('status', {})

    # Generic panels that all PAs have

    @property
    def panel_status(self) -> dict:
        """Get playbook alert panel_status.

        Returns:
            dict: Alert panel_status or {} if not found in the raw payload
        """
        return self.data.get('panel_status', {})

    @property
    def panel_evidence_summary(self) -> dict:
        """Get playbook alert panel_evidence_summary.

        Returns:
            dict: Alert panel_evidence_summary or {} if not found in the raw payload
        """
        return self.data.get('panel_evidence_summary', {})

    @property
    def panel_log_v2(self) -> list:
        """Get playbook alert panel_log_v2.

        Returns:
            list: Alert panel_log_v2 or [] if not found in the raw payload
        """
        return self.data.get('panel_log_v2', [])

    # Fields from the alert payload

    @property
    def alert_status(self) -> str:
        """Get playbook alert status.

        Returns:
            str: Alert status or None if not found in the raw payload
        """
        return self.panel_status.get('status', None)

    @property
    def priority(self) -> str:
        """Get playbook alert priority.

        Returns:
            str: Alert priority or None if not found in the raw payload
        """
        return self.panel_status.get('priority', None)

    @property
    def assignee_name(self) -> str:
        """Get playbook alert assignee name.

        Returns:
            str: Alert assignee name or None if not found in the raw payload
        """
        return self.panel_status.get('assignee_name', None)

    @property
    def assignee_id(self) -> str:
        """Get playbook alert assignee ID.

        Returns:
            str: Alert assignee ID or None if not found in the raw payload
        """
        return self.panel_status.get('assignee_id', None)

    @property
    def created(self):
        """Get playbook alert created timestamp.

        Returns:
            str: Alert created timestamp or None if not found in the raw payload
        """
        return self.panel_status.get('created', None)

    @property
    def updated(self) -> str:
        """Get playbook alert updated timestamp.

        Returns:
            str: Alert updated timestamp or None if not found in the raw payload
        """
        return self.panel_status.get('updated', None)

    @property
    def rule_id(self) -> str:
        """Get playbook alert rule ID.

        Returns:
            str: Alert rule ID or None if not found in the raw payload
        """
        return self.panel_status.get('case_rule_id', None)

    @property
    def rule_name(self) -> str:
        """Get playbook alert rule name.

        Returns:
            str: Alert rule name or None if not found in the raw payload
        """
        return self.panel_status.get('case_rule_label', None)

    @property
    def owner_id(self) -> str:
        """Get playbook alert owner ID.

        Returns:
            str: Alert owner ID or None if not found in the raw payload
        """
        return self.panel_status.get('owner_id', None)

    @property
    def owner_name(self) -> str:
        """Get playbook alert owner name. In this case its always the organisation name
        that "owns" the alert. So in case of a multi-org account, it will be the actual org
        that the alert belongs to.

        Returns:
            str: Alert owner name or None if not found in the raw payload
        """
        return self.panel_status.get('owner_name', None)

    @property
    def organisation_name(self) -> str:
        """Get playbook alert organisation name. This is the enterprise name.

        Returns:
            str: Alert organisation name or None if not found in the raw payload
        """
        return self.panel_status.get('organisation_name', None)

    @property
    def owner_organisations(self) -> list:
        """Get playbook owner organisations. Each element contains organisation uhash and name.

        Returns:
            list: Organisation uhashes and names or [] if not found in the raw payload
        """
        return self.panel_status.get('owner_organisation_details', {}).get('organisations', [])

    @property
    def owner_enterprise_id(self) -> str:
        """Get playbook alert owner enterprise ID.

        Returns:
            str: Enterprise uhash or None if not found in the raw payload
        """
        return self.panel_status.get('owner_organisation_details', {}).get('enterprise_id', None)

    @property
    def owner_enterprise_name(self) -> str:
        """Get playbook alert owner enterprise name.

        Returns:
            str: Enterprise name or None if not found in the raw payload
        """
        return self.panel_status.get('owner_organisation_details', {}).get('enterprise_name', None)

    @property
    def entity_id(self) -> str:
        """Get playbook alert entity ID.

        Returns:
            str: Alert entity ID or None if not found in the raw payload
        """
        return self.panel_status.get('entity_id', None)

    @property
    def entity_name(self) -> str:
        """Get playbook alert entity name.

        Returns:
            str: Alert entity or None if not found in the raw payload
        """
        return self.panel_status.get('entity_name', None)

    @property
    def entity_criticality(self) -> str:
        """Get playbook alert entity criticality.

        Returns:
            int: Alert entity criticality or None if not found in the raw payload
        """
        return self.panel_status.get('entity_criticality', None)

    @property
    def entity_risk_score(self) -> int:
        """Get playbook alert entity risk score.

        Returns:
            int: Alert entity risk score or None if not found in the raw payload
        """
        return self.panel_status.get('risk_score', None)

    @property
    def targets(self) -> list:
        """Get playbook alert targets list. According to the API team this is meant to be
        a generic list that is used to store info about the alert in the generic panel_status.

        Returns:
            list: Alert targets or [] if not found in the raw payload
        """
        return self.panel_status.get('targets', [])

    @property
    def log_v2_changes(self) -> list:
        """List of all the changes in panel log v2"""
        return [
            changes
            for panel in self.panel_log_v2
            if panel.get('changes')
            for changes in panel['changes']
        ]

    @property
    def log_v2_assignee_change(self) -> list:
        """Assignee change"""
        return [changes for changes in self.log_v2_changes if changes['type'] == 'assignee_change']

    @property
    def log_v2_status_change(self) -> list:
        """Status change"""
        return [changes for changes in self.log_v2_changes if changes['type'] == 'status_change']

    @property
    def log_v2_title_change(self) -> list:
        """Title change"""
        return [changes for changes in self.log_v2_changes if changes['type'] == 'title_change']

    @property
    def log_v2_reopen_strategy_change(self) -> list:
        """Reopen Strategy change"""
        return [
            changes
            for changes in self.log_v2_changes
            if changes['type'] == 'reopen_strategy_change'
        ]

    @property
    def log_v2_related_entities_change(self) -> list:
        """Related Entities change"""
        return [
            changes
            for changes in self.log_v2_changes
            if changes['type'] == 'related_entities_change'
        ]

    @property
    def log_v2_entities_change(self) -> list:
        """Entities change"""
        return [changes for changes in self.log_v2_changes if changes['type'] == 'entities_change']

    @property
    def log_v2_description_change(self) -> list:
        """Description change"""
        return [
            changes for changes in self.log_v2_changes if changes['type'] == 'description_change'
        ]

    @property
    def log_v2_comment_change(self) -> list:
        """Comment change"""
        return [changes for changes in self.log_v2_changes if changes['type'] == 'comment_change']

    @property
    def log_v2_action_change(self) -> list:
        """Action change"""
        return [changes for changes in self.log_v2_changes if changes['type'] == 'action_change']

    @property
    def log_v2_assessment_ids_change(self) -> list:
        """Assessment change"""
        return [
            changes for changes in self.log_v2_changes if changes['type'] == 'assessment_ids_change'
        ]

    @property
    def panel_log(self) -> list:
        """Get playbook alert panel_log.

        Returns:
            list: Alert panel_log or [] if not found in the raw payload
        """
        warnings.warn(DEPRECATION_MSG, DeprecationWarning)
        return self.data.get('panel_log', [])

    @property
    def log_changes(self) -> list:
        """Returns a list of log panels with changes to the status, priority, or alert's assignee.

        For a more direct list of change, timestamp tuples, use one of:
        - `status_changes`
        - `priority_changes`
        - `assignee_changes`

        Returns:
            list: list of panels with status, priority, and assignee changes
        """
        warnings.warn(DEPRECATION_MSG, DeprecationWarning)
        return [panel for panel in self.panel_log if panel.get('changes')]

    @property
    def log_context_changes(self) -> list:
        """Returns a list of log panels with context changes.

        Returns:
            list: list of panels with context changes
        """
        warnings.warn(DEPRECATION_MSG, DeprecationWarning)
        return [
            panel
            for panel in self.panel_log
            if panel.get('context') and panel.get('context', {}).get('changes')
        ]

    def _get_specific_context_changes(self, context_type: str, change_type: str) -> list:
        """Extract all the changes based on the context type and change required.

        Args:
            context_type (str): context of change
            change_type (str): type of change in log

        Returns:
            list: list of changes
        """
        context_changes = [
            panel
            for panel in self.log_context_changes
            if panel.get('context', {}).get('type') == context_type
        ]
        for panel in context_changes:
            panel['context']['changes'] = [
                change
                for change in panel['context']['changes']
                if change.get('type') == change_type
            ]

        return [panel for panel in context_changes if len(panel['context']['changes']) > 0]

    def _get_changes(self, change_type: str) -> list:
        """Returns a list of changes to status, priority, or assignee of the alert.

        Args:
            change_type (str): type of change to look for

        Returns:
            list: list of changes tuples
        """
        if change_type not in ('status_change', 'priority_change', 'assignee_change'):
            raise ValueError(
                'change_type must be one of: status_change, priority_change, assignee_change',
            )
        logs_with_changes = filter(
            lambda log: log.get('changes', {}).get(change_type),
            self.panel_log,
        )
        changes_of_interest = []
        for log in logs_with_changes:
            old = log.get('changes', {}).get(change_type, {}).get('old')
            new = log.get('changes', {}).get(change_type, {}).get('new')
            created = log.get('created')
            changes_of_interest.append((old, new, created))

        return changes_of_interest

    @property
    def log_status_changes(self) -> list:
        """Returns a list of changes to the status of the alert.

        Returns:
            list: list of changes tuples
        """
        warnings.warn(DEPRECATION_MSG, DeprecationWarning)
        return self._get_changes('status_change')

    @property
    def log_priority_changes(self) -> list:
        """Returns a list of priority changes to the priority of the alert.

        Returns:
            list: list of changes tuples
        """
        warnings.warn(DEPRECATION_MSG, DeprecationWarning)
        return self._get_changes('priority_change')

    @property
    def log_assignee_changes(self) -> list:
        """Returns a list of assignee changes to the status of the alert.

        Returns:
            list: list of changes tuples
        """
        warnings.warn(DEPRECATION_MSG, DeprecationWarning)
        return self._get_changes('assignee_change')

    @property
    def log_comments(self) -> list:
        """Returns a list of comments.

        Each comment contains:
            - id
            - actor_name
            - actor_id
            - created
            - modified
            - message (the comment itself)

        Return:
            list: list of comments
        """
        warnings.warn(DEPRECATION_MSG, DeprecationWarning)
        return list(filter(lambda log: log.get('message'), self.panel_log))
