##################################### TERMS OF USE ###########################################
# The following code is provided for demonstration purpose only, and should not be used      #
# without independent verification. Recorded Future makes no representations or warranties,  #
# express, implied, statutory, or otherwise, regarding any aspect of this code or of the     #
# information it may retrieve, and provides it both strictly “as-is” and without assuming    #
# responsibility for any information it may retrieve. Recorded Future shall not be liable    #
# for, and you assume all risk of using, the foregoing. By using this code, Customer         #
# represents that it is solely responsible for having all necessary licenses, permissions,   #
# rights, and/or consents to connect to third party APIs, and that it is solely responsible  #
# for having all necessary licenses, permissions, rights, and/or consents to any data        #
# accessed from any third party API.                                                         #
##############################################################################################

import warnings

from psengine.playbook_alerts.constants import DEPRECATION_MSG

from .rf_base_playbook_alert import RFBasePlaybookAlert


class RFPlaybookAlertDomainAbuse(RFBasePlaybookAlert):
    """Playbook Alert for Domain Abuse."""

    def __init__(self, category: str, raw_alert: dict):
        super().__init__(category, raw_alert)
        self._images = {}

    def store_image(self, image_id: str, image_bytes: bytes) -> None:
        """Domain Abuse: store image bytes in self.images dictionary.

        Args:
            image_id (str): image id
            image_bytes (bytes): image bytes
        """
        for screenshot in self.screenshots:
            if screenshot.get('image_id') == image_id:
                self._images[image_id] = {}
                self._images[image_id]['description'] = screenshot['description']
                self._images[image_id]['created'] = screenshot['created']
                self._images[image_id]['image_bytes'] = image_bytes

    @property
    def explanation(self) -> str:
        """Get the explanation why the alert triggered.

        Returns:
            str: Alert explanation or None if not found in the payload
        """
        return self.panel_evidence_summary.get('explanation', None)

    @property
    def context_list(self) -> list:
        """Domain Abuse: get the context_list.

        Returns:
            list: context_list or [] if not found in the payload
        """
        return self.panel_status.get('context_list', [])

    @property
    def panel_action(self) -> list:
        """Domain Abuse: get the panel_action.

        Returns:
            dict: panel_action or [] if not found in the payload
        """
        return self.data.get('panel_action', [])

    @property
    def panel_evidence_dns(self) -> dict:
        """Domain Abuse: get the panel_evidence_dns.

        Returns:
            dict: panel_evidence_dns or {} if not found in the payload
        """
        return self.data.get('panel_evidence_dns', {})

    @property
    def panel_evidence_whois(self) -> dict:
        """Domain Abuse: get the panel_evidence_whois.

        Returns:
            dict: panel_evidence_whois or {} if not found in the payload
        """
        return self.data.get('panel_evidence_whois', {})

    @property
    def resolved_record_list(self) -> list:
        """Domain Abuse: get the resolved_record_list.

        Returns:
            list: resolved_record_list or [] if not found in the payload
        """
        return self.panel_evidence_summary.get('resolved_record_list', [])

    @property
    def screenshots(self) -> list:
        """Domain Abuse: get the screenshots list.

        Returns:
            list: screenshots or [] if not found in the payload
        """
        return self.panel_evidence_summary.get('screenshots', [])

    @property
    def image_ids(self) -> list:
        """Domain Abuse: get the playbook alert image ids.

        Returns:
            list: Alert image ids or empty list if not found
        """
        ids = []
        if self.screenshots:
            ids = [screenshot.get('image_id', None) for screenshot in self.screenshots]
        return ids

    @property
    def images(self) -> dict:
        """Domain Abuse: get raw bytes of the screenshots. This data
        is stored in the following format:

        {
            image_id : {
                'description': "awesome image description",
                'created': "date",
                'image_bytes': b'xyz'
            }
        }

        Returns:
            dict: Alert images raw bytes or {} if not found
        """
        return self._images

    @property
    def log_v2_domain_abuse_dns_change(self) -> list:
        """DNS change"""
        return [changes for changes in self.log_v2_changes if changes['type'] == 'dns_change']

    @property
    def log_v2_domain_abuse_whois_change(self) -> list:
        """WHOIS change"""
        return [changes for changes in self.log_v2_changes if changes['type'] == 'whois_change']

    @property
    def log_v2_domain_abuse_logotype_change(self) -> list:
        """Logotype change"""
        return [
            changes
            for changes in self.log_v2_changes
            if changes['type'] == 'logotype_in_screenshot_change'
        ]

    @property
    def log_v2_domain_abuse_malicious_dns_change(self) -> list:
        """Malaicious DNS change"""
        return [
            changes for changes in self.log_v2_changes if changes['type'] == 'malicious_dns_change'
        ]

    @property
    def log_v2_domain_abuse_reregistration_change(self) -> list:
        """Reregistration change"""
        return [
            changes for changes in self.log_v2_changes if changes['type'] == 'reregistration_change'
        ]

    @property
    def log_v2_domain_abuse_malicious_url_change(self) -> list:
        """Malicious URL change"""
        return [
            changes for changes in self.log_v2_changes if changes['type'] == 'malicious_url_change'
        ]

    @property
    def log_v2_domain_abuse_screenshot_mentions_change(self) -> list:
        """Screenshot mentions change"""
        return [
            changes
            for changes in self.log_v2_changes
            if changes['type'] == 'screenshot_mentions_change'
        ]

    @property
    def log_domain_abuse_changes(self) -> list:
        """Domain Abuse: Filter for log panels with context type domain_abuse.

        Returns:
            list: log panel changes with context type domain_abuse, or [] if not found
        """
        warnings.warn(DEPRECATION_MSG, DeprecationWarning)
        return list(
            filter(
                lambda log: log.get('context', {}).get('type') == 'domain_abuse',
                self.log_context_changes,
            ),
        )

    @property
    def log_dns_changes(self) -> list:
        """Domain Abuse: Filter for log panels for dns_changes only.

        Returns:
            list: log panel changes with context type domain abuse containing only
                  dns_changes, or [] if not found
        """
        warnings.warn(DEPRECATION_MSG, DeprecationWarning)
        return self._get_specific_context_changes('domain_abuse', 'dns_change')

    @property
    def log_whois_changes(self) -> list:
        """Domain Abuse: Filter for log panels for whois_changes only.

        Returns:
            list: log panel changes with context type domain abuse containing only
                  whois_changes, or [] if not found
        """
        warnings.warn(DEPRECATION_MSG, DeprecationWarning)
        return self._get_specific_context_changes('domain_abuse', 'whois_change')

    @property
    def log_screenshot_mention_changes(self) -> list:
        """Domain Abuse: Filter for log panels for screenshot_mention_changes only.

        Returns:
            list: log panel changes with context type domain abuse containing only
                  screenshot_mention_changes, or [] if not found
        """
        warnings.warn(DEPRECATION_MSG, DeprecationWarning)
        return self._get_specific_context_changes('domain_abuse', 'screenshot_mention_change')

    @property
    def log_logotype_in_screenshot_changes(self) -> list:
        """Domain Abuse: Filter for log panels for logotype_in_screenshot_changes only.

        Returns:
            list: log panel changes with context type domain abuse containing only
                  logotype_in_screenshot_changes, or [] if not found
        """
        warnings.warn(DEPRECATION_MSG, DeprecationWarning)
        return self._get_specific_context_changes('domain_abuse', 'logotype_in_screenshot_change')

    @property
    def log_malicious_dns_changes(self) -> list:
        """Domain Abuse: Filter for log panels for malicious_dns_changes only.

        Returns:
            list: log panel changes with context type domain abuse containing only
                  malicious_dns_changes, or [] if not found
        """
        warnings.warn(DEPRECATION_MSG, DeprecationWarning)
        return self._get_specific_context_changes('domain_abuse', 'malicious_dns_change')
