##################################### TERMS OF USE ###########################################
# The following code is provided for demonstration purpose only, and should not be used      #
# without independent verification. Recorded Future makes no representations or warranties,  #
# express, implied, statutory, or otherwise, regarding any aspect of this code or of the     #
# information it may retrieve, and provides it both strictly “as-is” and without assuming    #
# responsibility for any information it may retrieve. Recorded Future shall not be liable    #
# for, and you assume all risk of using, the foregoing. By using this code, Customer         #
# represents that it is solely responsible for having all necessary licenses, permissions,   #
# rights, and/or consents to connect to third party APIs, and that it is solely responsible  #
# for having all necessary licenses, permissions, rights, and/or consents to any data        #
# accessed from any third party API.                                                         #
##############################################################################################

import json
import logging
import os
from datetime import datetime
from typing import Optional, Tuple, Union

import requests
from requests import ConnectionError, ConnectTimeout, HTTPError, ReadTimeout
from simplejson.errors import JSONDecodeError

from ..config.config import Config
from ..constants import (
    DEFAULT_KEY,
    ERR_MSG_KEY,
    RF_TOKEN_KWARG,
    ROOT_DIR,
    STANZA_FIELDS_KEY,
    TYPE_KEY,
    VALIDATOR_KEY,
)
from ..endpoints import (
    EP_PLAYBOOK_ALERT,
    EP_PLAYBOOK_ALERT_COMMON,
    EP_PLAYBOOK_ALERT_DOMAIN_ABUSE,
    EP_PLAYBOOK_ALERT_SEARCH,
)
from ..exceptions import WriteFileError
from ..helpers.helpers import OSHelpers, TimeHelpers
from ..rf_client import RFClient
from .errors import (
    PlaybookAlertFetchError,
    PlaybookAlertRetrieveImageError,
    PlaybookAlertSearchError,
    PlaybookAlertUpdateError,
)
from .mappings import CATEGORY_ENDPOINTS, CATEGORY_TO_OBJECT_MAP
from .rf_base_playbook_alert import RFBasePlaybookAlert
from .rf_pa_category import RFPACategory
from .rf_pa_code_repo_leakage import RFPlaybookAlertCodeRepoLeakage
from .rf_pa_cyber_vulnerability import RFPlaybookAlertCyberVulnerability
from .rf_pa_domain_abuse import RFPlaybookAlertDomainAbuse
from .rf_pa_third_party_risk import RFPlaybookAlertThirdPartyRisk

PLAYBOOK_ALERTS_STANZA_NAME = 'playbook_alerts'
LIMIT_SETTING = 'limit'
LIMIT_DEFAULT = 1000
FETCH_IMAGE_SETTING = 'fetch_image'
FETCH_IMAGE_DEFAULT = True
STATUSES_SETTING = 'statuses'
STATUSES_DEFAULT = []
CATEGORY_SETTING = 'category'
CATEGORY_DEFAULT = []
PRIORITY_SETTING = 'priority'
PRIORITY_DEFAULT = []
CREATED_SETTING = 'created'
CREATED_DEFAULT = None
UPDATED_SETTING = 'updated'
UPDATED_DEFAULT = None
OUTPUT_DIR_SETTING = 'output_dir'
UPDATE_STATUS_SETTING = 'update_status'
UPDATE_STATUS_DEFAULT = False
NEW_STATUS_SETTING = 'new_status'
NEW_STATUS_DEFAULT = 'InProgress'

DEFAULT_ALERTS_OUTPUT_DIR = os.path.join(ROOT_DIR, 'playbook_alerts')
PLAYBOOK_ALERTS_OUTPUT_FNAME = 'rf_playbook_alerts_'

ALERT_STATUSES = ['New', 'InProgress', 'Dismissed', 'Resolved']

PLAYBOOK_ALERTS_MGR_PARAMS = {
    STANZA_FIELDS_KEY: {
        LIMIT_SETTING: {
            DEFAULT_KEY: LIMIT_DEFAULT,
            TYPE_KEY: int,
            VALIDATOR_KEY: lambda x: x > 0,
            ERR_MSG_KEY: "[{}] '{}' is not a positive integer".format(
                PLAYBOOK_ALERTS_STANZA_NAME,
                LIMIT_SETTING,
            ),
        },
        FETCH_IMAGE_SETTING: {DEFAULT_KEY: FETCH_IMAGE_DEFAULT, TYPE_KEY: bool},
        STATUSES_SETTING: {
            DEFAULT_KEY: STATUSES_DEFAULT,
            TYPE_KEY: list,
        },
        CATEGORY_SETTING: {DEFAULT_KEY: CATEGORY_DEFAULT, TYPE_KEY: list},
        PRIORITY_SETTING: {DEFAULT_KEY: PRIORITY_DEFAULT, TYPE_KEY: list},
        CREATED_SETTING: {
            DEFAULT_KEY: CREATED_DEFAULT,
            TYPE_KEY: str,
            VALIDATOR_KEY: TimeHelpers.is_rel_time_valid,
            ERR_MSG_KEY: "[{}] '{}' is not a valid relative time".format(
                PLAYBOOK_ALERTS_STANZA_NAME,
                CREATED_SETTING,
            ),
        },
        UPDATED_SETTING: {
            DEFAULT_KEY: UPDATED_DEFAULT,
            TYPE_KEY: str,
            VALIDATOR_KEY: TimeHelpers.is_rel_time_valid,
            ERR_MSG_KEY: "[{}] '{}' is not a valid relative time".format(
                PLAYBOOK_ALERTS_STANZA_NAME,
                UPDATED_SETTING,
            ),
        },
        OUTPUT_DIR_SETTING: {
            DEFAULT_KEY: DEFAULT_ALERTS_OUTPUT_DIR,
            TYPE_KEY: str,
            VALIDATOR_KEY: OSHelpers.check_dir,
            ERR_MSG_KEY: "[{}] '{}' directory is invalid or has restricted permissions".format(
                PLAYBOOK_ALERTS_STANZA_NAME,
                OUTPUT_DIR_SETTING,
            ),
        },
        UPDATE_STATUS_SETTING: {
            DEFAULT_KEY: UPDATE_STATUS_DEFAULT,
            TYPE_KEY: bool,
            VALIDATOR_KEY: lambda x: x in [True, False],
            ERR_MSG_KEY: "[{}] '{}' must be one of [true, false]".format(
                PLAYBOOK_ALERTS_STANZA_NAME,
                UPDATE_STATUS_SETTING,
            ),
        },
        NEW_STATUS_SETTING: {
            DEFAULT_KEY: NEW_STATUS_DEFAULT,
            TYPE_KEY: str,
            VALIDATOR_KEY: lambda x: x in ALERT_STATUSES,
            ERR_MSG_KEY: "[{}] '{}' must be one of {}".format(
                PLAYBOOK_ALERTS_STANZA_NAME,
                NEW_STATUS_SETTING,
                ALERT_STATUSES,
            ),
        },
    },
}


class RFPlaybookAlertMgr:
    """Manages requests for Recorded Future playbook alerts."""

    def __init__(self, cfg: Config = None, **kwargs: dict):
        """Initializes the RFPlaybookAlertMgr object.

        Args:
            cfg (Config, optional): Config object. Defaults to None.
            limit (int, optional): Limit the number of alerts returned. Defaults to None.
            fetch_image (bool, optional): If True then fetch images. Defaults to False.
            statuses (list, optional): List of alert statuses. Defaults to None.
            category (RFPACategory, optional): Alert category. Defaults to None.
            priority (list, optional): List of alert priorities. Defaults to None.
            created (str, optional): Relative time to search for alerts created. Defaults to '3d'.
            updated (str, optional): Relative time to search for alerts updated. Defaults to None.
            rf_token (str, optional): Recorded Future API token. Defaults to None.
        """
        self.log = logging.getLogger(__name__)
        self.config = cfg if cfg is not None else Config()
        rf_token = kwargs.pop(RF_TOKEN_KWARG) if RF_TOKEN_KWARG in kwargs else None
        self.config.parse_stanza(PLAYBOOK_ALERTS_STANZA_NAME, PLAYBOOK_ALERTS_MGR_PARAMS)
        self.config.update_settings(
            [(PLAYBOOK_ALERTS_STANZA_NAME, PLAYBOOK_ALERTS_MGR_PARAMS)],
            kwargs,
        )
        self.rf_client = (
            RFClient(self.config, rf_token=rf_token)
            if rf_token is not None
            else RFClient(self.config)
        )
        self.config.log_settings(PLAYBOOK_ALERTS_STANZA_NAME)

    def fetch(
        self,
        query: Optional[str] = None,
        alert_id: Optional[str] = None,
        category: Optional[RFPACategory] = None,
        panels: Optional[dict] = None,
        fetch_image: Optional[bool] = None,
    ) -> None:
        """Fetch Playbook Alerts based on a query or do an individual lookup.

        Args:
            query (str, optional): Search query to search & fetch alerts. Defaults to None.
            alert_id (str, optional): Alert ID. Defaults to None.
            category (str, optional): Alert category. Defaults to None.
            panels (JSON, optional): Panels to fetch. Defaults to None.
            fetch_image (bool, optional): If True then fetch images. Defaults to None.

        Returns:
            PA Object - if individual lookup performed
            List of PA Objects - if query used for the lookup
        """
        if alert_id is None and category is None:
            # This covers two scenarios:
            # 1. no parameters were specified at all - this way _bulk will use settings.ini
            # 2. a user supplied a query
            p_alerts = self._bulk(query=query, panels=panels, fetch_image=fetch_image)
            return p_alerts

        if alert_id is None:
            raise ValueError('alert_id was not supplied')

        if category is None:
            raise ValueError('category was not supplied')

        p_alert = self._fetch(
            category=category,
            alert_id=alert_id,
            panels=panels,
            fetch_image=fetch_image,
        )
        return p_alert

    def _playbook_alert_factory(
        self,
        category: str,
        raw_alert: dict,
        alert_id: str,
    ) -> Union[
        RFPlaybookAlertDomainAbuse,
        RFPlaybookAlertThirdPartyRisk,
        RFPlaybookAlertCyberVulnerability,
        RFPlaybookAlertCodeRepoLeakage,
        RFBasePlaybookAlert,
    ]:
        """Return correct playbook alert type from raw alert, category, and ID.

        Args:
            category (string): string alert category
            raw_alert (dict): raw alert payload
            alert_id (string): string alert id

        Returns:
            RF playbook alert object
        """
        p_alert = None
        try:
            p_alert = CATEGORY_TO_OBJECT_MAP[category](category, raw_alert)
        except KeyError:
            # This way when we consume an unmapped(new) PA, we get a base object to work with
            self.log.warning(
                'Unmapped playbook alert category: {}. '.format(category)
                + 'Will initialize {} as a RFBasePlaybookAlert'.format(alert_id),
            )
            p_alert = RFBasePlaybookAlert(category, raw_alert)

        return p_alert

    def _fetch(
        self,
        category: RFPACategory,
        alert_id: str,
        panels: Optional[list] = None,
        fetch_image: Optional[bool] = None,
    ) -> RFBasePlaybookAlert:
        """Fetch a single playbook alert.

        Args:
            category  (str): Playbook alert category
            alert_id  (str): Playbook alert id
            panels (list, optional): List of panels to fetch. Defaults to None
            fetch_image (bool, optional): If True then fetch images. Defaults to None.

        Raises:
            PlaybookAlertFetchError: if the fetch fails
            PlaybookAlertFetchError: if the supplied arguments are invalid
            PlaybookAlertFetchError: if API response is invalid

        Returns:
            RFPlaybookAlert: instance of PlaybookAlert class
        """
        if alert_id is None or category is None:
            raise PlaybookAlertFetchError('alert_id and/or category was not supplied')

        category = category.lower()
        if category in CATEGORY_ENDPOINTS:
            endpoint = CATEGORY_ENDPOINTS[category]
        else:
            # Workaround to fetch new PAs that have not been officially supported
            self.log.warning(
                'Unknown playbook alert category: {}, for alert: {}. '.format(category, alert_id)
                + 'Using category as an endpoint for this lookup',
            )
            endpoint = '{}/{}'.format(EP_PLAYBOOK_ALERT, category)

        url = '{}/{}'.format(endpoint, alert_id)
        try:
            self.log.info('Fetching playbook alert: {}, category: {}'.format(alert_id, category))
            data = {}
            if panels:
                data = {'panels': panels}

            data = json.dumps(data)

            response = self.rf_client.post_request(url=url, data=data)
        except (HTTPError, ConnectionError, ConnectTimeout, ReadTimeout) as err:
            raise PlaybookAlertFetchError('Failed to fetch playbook alert: {}'.format(err)) from err

        try:
            self.log.debug('Getting playbook alert JSON data')
            raw_alert = response.json()
        except (AttributeError, JSONDecodeError) as err:
            self.log.debug(response.text)
            raise PlaybookAlertFetchError(
                'Invalid response from the API. Failed to fetch playbook alert: {}'.format(err)
            ) from err

        p_alert = self._playbook_alert_factory(category, raw_alert, alert_id)

        if isinstance(p_alert, RFPlaybookAlertDomainAbuse):
            self._handle_images(fetch_image, p_alert)

        return p_alert

    def _handle_images(self, fetch_image: bool, p_alert: RFBasePlaybookAlert) -> None:
        """Little helper function that simply checks if image fetching
        was requested and calls retrieve_images().

        Args:
            fetch_image (bool): True/False
            p_alert: Playbook Alert object
        """
        if fetch_image is None:
            fetch_image = self.fetch_image

        if fetch_image:
            self.retrieve_images(p_alert)

    def _bulk(
        self,
        query: Optional[dict] = None,
        panels: Optional[dict] = None,
        fetch_image: Optional[bool] = None,
    ) -> None:
        """Fetch alerts in bulk based on a search query.

        Args:
            query (dict, optional): Defaults to None.
            filter_id (str, optional): last ingested alert id. Defaults to None.
            panels (JSON, optional): Defaults to None.

        Raises:
            PlaybookAlertSearchError: if the search response is invalid and cannot be parsed

        Returns:
            list: PlaybookAlert objects
        """
        if query is None:
            self.log.info('Search query based on settings.ini input')
            query = self._prepare_search_query_from_config()
        else:
            self.log.info('Search query based on function input')

        if fetch_image is None:
            fetch_image = self.fetch_image

        search_results = self.search(query)
        alerts = []
        errors = 0
        for alert in search_results:
            alert_id = alert.get('playbook_alert_id')
            category = alert.get('category')
            try:
                alert = self._fetch(
                    category=category,
                    alert_id=alert_id,
                    panels=panels,
                    fetch_image=fetch_image,
                )
                alerts.append(alert)
            except PlaybookAlertFetchError:
                errors += 1

        if errors > 0:
            self.log.error('Failed to fetch {} alert(s). See errors above'.format(errors))

        return alerts

    def _check_type(
        self,
        func_args: dict,
        fields: Tuple,
        expected_types: Union[type, tuple],
        type_names: str,
    ) -> None:
        """Checks that the types of args are correctly assigned.
        If the types are incorrect it will print all the args with wrong type.

        Args:
            func_args (dict): locals()
            fields (Tuple): fields to verify
            expected_types (Union[type, tuple]): types expected, can be a single type or a tuple
            type_names (str): str repr of type, cant be easily retrieved if expected_types is tuple

        Raises:
            TypeError: If type is incorrect
        """
        if not all(
            isinstance(func_args.get(attr), expected_types) or func_args.get(attr) is None
            for attr in fields
        ):
            wrong_fields = [
                attr
                for attr in fields
                if func_args.get(attr) and not isinstance(func_args.get(attr), expected_types)
            ]
            raise TypeError('{} must be of type {}'.format(', '.join(wrong_fields), type_names))

    def _process_arg(
        self,
        attr: str,
        value: Union[str, list],
        list_or_str_args: str,
    ) -> Tuple[str, Union[str, list]]:
        """Return attribute and value normalized based on type of value.

        Args:
            attr (str): attribute to verify
            value (Union[str, list]): value of attribute
            list_or_str_args (str): args that are lists or str

        Returns:
            tuple: str, Union[str, list]
        """
        if attr == 'filter_from':
            return 'from', value
        elif attr == 'created_from':
            return 'created_range', {'from': value}
        elif attr == 'created_until':
            return 'created_range', {'until': value}
        elif attr == 'updated_from':
            return 'updated_range', {'from': value}
        elif attr == 'updated_until':
            return 'updated_range', {'until': value}
        elif attr in list_or_str_args and isinstance(value, str):
            return attr, [value]

        return attr, value

    def prepare_query(
        self,
        filter_from: int = 0,
        limit: int = LIMIT_DEFAULT,
        **kwargs: dict,
    ) -> dict:
        """Based on the supplied arguments returns a query that is ready to use
        for searching for Playbook Alerts. Functions taking the prepared query are:
            - search()
            - fetch().

        Args:
            filter_from (int, optional): Defaults to 0.
            limit (int, optional): Defaults to 1000.
            order_by (str, optional): Defaults to None.
            direction (str, optional): Defaults to None.
            entity (str or list, optional):  Defaults to None.
            statuses (str or list, optional): Defaults to None.
            priority (str or list, optional): Defaults to None.
            category (str or list, optional): Defaults to None.
            assignee (str or list, optional): Defaults to None.
            created_from (str, optional): Defaults to None.
            created_until (str, optional): Defaults to None.
            updated_from (str, optional): Defaults to None.
            updated_until (str, optional): Defaults to None.

        Raises:
            TypeError: when supplied arguments are wrong type

        Returns:
            dict: search query
        """
        created_time_args = ['created_from', 'created_until']
        updated_time_args = ['updated_from', 'updated_until']
        time_args = created_time_args + updated_time_args

        int_args = ['limit', 'filter_from']
        str_args = [
            'order_by',
            'direction',
        ]
        str_args += time_args

        list_or_str_args = ['entity', 'statuses', 'priority', 'category', 'assignee']

        kwargs['limit'] = limit
        kwargs['filter_from'] = filter_from
        self._check_type(kwargs, int_args, int, 'int')
        self._check_type(kwargs, str_args, str, 'str')
        self._check_type(kwargs, list_or_str_args, (list, str), 'list or str')

        # Construct default query
        query = {'created_range': {}, 'updated_range': {}}

        for arg in kwargs:
            key, value = self._process_arg(arg, kwargs[arg], list_or_str_args)
            if isinstance(value, dict):
                query[key].update(value)
            else:
                query[key] = value

        # remove any keys that have empty dictionaries or lists
        query = {
            key: val
            for key, val in query.items()
            if not ((isinstance(val, (dict, list))) and len(val) == 0)
        }

        return query

    def search(self, query: dict = {}) -> dict:
        """Search playbook alerts.

        Args:
            query (dict, optional): Defaults to {}.

        Raises:
            TypeError: Raised if the specified query is not a dict and not a JSON
            PlaybookAlertSearchError: Raised when the search fails

        Returns:
            dict: search results
        """
        if isinstance(query, dict):
            try:
                query = json.dumps(query)
            except TypeError as err:
                raise TypeError('Invalid search query. Could not convert to JSON: {}'.format(err))
        else:
            raise TypeError('Search query must be a dict')

        try:
            self.log.info('Searching for playbook alerts. Query: {}'.format(query))
            response = self.rf_client.post_request(EP_PLAYBOOK_ALERT_SEARCH, query)
        except (HTTPError, ConnectionError, ConnectTimeout, ReadTimeout) as err:
            raise PlaybookAlertSearchError('Playbook alert search failed. Cause: {}'.format(err))

        try:
            search_results = response.json()
        except (AttributeError, JSONDecodeError) as err:
            raise PlaybookAlertSearchError(
                'Invalid response from the API. Playbook alert search failed: {}'.format(err)
            ) from err

        status_message = search_results.get('status', {}).get('status_message')
        count_returned = search_results.get('counts', {}).get('returned')
        count_total = search_results.get('counts', {}).get('total')
        self.log.info(
            'Status: {}, Returned {}, Total {}'.format(status_message, count_returned, count_total),
        )

        search_results = search_results.get('data', None)
        if search_results is None:
            self.log.debug("Search results are missing 'data' key: {}".format(search_results))
            raise PlaybookAlertSearchError("Search results are missing 'data' key")

        return search_results

    def _prepare_search_query_from_config(self) -> dict:
        """Return a basic search query based on the settings file."""
        query = {}

        if self.limit is not None:
            query['limit'] = self.limit

        if len(self.statuses) != 0:
            query['statuses'] = self.statuses

        if len(self.category) != 0:
            query['category'] = self.category

        if len(self.priority) != 0:
            query['priority'] = self.priority

        if self.created is not None:
            query['created_range'] = {}
            query['created_range']['from'] = TimeHelpers.rel_time_to_date(self.created)

        if self.updated is not None:
            query['updated_range'] = {}
            query['updated_range']['from'] = TimeHelpers.rel_time_to_date(self.updated)

        return query

    def update(
        self,
        alert: RFBasePlaybookAlert,
        priority: Optional[str] = None,
        status: Optional[str] = None,
        assignee: Optional[str] = None,
        log_entry: Optional[str] = None,
        reopen_strategy: Optional[str] = None,
    ) -> requests.Response:
        """Update a playbook alert.

        Args:
            alert (RFBasePlaybookAlert): Playbook alert to update
            priority (str, optional): Alert priority. Defaults to None.
            status (str, optional): Alert Status. Defaults to None.
            assignee (str, optional): Assignee. Defaults to None.
            log_entry (str, optional): Log entry. Defaults to None.
            reopen_strategy (str, optional): Reopen strategy. Defaults to None.

        Raises:
            ValueError: Playbook alert was not supplied
            ValueError: No update parameters were supplied

        Returns:
            response: API response
        """
        if alert is None:
            raise ValueError('playbook alert was not supplied')

        if (
            priority is None
            and status is None
            and assignee is None
            and log_entry is None
            and reopen_strategy is None
        ):
            raise ValueError('No update parameters were supplied')

        query = {}
        if priority is not None:
            query['priority'] = priority

        if status is not None:
            query['status'] = status

        if assignee is not None:
            query['assignee'] = assignee

        if log_entry is not None:
            query['log_entry'] = log_entry

        if reopen_strategy is not None:
            query['reopen'] = reopen_strategy

        return self._update(alert.id, query)

    def _update(self, alert_id: str, payload: dict) -> requests.Response:
        """Performs playbook alert update PUT request.

        Args:
            alert_id (str): ID of alert to update
            payload (dict): payload content for alert update

        Raises:
            PlaybookAlertUpdateError: Raised if the update request fails

        Returns:
            requests.Response: playbook alert update response
        """
        url = '{}/{}'.format(EP_PLAYBOOK_ALERT_COMMON, alert_id)
        try:
            self.log.info('Updating playbook alert: {}'.format(alert_id))
            response = self.rf_client.put_request(url=url, data=json.dumps(payload))

        except (HTTPError, ConnectionError, ConnectTimeout, ReadTimeout) as err:
            raise PlaybookAlertUpdateError(err=err)

        return response

    def write_file(self, p_alert_data: list) -> None:
        """Writes playbook alerts payload to disk.

        Args:
            playbook_alerts (list): Alerts from a call to fetch_alerts
        """
        if p_alert_data is None:
            self.log.warning(
                "Empty 'playbook_alerts' argument passed to RFPlaybookAlertMgr.write_file",
            )
            return

        p_alert_data = p_alert_data if isinstance(p_alert_data, list) else [p_alert_data]
        alert_list = [alert.raw for alert in p_alert_data]

        formatted_time = datetime.now().strftime('%Y-%m-%dT%H-%M-%S-%f')
        output_file = PLAYBOOK_ALERTS_OUTPUT_FNAME + formatted_time + '.json'
        self.log.info('Writing playbook alerts to file')
        try:
            dir_path = OSHelpers.mkdir(self.output_dir)
            full_path = os.path.join(dir_path, output_file)
            with open(full_path, 'w') as outfile:
                outfile.write(json.dumps(alert_list, indent=2))
                self.log.info('  Wrote: {}'.format(full_path))

        except (OSError, JSONDecodeError, WriteFileError) as err:
            raise WriteFileError(str(err))

    def retrieve_images(self, playbook_alert: RFPlaybookAlertDomainAbuse) -> None:
        """Domain Abuse: Retrieve the associated images, if any available.

        Args:
            pa (RFPlaybookAlertDomainAbuse): Domain Abuse Playbook Alert
        """
        if isinstance(playbook_alert, RFPlaybookAlertDomainAbuse):
            for image_id in playbook_alert.image_ids:
                image_bytes = self.retrieve_image(playbook_alert.id, image_id)
                playbook_alert.store_image(image_id, image_bytes)
        else:
            self.log.debug('Image fetching is only supported for Domain Abuse alerts')

    def retrieve_image(self, alert_id: str, image_id: str) -> bytes:
        """Retrieve image from Domain Abuse playbook alert.

        Args:
            alert_id (str): Playbook alert id
            image_id (str): Playbook alert image id

        Raises:
            PlaybookAlertRetrieveImageError: if the image fetch fails

        Returns:
            bytes: bytes of the image
        """
        url = os.path.join(EP_PLAYBOOK_ALERT_DOMAIN_ABUSE, alert_id, 'image', image_id)

        try:
            self.log.info('Retrieving image: {} for alert: {}'.format(image_id, alert_id))
            response = self.rf_client.get_request(url)
        except (HTTPError, ConnectionError, ConnectTimeout, ReadTimeout) as err:
            raise PlaybookAlertRetrieveImageError(err=err)

        return response.content

    def save_images(self, pa: RFPlaybookAlertDomainAbuse) -> None:
        """Save Domain Abuse images/screenshots to disk as a PNG.

        Args:
            pa (RFPlaybookAlertDomainAbuse): Domain Abuse Playbook Alert
        """
        if not isinstance(pa, RFPlaybookAlertDomainAbuse):
            self.log.warning('Image saving is only supported by Domain Abuse alerts')
            return

        if not pa.images:
            self.log.warning('No images to save for alert: {}'.format(pa.id))
            return

        self.log.info('Saving {} image(s) to disk for alert {}'.format(len(pa.images), pa.id))
        for image_id, meta in pa.images.items():
            file_name = '{}_{}'.format(pa.id[5:], image_id[4:])
            self.save_image(image_name=file_name, image_bytes=meta['image_bytes'])

    def save_image(self, image_name: str, image_bytes: bytes) -> None:
        """Save image to disk as a PNG.

        Args:
            image_name (str): file name without the extension
            image_bytes (bytes): image bytes

        Raises:
            WriteFileError: if the image save fails
        """
        try:
            self.log.debug("Saving image '{}' to disk".format(image_name))
            dir_path = OSHelpers.mkdir(self.output_dir)
            image_filepath = os.path.join(dir_path, '{}.png'.format(image_name))
            with open(image_filepath, 'wb') as file:
                file.write(image_bytes)
        except OSError as err:
            raise WriteFileError(
                'Failed to save playbook alert image to disk. Cause: {}'.format(err.args),
            )

    @property
    def output_dir(self) -> str:
        """Getter for output_dir."""
        output_dir = self.config.settings.get(PLAYBOOK_ALERTS_STANZA_NAME, {}).get(
            OUTPUT_DIR_SETTING,
        )
        if output_dir is None:
            output_dir = DEFAULT_ALERTS_OUTPUT_DIR
            self.log.warning(
                '{} was not set, using default: {}'.format(OUTPUT_DIR_SETTING, output_dir),
            )
        return output_dir

    @output_dir.setter
    def output_dir(self, value: str) -> None:
        """Setter for output_dir."""
        if not isinstance(value, str):
            raise TypeError('{} must be a string'.format(OUTPUT_DIR_SETTING))
        if not PLAYBOOK_ALERTS_MGR_PARAMS[STANZA_FIELDS_KEY][OUTPUT_DIR_SETTING][VALIDATOR_KEY](
            value,
        ):
            raise ValueError(
                'Invalid value for {}: {}. Must be a valid directory'.format(
                    OUTPUT_DIR_SETTING,
                    value,
                ),
            )
        self.config.update_setting(PLAYBOOK_ALERTS_STANZA_NAME, OUTPUT_DIR_SETTING, value)

    @property
    def limit(self) -> int:
        """Getter for limit."""
        limit = self.config.settings.get(PLAYBOOK_ALERTS_STANZA_NAME, {}).get(LIMIT_SETTING)
        if limit is None:
            limit = LIMIT_DEFAULT
            self.log.warning('limit was not set, using default: {}'.format(limit))
        return limit

    @limit.setter
    def limit(self, value: int) -> None:
        """Setter for limit."""
        if not isinstance(value, int):
            raise TypeError('limit must be an integer')
        if not PLAYBOOK_ALERTS_MGR_PARAMS[STANZA_FIELDS_KEY][LIMIT_SETTING][VALIDATOR_KEY](value):
            raise ValueError(
                'Invalid value for limit: {}. Must be a positive integer'.format(value),
            )
        self.config.update_setting(PLAYBOOK_ALERTS_STANZA_NAME, LIMIT_SETTING, value)

    @property
    def statuses(self) -> list:
        """Getter for statuses."""
        statuses = self.config.settings.get(PLAYBOOK_ALERTS_STANZA_NAME, {}).get(STATUSES_SETTING)
        if statuses is None:
            statuses = STATUSES_DEFAULT
            self.log.warning('statuses was not set, using default: {}'.format(statuses))
        return statuses

    @statuses.setter
    def statuses(self, value: list) -> None:
        """Setter for statuses."""
        if not isinstance(value, list):
            raise TypeError('statuses must be a list')
        self.config.update_setting(PLAYBOOK_ALERTS_STANZA_NAME, STATUSES_SETTING, value)

    @property
    def category(self) -> list:
        """Getter for category."""
        category = self.config.settings.get(PLAYBOOK_ALERTS_STANZA_NAME, {}).get(CATEGORY_SETTING)
        if category is None:
            category = CATEGORY_DEFAULT
            self.log.warning('category was not set, using default: {}'.format(category))
        return category

    @category.setter
    def category(self, value: list) -> None:
        """Setter for category."""
        if not isinstance(value, list):
            raise TypeError('category must be a list')
        self.config.update_setting(PLAYBOOK_ALERTS_STANZA_NAME, CATEGORY_SETTING, value)

    @property
    def priority(self) -> list:
        """Getter for priority."""
        priority = self.config.settings.get(PLAYBOOK_ALERTS_STANZA_NAME, {}).get(PRIORITY_SETTING)
        if priority is None:
            priority = PRIORITY_DEFAULT
            self.log.warning('priority was not set, using default: {}'.format(priority))
        return priority

    @priority.setter
    def priority(self, value: list) -> None:
        """Setter for priority."""
        if not isinstance(value, list):
            raise TypeError('priority must be a list')
        self.config.update_setting(PLAYBOOK_ALERTS_STANZA_NAME, PRIORITY_SETTING, value)

    @property
    def created(self) -> str:
        """Getter for created."""
        created = self.config.settings.get(PLAYBOOK_ALERTS_STANZA_NAME, {}).get(CREATED_SETTING)
        if created is None:
            created = CREATED_DEFAULT
            self.log.warning('created was not set, using default: {}'.format(created))
        return created

    @created.setter
    def created(self, value: str) -> None:
        """Setter for created."""
        if not isinstance(value, str):
            raise TypeError('created must be a string')
        if not PLAYBOOK_ALERTS_MGR_PARAMS[STANZA_FIELDS_KEY][CREATED_SETTING][VALIDATOR_KEY](value):
            raise ValueError(
                'Invalid value for created: {}. Must be a valid relative time'.format(value),
            )
        self.config.update_setting(PLAYBOOK_ALERTS_STANZA_NAME, CREATED_SETTING, value)

    @property
    def updated(self) -> str:
        """Getter for updated."""
        updated = self.config.settings.get(PLAYBOOK_ALERTS_STANZA_NAME, {}).get(UPDATED_SETTING)
        if updated is None:
            updated = UPDATED_DEFAULT
            self.log.warning('updated was not set, using default: {}'.format(updated))
        return updated

    @updated.setter
    def updated(self, value: str) -> None:
        """Setter for updated."""
        if not isinstance(value, str):
            raise TypeError('updated must be a string')
        if not PLAYBOOK_ALERTS_MGR_PARAMS[STANZA_FIELDS_KEY][UPDATED_SETTING][VALIDATOR_KEY](value):
            raise ValueError(
                'Invalid value for updated: {}. Must be a valid relative time'.format(value),
            )
        self.config.update_setting(PLAYBOOK_ALERTS_STANZA_NAME, UPDATED_SETTING, value)

    @property
    def fetch_image(self) -> bool:
        """Getter to enable/disable image fetchning."""
        fetch_image = self.config.settings.get(PLAYBOOK_ALERTS_STANZA_NAME, {}).get(
            FETCH_IMAGE_SETTING,
        )
        if fetch_image is None:
            fetch_image = FETCH_IMAGE_DEFAULT
            self.log.warning('fetch_image was not set, using default: {}'.format(fetch_image))
        return fetch_image

    @fetch_image.setter
    def fetch_image(self, value: bool) -> None:
        """Setter to enable/disable image fetchning."""
        if not isinstance(value, bool):
            raise TypeError('fetch_image must be a boolean')
        self.config.update_setting(PLAYBOOK_ALERTS_STANZA_NAME, FETCH_IMAGE_SETTING, value)

    @property
    def update_status(self):
        """Getter for update_status."""
        update_status = self.config.settings.get(PLAYBOOK_ALERTS_STANZA_NAME, {}).get(
            UPDATE_STATUS_SETTING
        )
        if update_status is None:
            update_status = UPDATE_STATUS_DEFAULT
            self.log.warning('update_status was not set, using default: {}'.format(update_status))
        return update_status

    @update_status.setter
    def update_status(self, value) -> None:
        """Setter for update_status."""
        if not isinstance(value, bool):
            raise TypeError('update_status must be a boolean')
        if not PLAYBOOK_ALERTS_MGR_PARAMS[STANZA_FIELDS_KEY][UPDATE_STATUS_SETTING][VALIDATOR_KEY](
            value
        ):
            raise ValueError('Invalid value for update_status: {}'.format(value))
        self.config.update_setting(PLAYBOOK_ALERTS_STANZA_NAME, UPDATE_STATUS_SETTING, value)

    @property
    def new_status(self):
        """Getter for new_status."""
        new_status = self.config.settings.get(PLAYBOOK_ALERTS_STANZA_NAME, {}).get(
            NEW_STATUS_SETTING
        )
        if new_status is None:
            new_status = NEW_STATUS_DEFAULT
            self.log.warning('new_status was not set, using default: {}'.format(new_status))
        return new_status

    @new_status.setter
    def new_status(self, value) -> None:
        """Setter for new_status."""
        if not isinstance(value, str):
            raise TypeError('new_status must be a string')
        if not PLAYBOOK_ALERTS_MGR_PARAMS[STANZA_FIELDS_KEY][NEW_STATUS_SETTING][VALIDATOR_KEY](
            value
        ):
            raise ValueError('Invalid value for new_status: {}'.format(value))
        self.config.update_setting(PLAYBOOK_ALERTS_STANZA_NAME, NEW_STATUS_SETTING, value)
