##################################### TERMS OF USE ###########################################
# The following code is provided for demonstration purpose only, and should not be used      #
# without independent verification. Recorded Future makes no representations or warranties,  #
# express, implied, statutory, or otherwise, regarding any aspect of this code or of the     #
# information it may retrieve, and provides it both strictly “as-is” and without assuming    #
# responsibility for any information it may retrieve. Recorded Future shall not be liable    #
# for, and you assume all risk of using, the foregoing. By using this code, Customer         #
# represents that it is solely responsible for having all necessary licenses, permissions,   #
# rights, and/or consents to connect to third party APIs, and that it is solely responsible  #
# for having all necessary licenses, permissions, rights, and/or consents to any data        #
# accessed from any third party API.                                                         #
##############################################################################################

import json
import logging
import os
import re
from distutils import util as distutils_util
from typing import Union

import jsonpath_ng
from jsonpath_ng.exceptions import JsonPathParserError
from requests import (
    ConnectionError,
    ConnectTimeout,
    HTTPError,
    ReadTimeout,
    Response,
    Session,
    adapters,
)
from requests.adapters import HTTPAdapter, Retry
from requests.exceptions import JSONDecodeError

from psengine.endpoints import BASE_URL

from ._sdk_id import SDK_ID
from .config import Config
from .constants import _API_RESPONSE_LIMIT, STANZA_FIELDS_KEY, VALIDATOR_KEY
from .helpers import OSHelpers

# RF token origin
KWARG_ORIGIN = 'RFClient init kwarg'
CONFIG_ORIGIN = 'psengine Config object'
SETTER_ORIGIN = 'RFClient object setter'

# In seconds
REQUEST_TIMEOUT = 120
POOL_MAX_SIZE = 120

# Request RETRY configuration
RETRY_TOTAL = 5
BACKOFF_FACTOR = 1
STATUS_FORCELIST = [502, 503, 504]

# Params
PARAM_LIMIT = 'limit'
PARAM_FROM = 'from'

# Do not edit, this specifies the env variable name to get the key from
RF_API_KEY = 'RF_API_KEY'

RFAPI_STANZA_NAME = 'rfapi'
PROXY_ENABLED_SETTING = 'proxy_enabled'
PROXY_ENABLED_DEFAULT = False
PROXY_SERVERS_SETTING = 'proxy_servers'
PROXY_SERVERS_DEFAULT = []
VERIFY_SSL_SETTING = 'verify_ssl'
VERIFY_SSL_DEFAULT = True
CERT_SETTING = 'cert'
CERT_DEFAULT = (None, None)
AUTH_SETTING = 'auth'
AUTH_DEFAULT = (None, None)
RFAPI_STANZA_PARAMS = {
    STANZA_FIELDS_KEY: {
        PROXY_ENABLED_SETTING: {
            'default': PROXY_ENABLED_DEFAULT,
            'type': bool,
            'validator': lambda x: x in [True, False],
            'err_message': "[{}] '{}' must be one of [true, false]".format(
                RFAPI_STANZA_NAME,
                PROXY_ENABLED_SETTING,
            ),
        },
        PROXY_SERVERS_SETTING: {
            'default': PROXY_SERVERS_DEFAULT,
            'type': list,
            'validator': lambda x: all(isinstance(y, str) for y in x),
            'err_message': "[{}] '{}' must be a list of string URLs".format(
                RFAPI_STANZA_NAME,
                PROXY_SERVERS_SETTING,
            ),
        },
        VERIFY_SSL_SETTING: {
            'default': VERIFY_SSL_DEFAULT,
            'type': object,
        },
        CERT_SETTING: {
            'default': CERT_DEFAULT,
            'type': tuple,
        },
        AUTH_SETTING: {
            'default': AUTH_DEFAULT,
            'type': tuple,
        },
    },
}


def is_key_valid(key: str):
    """Checks if the key is valid.

    Args:
        key (str): Recorded Future API key

    Returns:
        bool: True if key is valid, False otherwise
    """
    if not isinstance(key, str):
        return False
    return re.match(r'^[a-zA-Z0-9]{32}$', key) is not None


class RFClient:
    """Makes HTTPS request to the Recorded Future API and validates results."""

    def __init__(self, config: Config = None, **kwargs):  # noqa: C901
        """Initializes the RFClient object.

        Args:
            config (Config, optional): psengine.Config object
            rf_token (str, optional): Recorded Future API key
            cert (tuple, optional): cert and key tuple
            verify_ssl (Union[bool, str], optional): verify SSL boolean or cert bundle path
            proxy_enabled (bool, optional): use proxy
            proxy_servers (list, optional): proxy settings

        Raises:
            ValueError: if RF API key is not found
        """
        self.log = logging.getLogger(__name__)
        self.config = config if config is not None else Config()
        self.config.parse_stanza(RFAPI_STANZA_NAME, RFAPI_STANZA_PARAMS)
        proxies_dict = self._create_proxies_dict(
            self.config.settings[RFAPI_STANZA_NAME].get(PROXY_SERVERS_SETTING),
        )
        if kwargs is not None:
            kwargs = self._validate_and_set_args(kwargs)
            self.config.update_settings([(RFAPI_STANZA_NAME, RFAPI_STANZA_PARAMS)], kwargs)
            if PROXY_SERVERS_SETTING in kwargs:
                proxies_dict = self._create_proxies_dict(kwargs.get(PROXY_SERVERS_SETTING))

        if kwargs.get('rf_token', None) is not None:
            if not isinstance(kwargs.get('rf_token'), str):
                raise TypeError('RF API key must be a string')
            if not is_key_valid(kwargs.get('rf_token')):
                raise ValueError('Invalid Recorded Future API key')
            self.__rf_token = (kwargs.get('rf_token'), KWARG_ORIGIN)
        elif self.config.rf_token is not None:
            if not is_key_valid(self.config.rf_token):
                raise ValueError('Invalid Recorded Future API key')
            self.__rf_token = (self.config.rf_token, CONFIG_ORIGIN)
        else:
            env_token = os.environ.get(RF_API_KEY)
            if env_token is not None:
                if not is_key_valid(env_token):
                    raise ValueError('Invalid Recorded Future API key')
                self.__rf_token = (env_token, '{} environment variable'.format(RF_API_KEY))
            else:
                self.log.warning('No Recorded Future API key found')

        self.session = self._config_session(proxies_dict)
        if not isinstance(self.verify_ssl, bool) and self.verify_ssl.lower() in ('true', 'false'):
            self.verify_ssl = bool(distutils_util.strtobool(self.verify_ssl))
        if self.cert[0] is not None:
            self.session.cert = self.cert
        if self.auth[0] is not None:
            self.session.auth = self.auth
        self._set_retry_config()

    def _validate_and_set_args(self, kwargs):
        if PROXY_SERVERS_SETTING in kwargs and PROXY_ENABLED_SETTING not in kwargs:
            # Assume if proxies are set, proxy_enabled is True
            kwargs[PROXY_ENABLED_SETTING] = True

        return kwargs

    def _set_retry_config(self) -> None:
        retries = Retry(
            total=RETRY_TOTAL,
            backoff_factor=BACKOFF_FACTOR,
            status_forcelist=STATUS_FORCELIST,
        )
        for prefix in 'http://', 'https://':
            self.session.mount(prefix, HTTPAdapter(max_retries=retries))

    def _create_proxies_dict(self, proxy_servers: list):
        """Get proxies.

        Arguments:
            * proxy_servers - an array of proxies

        Returns:
            proxies - python dict object of http/https proxies formatted for urlib3.requests
        """
        proxies = {}
        for proxy in proxy_servers:
            if 'https:' in proxy:
                proxies['https'] = proxy
            else:
                proxies['http'] = proxy

        return proxies

    def _config_session(self, proxies):
        """Creates a requests.Session(), utilises the Requests HTTP Library and sets up proxies.

        Returns:
            * session - a requests session object
        """
        self.log.debug('Create a client session')
        session = Session()
        adapter = adapters.HTTPAdapter(pool_maxsize=POOL_MAX_SIZE)
        session.mount('https://', adapter)
        if self.proxy_enabled:
            self.log.debug('Set proxy settings: {}'.format(proxies))
            session.proxies.update(proxies)
        return session

    def end_session(self) -> None:
        """Close requests.Session() session."""
        self.session.close()

    def _prepare_headers(self):
        user_agent = self._prepare_user_agent_header()
        headers = {
            'User-Agent': user_agent,
            'Content-Type': 'application/json',
            'accept': 'application/json',
        }
        try:
            headers['X-RFToken'] = self.rf_token
            self.log.debug('Using Recorded Future API key from {}'.format(self.rf_token_origin))
        except AttributeError:
            self.log.warning('Request being made with no Recorded Future API key set')

        return headers

    def _prepare_user_agent_header(self):
        os_info = OSHelpers.os_platform()
        app_id = self.config.app_id
        platform_id = self.config.platform_id
        sdk_id = SDK_ID
        user_agent_list = []
        if app_id is not None:
            user_agent_list.append(app_id)

        if os_info is not None:
            user_agent_list.append('({})'.format(os_info))

        if sdk_id is not None:
            user_agent_list.append(sdk_id)

        if platform_id is not None:
            user_agent_list.append(platform_id)

        user_agent = ' '.join(user_agent_list)
        return user_agent

    def _get_error_message(self, response):
        """Safely extract error message from response."""
        try:
            return response.json().get('message') or response.json().get('error', {}).get(
                'message',
                '',
            )
        except (KeyError, JSONDecodeError):
            return ''

    def set_urllib_log_level(self, level: str) -> None:
        """Set log level for urllib3 library.

        Args:
            level (str): log level to be set
        """
        if not level or level.upper() not in ('CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG'):
            self.log.warning('Log level is empty or not valid')
            return
        logging.getLogger('urllib3').setLevel(level.upper())

    def get_request(self, url: str, params=None):
        """Makes a request to the Recorded Future API, returns JSON response.

        Args:
            url - string completing the URL path for the RF API request
            params   - dictionary containing query parameter names and values

        Returns:
            esponse - raw response object

        Raises:
            ValueError
        """
        self.log.debug("GET -> '{0}' PARAMS: {1}".format(url, params))
        headers = self._prepare_headers()
        try:
            response = self.session.get(
                url,
                params=params,
                headers=headers,
                verify=self.verify_ssl,
                timeout=REQUEST_TIMEOUT,
            )
            self.log.debug('HTTP Status Code: {}'.format(response.status_code))
        except (ConnectionError, ConnectTimeout, ReadTimeout):
            raise
        try:
            response.raise_for_status()
        except HTTPError as err:
            msg = str(err)
            additional_msg = self._get_error_message(response)
            if additional_msg:
                msg += ', Cause: {}'.format(additional_msg)
                self.log.error(msg)
            raise

        return response

    def delete_request(self, url: str, params: dict = None) -> Response:
        """Makes a delete request to the Recorded Future API, returns JSON response.

        Args:
            url(str): string completing the URL path for the RF API request
            params(dict): contains query parameter names and values

        Returns:
            esponse - raw response object

        Raises:
            (ConnectionError, ConnectTimeout, ReadTimeout) : if connection fails
            HTTPError: if response returned with a non 200 return code
        """
        self.log.debug("DELETE -> '{}' PARAMS: {}".format(url, params))
        headers = self._prepare_headers()
        try:
            response = self.session.delete(
                url,
                params=params,
                headers=headers,
                verify=self.verify_ssl,
                timeout=REQUEST_TIMEOUT,
            )
            self.log.debug('HTTP Status Code: {}'.format(response.status_code))
        except (ConnectionError, ConnectTimeout, ReadTimeout):
            raise

        try:
            response.raise_for_status()
        except HTTPError as err:
            msg = str(err)
            additional_msg = self._get_error_message(response)
            if additional_msg:
                msg += ', Cause: {}'.format(additional_msg)
                self.log.error(msg)
            raise

        return response

    def make_paged_request(self, url, params=None):
        """Makes an HTTP request to target URL and returns data.results field
        This function is to be used instead of get_request() when there are more
        than 100 results. It uses get_request() for each request.

        Args:
            * url     - string target URL
            * params  - dictionary containing query parameters to be urlencoded

        Returns:
            JSON results array
        """
        if params is not None and PARAM_LIMIT in params:
            params[PARAM_LIMIT] = min(params[PARAM_LIMIT], _API_RESPONSE_LIMIT)
        elif params is not None:
            params[PARAM_LIMIT] = _API_RESPONSE_LIMIT
        else:
            params = {PARAM_LIMIT: _API_RESPONSE_LIMIT}

        response = self.get_request(url, params)
        try:
            json_response = response.json()
        except JSONDecodeError:
            self.log.debug('Paged request does not contain valid JSON:\n{}'.format(response.text))
            raise

        if (
            'counts' not in json_response
            or 'total' not in json_response['counts']
            or 'returned' not in json_response['counts']
        ):
            return json_response

        seen = json_response['counts']['returned']
        if json_response['counts']['total'] > _API_RESPONSE_LIMIT:
            total = _API_RESPONSE_LIMIT
        else:
            total = json_response['counts']['total']

        results = json_response['data']['results']
        while seen < total:
            params[PARAM_FROM] = seen
            params[PARAM_LIMIT] = min(params[PARAM_LIMIT], _API_RESPONSE_LIMIT - seen)
            response = self.get_request(url, params)
            json_response = response.json()
            results += json_response['data']['results']
            seen += json_response['counts']['returned']

        return results

    def raw_post_request(self, url: str, data):
        """Makes a post request, returns raw response.

        Args:
            url - URL for the RF API request
            data - JSON data to be sent in the body of the request

        Returns:
            response - raw response object from Recorded Future API

        Raises:
            ValueError
        """
        self.log.debug("POST -> '{0}' DATA: {1}".format(url, data))
        headers = self._prepare_headers()
        try:
            response = self.session.post(
                url,
                data=data,
                headers=headers,
                verify=self.verify_ssl,
                timeout=REQUEST_TIMEOUT,
            )
        except (ConnectionError, ReadTimeout, ConnectTimeout):
            raise
        self.log.debug('HTTP Status Code: {}'.format(response.status_code))
        try:
            response.raise_for_status()
        except HTTPError as err:
            msg = str(err)
            additional_msg = self._get_error_message(response)
            if additional_msg:
                msg += ', Cause: {}'.format(additional_msg)
                self.log.error(msg)
            raise

        return response

    def _get_root_key(self, path: jsonpath_ng.jsonpath.Child) -> str:
        try:
            return self._get_root_key(path.left)
        except AttributeError:
            return str(path)

    def _get_matches(
        self, results_expr: jsonpath_ng.jsonpath.Fields, results: Union[list, dict]
    ) -> list:
        """Get matches from results

        Args:
            results_expr (jsonpath_ng): jsonpath_ng object
            results (dict): results

        Raises:
            KeyError: if no results are found

        Returns:
            list: list of matches
        """
        matches = results_expr.find(results)
        if len(matches) == 0:
            self.log.warning('No results found for path: {}'.format(str(results_expr)))
        results = []
        if len(matches) == 0:
            raise KeyError(str(results_expr))
        for match in matches:
            if isinstance(match.value, list):
                results += match.value
            else:
                results.append(match.value)
        return results

    def post_request(self, url: str, data, do_page: bool = False, results_path: str = 'data'):
        """Makes a POST request. Returns list when paging enabled, raw response otherwise

        Args:
            url (str): Full URL for RF API request
            data (_type_): JSON data to be sent in the body of the request
            do_page (bool, optional): Bool indicating whether to page results. Defaults to False
            results_path (str, optional): Path to extract paged results from. Defaults to 'data'

        Raises:
            ValueError: Raised if request data is not valid JSON, or if results path is invalid
            ValueError: _description_

        Returns:
            requests.models.Response or list: raw response, or list if paging enabled
        """
        if not isinstance(data, list) and not isinstance(data, dict):
            try:
                data = json.loads(data)
            except json.decoder.JSONDecodeError:
                raise ValueError('Invalid JSON in data: {}'.format(data))
        if do_page:
            try:
                results_expr = jsonpath_ng.parse(results_path)
            except JsonPathParserError:
                raise ValueError('Invalid results_path: {}'.format(results_path))
            root_key = self._get_root_key(results_expr)
            all_results = []
            self.log.debug('Making POST request with paging enabled')
            results = self.raw_post_request(url, json.dumps(data)).json()
            if root_key not in results:
                raise KeyError(results_path)
            if len(results[root_key]) == 0:
                return []
            all_results += self._get_matches(results_expr, results)
            while 'next_offset' in results:
                data['offset'] = results['next_offset']
                results = self.raw_post_request(url, json.dumps(data)).json()
                all_results += self._get_matches(results_expr, results)
            return all_results
        else:
            self.log.debug('Making POST request with paging disabled')
            return self.raw_post_request(url, json.dumps(data))

    def put_request(self, url: str, data):
        """Makes a put request, returns JSON response.

        Args:
            url - string completing the URL path for the RF API request
            data -

        Returns:
            json_response - raw JSON response object from Recorded Future API

        Raises:
            ValueError
        """
        self.log.debug("PUT -> '{0}' DATA: {1}".format(url, data))
        headers = self._prepare_headers()
        try:
            response = self.session.put(
                url,
                data=data,
                headers=headers,
                verify=self.verify_ssl,
                timeout=REQUEST_TIMEOUT,
            )
        except (ConnectionError, ReadTimeout, ConnectTimeout):
            raise
        self.log.debug('HTTP Status Code: {}'.format(response.status_code))
        try:
            response.raise_for_status()
        except HTTPError as err:
            msg = str(err)
            additional_msg = self._get_error_message(response)
            if additional_msg:
                msg += ', Cause: {}'.format(additional_msg)
                self.log.error(msg)
            raise

        return response

    def get_pool_max_size(self) -> int:
        """Return POOL_MAX_SIZE."""
        return POOL_MAX_SIZE

    def is_token_valid(self, url: str, method: str = 'get', data: dict = None) -> bool:
        """Validate token call given a URL

        Args:
            url (str): url to be validated
            method (str, optional): HTTP Method to be used.
                Allowed are 'get', 'post', 'delete', 'put'. Defaults to 'get'.
            data (dict, optional): data to be sent with the request. Defaults to None.

        Raises:
            ValueError: if method is not valid

        Returns:
            bool: True if token is valid, else otherwise
        """
        allowed_methods = ('get', 'post', 'delete', 'put')
        if method not in allowed_methods:
            raise ValueError(f'method should be one of {allowed_methods}')
        func = getattr(self, f'{method}_request')

        try:
            if data:
                func(url, data)
                return True

            func(url)
            return True
        except Exception as err:  # noqa: BLE001
            self.log.error(f'Error during validation: {err}')
            return False

    def is_connectivity_valid(self, url: str = BASE_URL) -> bool:
        """Validate if the url can be reached

        Args:
            url (str, optional): url to test. Defaults to BASE_URL (api.recordedfuture.com).

        Returns:
            bool: True if connection is 200 else False
        """
        try:
            request = self.get_request(url)
            request.raise_for_status()
            return True
        except Exception as err:  # noqa: BLE001
            self.log.error(f'Error during connectivity test: {err}')
            return False

    @property
    def verify_ssl(self):
        """Getter for verify_ssl."""
        verify_ssl = self.config.settings.get(RFAPI_STANZA_NAME, {}).get(VERIFY_SSL_SETTING)
        if verify_ssl is None:
            verify_ssl = VERIFY_SSL_DEFAULT
            self.log.warning(
                '{} was not set, using default: {}'.format(VERIFY_SSL_SETTING, verify_ssl),
            )
        return verify_ssl

    @verify_ssl.setter
    def verify_ssl(self, value) -> None:
        """Setter for verify_ssl."""
        if not isinstance(value, bool) and not isinstance(value, str):
            raise TypeError('{} must be a boolean or a string'.format(VERIFY_SSL_SETTING))
        self.config.settings[RFAPI_STANZA_NAME][VERIFY_SSL_SETTING] = value
        self.session.verify = value

    @property
    def cert(self):
        """Getter for cert."""
        cert = self.config.settings.get(RFAPI_STANZA_NAME, {}).get(CERT_SETTING)

        return cert

    @cert.setter
    def cert(self, value) -> None:
        """Setter for cert."""
        if not isinstance(value, tuple):
            raise TypeError('{} must be a tuple'.format(CERT_SETTING))
        self.config.settings[RFAPI_STANZA_NAME][CERT_SETTING] = value
        self.session.cert = value

    @property
    def auth(self):
        """Getter for auth."""
        auth = self.config.settings.get(RFAPI_STANZA_NAME, {}).get(AUTH_SETTING)

        return auth

    @auth.setter
    def auth(self, value) -> None:
        """Setter for auth."""
        if not isinstance(value, tuple):
            raise TypeError('{} must be a tuple'.format(AUTH_SETTING))
        self.config.settings[RFAPI_STANZA_NAME][AUTH_SETTING] = value
        self.session.auth = value

    @property
    def proxy_enabled(self):
        """Getter for proxy_enabled."""
        proxy_enabled = self.config.settings.get(RFAPI_STANZA_NAME, {}).get(PROXY_ENABLED_SETTING)
        if proxy_enabled is None:
            proxy_enabled = PROXY_ENABLED_DEFAULT
            self.log.warning(
                '{} was not set, using default: {}'.format(PROXY_ENABLED_SETTING, proxy_enabled),
            )
        return proxy_enabled

    @proxy_enabled.setter
    def proxy_enabled(self, value) -> None:
        """Setter for proxy_enabled."""
        if not isinstance(value, bool):
            raise TypeError('{} must be a boolean'.format(PROXY_ENABLED_SETTING))
        if not RFAPI_STANZA_PARAMS[STANZA_FIELDS_KEY][PROXY_ENABLED_SETTING][VALIDATOR_KEY](value):
            raise ValueError('Invalid value for proxy_enabled: {}'.format(value))
        self.config.settings[RFAPI_STANZA_NAME][PROXY_ENABLED_SETTING] = value

    @property
    def proxy_servers(self):
        """Getter for proxy_servers."""
        proxies = self.config.settings.get(RFAPI_STANZA_NAME, {}).get(PROXY_SERVERS_SETTING, {})
        if proxies is None:
            proxies = PROXY_SERVERS_DEFAULT
            self.log.warning(
                '{} was not set, using default: {}'.format(PROXY_SERVERS_SETTING, proxies),
            )
        return proxies

    @proxy_servers.setter
    def proxy_servers(self, value) -> None:
        """Setter for proxy_servers."""
        if not isinstance(value, list):
            raise TypeError('{} must be a list'.format(PROXY_SERVERS_SETTING))
        if not RFAPI_STANZA_PARAMS[STANZA_FIELDS_KEY][PROXY_SERVERS_SETTING][VALIDATOR_KEY](value):
            raise ValueError('Invalid value for proxies: {}'.format(value))
        self.config.settings[RFAPI_STANZA_NAME][PROXY_SERVERS_SETTING] = value
        proxies = self._create_proxies_dict(value)
        self.session.proxies = proxies

    @property
    def rf_token(self):
        """Getter for rf_token."""
        return self.__rf_token[0]

    @property
    def rf_token_origin(self):
        """Getter for rf_token origin."""
        return self.__rf_token[1]

    @rf_token.setter
    def rf_token(self, value) -> None:
        """Setter for rf_token."""
        if not isinstance(value, str):
            raise TypeError('rf_token must be a string')
        if not is_key_valid(value):
            raise ValueError('Invalid value for rf_token: {}'.format(value))
        self.__rf_token = (value, SETTER_ORIGIN)
