##################################### TERMS OF USE ###########################################
# The following code is provided for demonstration purpose only, and should not be used      #
# without independent verification. Recorded Future makes no representations or warranties,  #
# express, implied, statutory, or otherwise, regarding any aspect of this code or of the     #
# information it may retrieve, and provides it both strictly “as-is” and without assuming    #
# responsibility for any information it may retrieve. Recorded Future shall not be liable    #
# for, and you assume all risk of using, the foregoing. By using this code, Customer         #
# represents that it is solely responsible for having all necessary licenses, permissions,   #
# rights, and/or consents to connect to third party APIs, and that it is solely responsible  #
# for having all necessary licenses, permissions, rights, and/or consents to any data        #
# accessed from any third party API.                                                         #
##############################################################################################

import copy
import logging
import os
import re

from requests import ConnectionError, ConnectTimeout, HTTPError, ReadTimeout

from ..config.config import Config
from ..constants import (
    _API_RESPONSE_LIMIT,
    DEFAULT_KEY,
    ERR_MSG_KEY,
    RF_TOKEN_KWARG,
    ROOT_DIR,
    STANZA_FIELDS_KEY,
    TYPE_KEY,
    VALIDATOR_KEY,
)
from ..endpoints import API_VERSION, BASE_URL, EP_FUSION_FILES
from ..exceptions import WriteFileError
from ..helpers.helpers import OSHelpers
from ..rf_client import RFClient
from .constants import DEFAULT_RISKLIST_FORMAT
from .errors import RiskListNotAvailable

ENTITY_TYPES = ['ip', 'domain', 'url', 'hash', 'vulnerability']

DEFAULT_IP_STANZA_NAME = 'risklist_ip_default'
DEFAULT_DOMAIN_STANZA_NAME = 'risklist_domain_default'
DEFAULT_URL_STANZA_NAME = 'risklist_url_default'
DEFAULT_HASH_STANZA_NAME = 'risklist_hash_default'
DEFAULT_VULNERABILITY_STANZA_NAME = 'risklist_vulnerability_default'
ENTITY_TYPE_SETTING = 'entity_type'
ENABLED_SETTING = 'enabled'
ENABLED_DEFAULT = True
INTERVAL_SETTING = 'interval'
INTERVAL_DEFAULT = 3300
FUSION_FILE_SETTING = 'fusion_file'
LIST_SETTING = 'list'

RISKLIST_SETTINGS = {
    ENTITY_TYPE_SETTING: {
        TYPE_KEY: str,
        VALIDATOR_KEY: lambda x: x in ENTITY_TYPES,
        ERR_MSG_KEY: "Risklist 'entity_type' must be one of {}".format(ENTITY_TYPES),
    },
    ENABLED_SETTING: {
        DEFAULT_KEY: ENABLED_DEFAULT,
        TYPE_KEY: bool,
    },
    INTERVAL_SETTING: {
        DEFAULT_KEY: INTERVAL_DEFAULT,
        TYPE_KEY: int,
        VALIDATOR_KEY: lambda x: x > 0,
        ERR_MSG_KEY: "Risklist 'interval' must be a positive integer",
    },
}

LIST_SETTING_VALUE = {
    TYPE_KEY: str,
    VALIDATOR_KEY: lambda x: x != '',
    ERR_MSG_KEY: "Risklist 'list' must be a non-empty string",
}

FUSION_FILE_SETTING_VALUE = {
    TYPE_KEY: str,
    VALIDATOR_KEY: lambda x: x.startswith(('/home', '/public')),
    ERR_MSG_KEY: "Risklist 'fusion_file' must be a path starting with /home or /public",
}

DEFAULT_IP_STANZA_PARAMS = {STANZA_FIELDS_KEY: copy.deepcopy(RISKLIST_SETTINGS)}
DEFAULT_IP_STANZA_PARAMS[STANZA_FIELDS_KEY][ENTITY_TYPE_SETTING][DEFAULT_KEY] = 'ip'
DEFAULT_DOMAIN_STANZA_PARAMS = {STANZA_FIELDS_KEY: copy.deepcopy(RISKLIST_SETTINGS)}
DEFAULT_DOMAIN_STANZA_PARAMS[STANZA_FIELDS_KEY][ENTITY_TYPE_SETTING][DEFAULT_KEY] = 'domain'
DEFAULT_URL_STANZA_PARAMS = {STANZA_FIELDS_KEY: copy.deepcopy(RISKLIST_SETTINGS)}
DEFAULT_URL_STANZA_PARAMS[STANZA_FIELDS_KEY][ENTITY_TYPE_SETTING][DEFAULT_KEY] = 'url'
DEFAULT_HASH_STANZA_PARAMS = {STANZA_FIELDS_KEY: copy.deepcopy(RISKLIST_SETTINGS)}
DEFAULT_HASH_STANZA_PARAMS[STANZA_FIELDS_KEY][ENTITY_TYPE_SETTING][DEFAULT_KEY] = 'hash'
DEFAULT_VULNERABILITY_STANZA_PARAMS = {STANZA_FIELDS_KEY: copy.deepcopy(RISKLIST_SETTINGS)}
DEFAULT_VULNERABILITY_STANZA_PARAMS[STANZA_FIELDS_KEY][ENTITY_TYPE_SETTING][
    DEFAULT_KEY
] = 'vulnerability'

RISKLIST_OUTPUT_STANZA_NAME = 'output'
RISKLIST_OUTPUT_DIRECTORY_SETTING = 'risklists'
RISKLIST_OUTPUT_DIRECTORY_DEFAULT = os.path.join(ROOT_DIR, 'risklists')

RISKLIST_OUTPUT_STANZA_PARAMS = {
    STANZA_FIELDS_KEY: {
        RISKLIST_OUTPUT_DIRECTORY_SETTING: {
            DEFAULT_KEY: RISKLIST_OUTPUT_DIRECTORY_DEFAULT,
            TYPE_KEY: str,
            VALIDATOR_KEY: OSHelpers.check_dir,
            ERR_MSG_KEY: 'Specified risklists directory is invalid or has restricted permissions',
        },
    },
}


class RFRiskListMgr:
    """Manages requests for Recorded Future risk lists."""

    def __init__(self, cfg: Config = None, **kwargs):
        """Initializes RFRiskListMgr.

        Args:
            cfg (Config, optional): Config object
            risklist* (dict, optional): kwargs allow risklist settings
                                            entity_type (str): entity type
                                            enabled (bool): enabled
                                            interval (int): interval
                                            list (str, optional): list name
                                            fusion_file (str, optional): fusion file name
            rf_token (str, optional): Recorded Future API token. Defaults to None
        """
        self.log = logging.getLogger(__name__)
        self.config = cfg if cfg is not None else Config()
        rf_token = kwargs.pop(RF_TOKEN_KWARG) if RF_TOKEN_KWARG in kwargs else None
        self.stanza_params_mapping = {}
        # This is the only class where kwargs are set before settings.ini is parsed
        # Doing this so default lists aren't parsed if lists are provided as kwargs
        self._validate_and_set_args(kwargs)

        self.config.parse_stanza(RISKLIST_OUTPUT_STANZA_NAME, RISKLIST_OUTPUT_STANZA_PARAMS)
        if self.risklists == []:
            self.config.parse_stanza(DEFAULT_IP_STANZA_NAME, DEFAULT_IP_STANZA_PARAMS)
            self.config.parse_stanza(DEFAULT_DOMAIN_STANZA_NAME, DEFAULT_DOMAIN_STANZA_PARAMS)
            self.config.parse_stanza(DEFAULT_URL_STANZA_NAME, DEFAULT_URL_STANZA_PARAMS)
            self.config.parse_stanza(DEFAULT_HASH_STANZA_NAME, DEFAULT_HASH_STANZA_PARAMS)
            self.config.parse_stanza(
                DEFAULT_VULNERABILITY_STANZA_NAME,
                DEFAULT_VULNERABILITY_STANZA_PARAMS,
            )
        else:
            for risklist in self.risklists:
                if risklist in self.stanza_params_mapping:
                    # kwarg risklists parsed this way
                    self.config.parse_stanza(
                        risklist,
                        {STANZA_FIELDS_KEY: self.stanza_params_mapping[risklist]},
                    )
                else:
                    # risklists from settings.ini parsed this way
                    self.config.parse_stanza(risklist, {STANZA_FIELDS_KEY: RISKLIST_SETTINGS})

        self.rf_client = (
            RFClient(self.config, rf_token=rf_token)
            if rf_token is not None
            else RFClient(self.config)
        )
        self.api_response_limit = _API_RESPONSE_LIMIT

    def _type_and_value_check_risklist_setting(self, name, risklist_setting) -> None:  # noqa: C901
        if not isinstance(risklist_setting, dict):
            raise TypeError('Expected dict for risklist setting {}'.format(name))
        if ENTITY_TYPE_SETTING in risklist_setting:
            if not isinstance(risklist_setting[ENTITY_TYPE_SETTING], str):
                raise TypeError("Expected str for risklist '{}' 'entity_type' setting".format(name))
        else:
            raise ValueError("Missing 'entity_type' for risklist '{}'".format(name))
        if ENABLED_SETTING in risklist_setting and not isinstance(
            risklist_setting[ENABLED_SETTING],
            bool,
        ):
            raise TypeError("Expected bool for risklist '{}' 'enabled' setting".format(name))
        if INTERVAL_SETTING in risklist_setting and not isinstance(
            risklist_setting[INTERVAL_SETTING],
            int,
        ):
            raise TypeError("Expected int for risklist '{}' 'interval' setting".format(name))
        if LIST_SETTING in risklist_setting and FUSION_FILE_SETTING in risklist_setting:
            raise ValueError(
                "Cannot specify both 'list' and 'fusion_file' for risklist '{}'".format(name),
            )
        if LIST_SETTING in risklist_setting and not isinstance(risklist_setting[LIST_SETTING], str):
            raise TypeError("Expected str for risklist '{}' 'list' setting".format(name))
        if FUSION_FILE_SETTING in risklist_setting and not isinstance(
            risklist_setting[FUSION_FILE_SETTING],
            str,
        ):
            raise TypeError("Expected str for risklist '{}' 'fusion_file' setting".format(name))

    def _validate_and_set_args(self, kwargs) -> None:
        if kwargs is None:
            return
        for key, value in kwargs.items():
            # Raises TypeError if value is not a dict, and for incorrect types of dict values
            # Raises ValueError if ENTITY_TYPE_SETTING is missing
            # Raises ValueError for conflicting LIST_SETTING and FUSION_FILE_SETTING
            self._type_and_value_check_risklist_setting(key, value)
            stanza_params = copy.deepcopy(RISKLIST_SETTINGS)
            stanza_params[ENTITY_TYPE_SETTING][DEFAULT_KEY] = value[ENTITY_TYPE_SETTING]
            if LIST_SETTING in value:
                stanza_params[LIST_SETTING] = copy.deepcopy(LIST_SETTING_VALUE)
                stanza_params[LIST_SETTING][DEFAULT_KEY] = value[LIST_SETTING]
            if FUSION_FILE_SETTING in value:
                stanza_params[FUSION_FILE_SETTING] = copy.deepcopy(FUSION_FILE_SETTING_VALUE)
                stanza_params[FUSION_FILE_SETTING][DEFAULT_KEY] = value[FUSION_FILE_SETTING]
            for setting, setting_value in value.items():
                self.config.update_setting(key, setting, setting_value)
            self.stanza_params_mapping[key] = stanza_params

    def write_file(self, risklist, filename) -> None:
        """Writes risklist to file filename.
        The output directory is set in settings.ini.

        Args:
            risklist (str): risklist contents to write to file
            filename (str): string filename, should end in .csv
        """
        if risklist is None:
            self.log.warning("Empty 'risklist' argument passed to RFRiskListMgr.write_file")
            return
        self.log.info('Writing risklist to file')
        try:
            dir_path = OSHelpers.mkdir(self.risklist_output)
            full_path = os.path.join(dir_path, filename)
            with open(full_path, 'w') as outfile:
                outfile.write(risklist)
                self.log.info('  Wrote: {}'.format(full_path))
        except (OSError, WriteFileError) as err:
            raise WriteFileError('Could not write file {}: {}'.format(filename, err)) from err

    def get_risklists(self):
        """Fetches all configured risk lists."""
        risklists = []
        for risklist in self.risklists:
            try:
                risklists.append(self.get_risklist(risklist))
            except RiskListNotAvailable:  # noqa: PERF203 - no need to optimize, few risklists
                self.log.error(
                    'Risklist {} not available from RF API. Skipping...'.format(risklist),
                )
                continue

        return risklists

    def get_risklist(self, risklist_name):
        """Get a Recorded Future RiskList. Specify a fusion_path to get a custom
            risklist instead - format field is ignored when custom risklists are used.

        Gotchas:
            - if a specified list does not exists, RF API returns the default risklist
            - an empty risklist can be returned and the contents would look like this:
                <"Name","Risk","RiskString","EvidenceDetails"
                ...>

        Args:
            risklist_name (str): name of the risklist to download. Must be set in config

        Returns:
            risklist: raw risklist
        """
        if risklist_name not in self.risklists:
            raise ValueError(
                'Requested risklist {} does not exist in the configuration.'.format(risklist_name),
            )

        settings = self.get_risklist_settings(risklist_name)
        if not settings.get('enabled'):
            raise RiskListNotAvailable(
                'Risklist {} not enabled or is invalid'.format(risklist_name),
            )

        entity_type = settings.get('entity_type')
        # list, fusion_file, and format may not be present, so use null-safe access
        # set format if it is not set
        lst = settings.get('list')
        fusion_file = settings.get('fusion_file')
        frmt = settings.get('format')
        frmt = DEFAULT_RISKLIST_FORMAT if frmt is None else frmt
        self.log.info('Getting risklist: {}'.format(risklist_name))
        self.log.debug('  Entity Type: {}'.format(entity_type))
        self.log.debug('  Format: {}'.format(frmt))
        self.log.debug('  List: {}'.format(lst))
        self.log.debug('  Fusion File: {}'.format(fusion_file))
        if lst is not None and fusion_file is not None:
            raise ValueError('Parameters list and fusion_file can not be used in conjunction.')

        if entity_type not in ENTITY_TYPES:
            raise ValueError('Parameter entity_type is not valid.')

        if fusion_file is None:
            url = '{}/{}/{}/risklist'.format(BASE_URL, API_VERSION, entity_type)
            query = {
                'format': frmt,
            }
        else:
            url = EP_FUSION_FILES
            query = {
                'path': fusion_file,
            }

        if lst is not None:
            query['list'] = lst

        try:
            response = self.rf_client.get_request(url, query)

        except (HTTPError, ConnectionError, ConnectTimeout, ReadTimeout) as err:
            raise RiskListNotAvailable(
                'An error ocurred while fetching a risklist: {}'.format(err)
            ) from err

        return response.text

    def get_risklist_settings(self, risklist_name):
        """Get settings for a risklist."""
        risklist_settings = self.config.settings.get(risklist_name, {})

        return risklist_settings

    @property
    def risklists(self):
        """Get list of all risklists enabled."""
        risklist_p = re.compile('^risklist_')
        risklists = list(filter(risklist_p.match, self.config.settings.keys()))

        return risklists

    @property
    def risklist_output(self):
        """risklist_output getter."""
        risklist_output = self.config.settings.get(RISKLIST_OUTPUT_STANZA_NAME, {}).get(
            RISKLIST_OUTPUT_DIRECTORY_SETTING,
        )
        if risklist_output is None:
            risklist_output = RISKLIST_OUTPUT_DIRECTORY_DEFAULT
            self.log.warning(
                'risklist_output was not set, using default: {}'.format(risklist_output),
            )
        return risklist_output
