##################################### TERMS OF USE ###########################################
# The following code is provided for demonstration purpose only, and should not be used      #
# without independent verification. Recorded Future makes no representations or warranties,  #
# express, implied, statutory, or otherwise, regarding any aspect of this code or of the     #
# information it may retrieve, and provides it both strictly “as-is” and without assuming    #
# responsibility for any information it may retrieve. Recorded Future shall not be liable    #
# for, and you assume all risk of using, the foregoing. By using this code, Customer         #
# represents that it is solely responsible for having all necessary licenses, permissions,   #
# rights, and/or consents to connect to third party APIs, and that it is solely responsible  #
# for having all necessary licenses, permissions, rights, and/or consents to any data        #
# accessed from any third party API.                                                         #
##############################################################################################

import csv
import json
import logging
from io import StringIO

from stix2 import Bundle, Identity, Report
from stix2.exceptions import (
    ExtraPropertiesError,
    InvalidValueError,
    MissingPropertiesError,
    STIXError,
)

from ..analyst_notes import RFNote
from .constants import ENTITY_TYPE_MAP, REPORT_TYPE_MAPPER, SUPPORTED_HUNTING_RULES, TLP_MAP
from .errors import STIX2TransformError, UnsupportedConversionType
from .rf_complex_entity import DetectionRule, Indicator
from .rf_convert_entity import convert_entity
from .rf_enriched_entity import EnrichedIndicator
from .util import create_rf_author

LOG = logging.getLogger(__name__)


class RFBundle:
    """Class for creating STIX2 bundles from Recorded Future objects."""

    @classmethod
    def from_default_risklist(
        cls,
        risklist: str,
        entity_type: str,
        identity: Identity = None,
    ) -> Bundle:
        """Creates STIX2 bundle from a Recorded Future default risklist.

        Args:
            - risklist (str): Recorded Future default risklist (contains the standard 5 columns)
            - entity_type (str): entity type
            - identity (_type_, optional): Author identity. Defaults to Recorded Future

        Raises:
            - STIX2TransformError: if the risklist is not valid
            - STIX2TransformError: if EvidenceDetails is not valid JSON
            - STIX2TransformError: if the bundle cannot be created

        Returns:
            list[Bundle]: list of STIX2 bundles made from the risklist
        """
        if not identity:
            identity = create_rf_author()
        objects = [identity]
        LOG.info('Creating STIX2 bundle from {} risklist'.format(entity_type))

        csv_risklist = csv.DictReader(StringIO(risklist), delimiter=',')

        try:
            enriched_indicators = []
            for ioc in csv_risklist:
                indicator = EnrichedIndicator(
                    name=ioc['Name'],
                    type_=ENTITY_TYPE_MAP[entity_type],
                    confidence=ioc['Risk'],
                    evidence_details=json.loads(ioc['EvidenceDetails']).get('EvidenceDetails', []),
                )
                enriched_indicators.append(indicator)

        except KeyError as ke:
            raise STIX2TransformError('Risklist missing header: {}'.format(ke))

        except json.JSONDecodeError as jse:
            raise STIX2TransformError('EvidenceDetails is not valid JSON: {}'.format(jse))

        for i in enriched_indicators:
            objects.extend(i.stix_objects)

        try:
            bundle = Bundle(objects=objects, allow_custom=True)
        except (
            ValueError,
            ExtraPropertiesError,
            InvalidValueError,
            MissingPropertiesError,
            STIXError,
        ) as e:
            raise STIX2TransformError('Failed to create STIX2 bundle from risklist. {}'.format(e))

        return bundle

    @classmethod
    def from_analyst_note(
        cls,
        note: RFNote,
        attachment: bytes = None,
        split_snort: bool = False,
        identity: Identity = None,
    ) -> Bundle:
        """Creates a STIX2 bundle from a Recorded Future analyst note.

        Args:
            note (RFNote): Recorded Future analyst note
            attachment (bytes, optional): Note attachment. Defaults to None.
            split_snort (bool, optional): Split snort rules into separate DetectionRule objects.
            Defaults to False.
            identity (stix2.Identity, optional): Author. Defaults to Recorded Future.

        Returns:
            Bundle: STIX2 bundle
        """
        LOG.info('Creating STIX2 bundle from analyst note {}'.format(note.note_id))

        if not identity:
            identity = create_rf_author()
        objects = [identity]

        topics = [topic['name'] for topic in note.topics]
        report_types = create_report_types(topics)
        for entity in note.note_entities:
            try:
                stix_entity = convert_entity(entity['name'], entity['type'])
                if isinstance(stix_entity, Indicator):
                    objects.extend(stix_entity.stix_objects)
                else:
                    objects.append(stix_entity.stix_obj)

            except UnsupportedConversionType as err:  # noqa: PERF203
                LOG.warning(str(err) + '. Skipping...')
                continue

        if attachment and note.detection_rule_type in SUPPORTED_HUNTING_RULES:
            # This is a workaround for OpenCTI
            # OpenCTI does not support multiple Snort rules within a single DetectionRule object
            # so we split them into separate objects (split_snort = True)
            if note.detection_rule_type == 'snort' and split_snort is True:
                objects.extend(split_snort_rules(note, str(attachment, 'UTF-8')))
            else:
                rule = DetectionRule(
                    name=note.title,
                    type_=note.detection_rule_type,
                    content=str(attachment, 'UTF-8'),
                )
                objects.append(rule.stix_obj)

        external_references = generate_external_references(note.validation_urls)
        external_references.append(
            {
                'source_name': 'Recorded Future',
                'url': note.portal_url,
            },
        )

        report = Report(
            name=note.title,
            created_by_ref=identity.id,
            description=note.text,
            published=note.published,
            labels=topics,
            report_types=report_types,
            object_refs=[obj.id for obj in objects],
            object_marking_refs=TLP_MAP['amber'],
            external_references=external_references,
        )
        objects.append(report)

        bundle = Bundle(objects=objects, allow_custom=True)

        return bundle


# Helpers for bundle creation


def split_snort_rules(note: RFNote, attachment: str) -> list:
    """Splits snort rules into multiple DetectionRule objects.

    Args:
        note (RFNote): original note object from psengine
        attachment (str): snort rule payload
    """
    rules = []
    ctr = 1
    temp_description = []
    for line in attachment.split('\n'):
        line = line.strip()

        # skip comments and empty lines
        if line.startswith('#') or not line:
            temp_description.append(line[1:].strip())
            continue
        rules.append(
            DetectionRule(
                name=note.title + ', Rule {}'.format(ctr),
                type_=note.detection_rule_type,
                content=line,
                description='\n'.join(temp_description),
            ).stix_obj,
        )
        # reset description for next rule
        temp_description = []
        ctr += 1
    return rules


def create_report_types(topics: list) -> list:
    ret = set()
    for topic in topics:
        if topic not in REPORT_TYPE_MAPPER:
            LOG.warning('Could not map a report type for type {}'.format(topic))
            continue
        ret.add(REPORT_TYPE_MAPPER[topic])
    return list(ret)


def generate_external_references(urls: list) -> list:
    """Generate External references from validation urls."""
    refs = []
    for url in urls:
        source_name = url.split('/')[2].split('.')[-2].capitalize()
        refs.append({'source_name': source_name, 'url': url})
    return refs
