##################################### TERMS OF USE ###########################################
# The following code is provided for demonstration purpose only, and should not be used      #
# without independent verification. Recorded Future makes no representations or warranties,  #
# express, implied, statutory, or otherwise, regarding any aspect of this code or of the     #
# information it may retrieve, and provides it both strictly “as-is” and without assuming    #
# responsibility for any information it may retrieve. Recorded Future shall not be liable    #
# for, and you assume all risk of using, the foregoing. By using this code, Customer         #
# represents that it is solely responsible for having all necessary licenses, permissions,   #
# rights, and/or consents to connect to third party APIs, and that it is solely responsible  #
# for having all necessary licenses, permissions, rights, and/or consents to any data        #
# accessed from any third party API.                                                         #
##############################################################################################

import warnings

from psengine.playbook_alerts.constants import DEPRECATION_MSG

from .rf_base_playbook_alert import RFBasePlaybookAlert


class RFPlaybookAlertCyberVulnerability(RFBasePlaybookAlert):
    """Playbook Alert for Cyber Vulnerability."""

    @property
    def lifecycle_stage(self) -> str:
        """Cyber Vulnerability: Get playbook alert lifecycle_stage.

        Returns:
            str: vulnerability lifecycle stage or None if not found
        """
        stage = self.panel_status.get('lifecycle_stage', None)
        if stage is None:
            stage = self.panel_evidence_summary.get('summary', {}).get('lifecycle_stage', None)

        return stage

    @property
    def risk_rules(self) -> list:
        """Cyber Vulnerability: Get playbook alert risk_rules.

        Each rule contains:
            rule
            description

        Returns:
            list: rule and description or None if not found
        """
        return self.panel_evidence_summary.get('summary', {}).get('risk_rules', [])

    @property
    def affected_products(self) -> list:
        """Cyber Vulnerability: Get playbook alert affected_products.

        Returns:
            list: affected products or None if not found
        """
        return self.panel_evidence_summary.get('affected_products', [])

    @property
    def insikt_notes(self) -> list:
        """Cyber Vulnerability: Get vulnerability Insikt notes.

        Each note contains:
            id
            title
            published
            topic
            fragment

        Returns:
            list: Insikt notes or None if not found
        """
        return self.panel_evidence_summary.get('insikt_notes', [])

    @property
    def log_v2_vulnerability_lifecycle_change(self) -> list:
        """Vulnerability Lifecycle change"""
        return [
            changes
            for changes in self.log_v2_changes
            if changes['type'] == 'lifecycle_in_cve_change'
        ]

    @property
    def log_vuln_changes(self) -> list:
        """Cyber Vulnerability: Filter for log panels with context type cyber_vulnerability.

        Returns:
            list: vulnerability changes or [] if not found
        """
        warnings.warn(DEPRECATION_MSG, DeprecationWarning)
        return list(
            filter(
                lambda log: log.get('context', {}).get('type') == 'cyber_vulnerability',
                self.log_context_changes,
            ),
        )

    @property
    def log_related_entities_changes(self) -> list:
        """Cyber Vulnerability: Filter for log panels with changes to related entities.

        Return:
            list: related entities changes or [] if not found
        """
        warnings.warn(DEPRECATION_MSG, DeprecationWarning)
        return list(
            filter(
                lambda log: log.get('changes', {}).get('related_entities_change'),
                self.log_changes,
            ),
        )

    @property
    def log_lifecycle_in_cve_change(self) -> list:
        """Cyber Vulnerability: Filter for log panels with lifecycle_in_cve_changes.

        Return:
            list: log panel changes with context type cyber_vulnerability containing only
                  lifecycle_in_cve_changes, or [] if not found
        """
        warnings.warn(DEPRECATION_MSG, DeprecationWarning)
        return self._get_specific_context_changes('cyber_vulnerability', 'lifecycle_in_cve_change')
